﻿# -*- coding: utf-8 -*-

import sys, math

import bpy
from mathutils import Matrix

from .. import Chunk_Object
from . import Meshes, Curves, Opts, Material

if(sys.modules["Load_LWO"].DO_RELOAD):
	import importlib
	loc = set(locals().keys())
	for f in ("Meshes", "Curves", "Opts", "Material", "Chunk_Object",):
		if(f in loc):
			importlib.reload(eval(f))
	del f, loc

from ..LXO_ParseHeader import PolyIdT

#### =======================================================================
__all__ = ("BuildObjects")

class __RotateOrderT(Chunk_Object.ChunkObject):
	# obj.rotateOrder -> ob.rotation_mode 座標系自体の差異を吸収.
	xyz = "XZY"
	yzx = "ZYX"
	zxy = "YXZ"
	xzy = "XYZ"
	yxz = "ZXY"
	zyx = "YZX"
	__slots__=()

#### =======================================================================
def BuildObjects(context, lxo, opts= Opts.OptsObject()):
	u'''ブレンダーデータ作成, LXO'''
	if(not (opts.buildMesh | opts.buildCrv)):
		return

	deteminedMtlDct = Material.Make(lxo.surfs, lxo.clips, opts)	# 先行して作成

	for lyr in lxo.layers:
		if(not (lyr.isHide & opts.ignoreHide)):
			# 先にpivとpntsを改変してしまう
			obLocation = Meshes.PreCalcVertsBasePos(lyr, opts.scale, opts.usePivot)

			if(opts.buildMesh and lyr.polNum):
				(meOB, _) = Meshes.MakeMesh(context, lyr, lxo.tags, lxo.surfs, deteminedMtlDct, opts, obLocation)
				if(meOB):
					obj = lxo.meshes[lyr.uniqueIdentifier]
					obj.buildedObject = meOB
					obj.sdsLvl = lyr.sdsLvl
					obj.isSds = bool(lyr.lwid in {PolyIdT.PSUB, PolyIdT.SUBD})

			if(opts.buildCrv and lyr.crvNum):
				crvOBs = Curves.MakeCurve(context, lyr, opts, obLocation)
				obj = lxo.meshes[lyr.uniqueIdentifier]
				obj.buildedCrvs.extend(crvOBs)

	# mesh(instance)作成
	__MeshInstancing(context, lxo)

	totalDct = lxo.transforms.copy()
	totalDct.update(lxo.locators)
	totalDct.update(lxo.meshes)

	__SetMeshesNodeAttr(context, lxo, True, opts.scaling)

	__MakeNodes(context, lxo, opts.scaling, totalDct)

	__ParentAndMove(lxo, opts.scaling, totalDct)

	__Rename(lxo)

#### ========================================================================
def __Make_Transform(context, obj, optScl):
	ob = bpy.data.objects.new(obj.name, None)
	context.view_layer.active_layer_collection.collection.objects.link(ob)	# link->active->modeを遵守
	ob.hide_viewport = obj.isVis ^ 1

	# Blenderに空のtransformは無い
	ob.empty_display_type = "PLAIN_AXES"
	ob.empty_display_size = 0.0001

	# Blenderにtransform層のpivotは無い. ツール用transform_pivot_pointは無意味.
	# vert移動のフェイク(ob.set_origin)はあるけど微妙. 3Dカーソル移動→set_origin(cur)の流れ.
	# FnCalc.VScl(obj.pivot, optScl)
	return(ob)

#### ------------------------------------------------------------------------
def __Make_Locator(context, obj, optScl):
	ob = __Make_Transform(context, obj, optScl)
	ob.empty_display_type = "PLAIN_AXES"
	ob.empty_display_size = obj.size * optScl
	return(ob)

#### ------------------------------------------------------------------------
def __SetMeshesNodeAttr(context, lxo, useWire, optScl):
	# layerから作成された/インスタンス作成した分のattrを設定
	for obj in lxo.meshes.values():
		if(obj.buildedObject):
			obj.buildedObject.hide_viewport = obj.isVis ^ 1		# Meshのtransform
			if(obj.isSds):
				# bpy.types.SubsurfModifier
				md = obj.buildedObject.modifiers.new("Subdivision", "SUBSURF")
				md.subdivision_type = "CATMULL_CLARK"
				md.levels = md.render_levels = int(obj.sdsLvl)
				md.uv_smooth = "PRESERVE_CORNERS"

			if(useWire):
				if(obj.lyrRGBA is not None):
					obj.buildedObject.color = obj.lyrRGBA

#### ------------------------------------------------------------------------
def __MakeNodes(context, lxo, optScl, totalDct):
	u'''trs/locの既存Nodeとの重複を考慮しない場合の作成'''
	tmp = []
	EXT = tmp.extend
	for t in [ obj.treeParentsList for (uniq, obj) in lxo.meshes.items() if(obj.buildedObject) ]:
		EXT(t)
	meshUniqNamesSet = set(tmp)

	# 未作成のmeshが作成済meshの階層経路になる場合は代替の空transformを作る
	# 経路判定をしない場合は, curveのみのmeshを含まないlyrを考慮しないといけない
	# 必要な経路に含まれているなら作る

	for (uniq, obj) in lxo.meshes.items():
		if(obj.buildedObject is None):
			if(uniq in meshUniqNamesSet):
				obj.buildedObject = __Make_Transform(context, obj, optScl)
				obj.isPlaceHolder = True
				break

	for (uniq, obj) in lxo.transforms.items():
		if(uniq in meshUniqNamesSet):
			obj.buildedObject = __Make_Transform(context, obj, optScl)

	for (uniq, obj) in lxo.locators.items():
		if(uniq in meshUniqNamesSet):
			obj.buildedObject = __Make_Locator(context, obj, optScl)

#### ========================================================================
def __ParentAndMove(lxo, optScl, totalDct):
	u'''親子組'''
	def __RoundRadian(rad):
		return(math.radians(round(math.degrees(rad) * 100000.0) * 0.00001))
	#### ----------------
	def __Round_Scaling(v, scl):
		return(round(v * 100000000.0) * 0.00000001 * scl)
	#### ----------------
	for obj in totalDct.values():
		hasCrv = False if(obj.type != "mesh") else bool(obj.buildedCrvs)

		if(obj.buildedObject is None):		# mesh/instance無し?
			if(not hasCrv):					# カーブ要素も無いなら次へ
				continue
		# mesh/curve何れかがある

		if(obj.parentUniqueIdentifier is not None):
			parentOb = totalDct[obj.parentUniqueIdentifier].buildedObject
			if(obj.buildedObject):
				obj.buildedObject.parent = parentOb
			if(hasCrv):
				for ob in obj.buildedCrvs:
					ob.parent = parentOb

		# lxoは単精度なので精度が低い. 3.0→2.9~9nnnnなので丸めて8桁で丸めてからスケールする
		tx = __Round_Scaling(obj.translate[0], optScl)
		tz = __Round_Scaling(obj.translate[1], optScl)
		ty = __Round_Scaling(obj.translate[2],-optScl)
		pos = (tx,ty,tz)

		# lxoは単精度のradianなので精度が低い. deg変換してから丸めてradianに戻す
		rx = __RoundRadian(obj.rotate[0])
		rz = __RoundRadian(obj.rotate[1])
		ry = __RoundRadian(-obj.rotate[2])
		rot = (rx,ry,rz)

		scl = (obj.scale[0], obj.scale[2], obj.scale[1])

		if(obj.buildedObject):
			ob = obj.buildedObject
			ob.location = pos
			ob.rotation_euler = rot
			ob.rotation_mode = __RotateOrderT[obj.rotateOrder]
			ob.scale = scl

		if(hasCrv):
			for ob in obj.buildedCrvs:
				ob.location = pos
				ob.rotation_euler = rot
				ob.rotation_mode = __RotateOrderT[obj.rotateOrder]
				ob.scale = scl

#### ========================================================================
def __Rename(lxo):
	u'''リネーム'''
	def __Rename_Transform(obj):
		if(obj.buildedObject):
			obj.buildedObject.name = obj.name
	#### ----------------
	def __Rename_Mesh(obj):
		if(obj.buildedObject):
			if(obj.isPlaceHolder):				# 未作成Meshのダミーtransform
				obj.buildedObject.name = "{}_PlaceHolder".format(obj.name)
			else:								# 通常作成/インスタンス
				obj.buildedObject.name = obj.name
				if(obj.ownItem.type != "meshInst"):		# インスタンス時はシェイプの変更はしない
					# transformの変化後の名を使用
					obj.buildedObject.data.name = "{}Shape".format(obj.buildedObject.name)
	#### ----------------
	for obj in lxo.transforms.values():
		__Rename_Transform(obj)

	for obj in lxo.meshes.values():
		__Rename_Mesh(obj)

	for obj in lxo.locators.values():
		__Rename_Transform(obj)

#### ========================================================================
def __MeshInstancing(context, lxo):
	u'''インスタンス作成'''
	for (uniq, obj) in lxo.meshes.items():
		if("meshInst" == obj.ownItem.type):
			if("source" in obj.ownItem.itemLinks):
				srcObj = lxo.meshes[lxo.items[obj.ownItem.itemLinks.source[0].id].uniqueIdentifier]
				if(srcObj.buildedObject is not None):
					# srcMeshが作成されていたら作成. Shapeだけインスタンス.
					ob = bpy.data.objects.new(obj.name, srcObj.buildedObject.data)
					context.view_layer.active_layer_collection.collection.objects.link(ob)	# link->active->modeを遵守
					obj.buildedObject = ob

					# SDSかどうかの継承. BlenderはObject(transform)で処理する.
					obj.isSds  = srcObj.isSds
					obj.sdsLvl = srcObj.sdsLvl

					# Modoでのインスタンス→変更可能なtrs&mesh.
					# Modoではtrsに明示的な色指定が無い場合は元trsの状態が継承される.
					if(srcObj.lyrRGBA is not None):
						if(isinstance(obj.lyrRGBA, list)):	# list時: ModoのUI設定値
							obj.lyrRGBA = srcObj.lyrRGBA

				if(srcObj.buildedCrvs):		# [ob,ob,...]
					# カーブがあったら作成. Shapeだけインスタンス.
					for crv in srcObj.buildedCrvs:
						ob = bpy.data.objects.new(crv.name, crv.data)
						context.view_layer.active_layer_collection.collection.objects.link(ob)	# link->active->modeを遵守
						obj.buildedCrvs.append(ob)
