﻿# -*- coding: utf-8 -*-

import bpy

from .. import FnCalc

#### =======================================================================
__all__ = ("MakeCurve",)

#### =======================================================================
def MakeCurve(context, lyr, opts, obLoc):
	u'''カーブ作成'''
	if(opts.crvPerLyr):
		obs = __MakeCurve_PerLyr(context, lyr, opts.mxAxisConv, opts.degree)
	else:
		obs = __MakeCurve_PerCrv(context, lyr, opts.mxAxisConv, opts.degree)

	FnObjectLink = context.view_layer.active_layer_collection.collection.objects.link
	for ob in obs:
		FnObjectLink(ob)
		ob.hide_viewport = lyr.isHide	# 不可視だとEDITに入れなくなるので最後に.
		if(opts.usePivot):
			ob.location = obLoc

	return(obs)

#### --------------------------------
def __Container(name):
	co = bpy.data.curves.new(name, "CURVE")
	ob = bpy.data.objects.new(name, co)
	co.dimensions = "3D"
	return(co, ob)

#### --------------------------------
def __MakeCurve_PerLyr(context, lyr, mxAxisConv, deg):
	u'''1オブジェクトに全部入り作成, → 1 obj: 1 curve: N spline'''
	name = "{}_crv".format(lyr.name)
	(co, ob) = __Container(name)

	for crv in lyr.crvs:
		__MakeCurve_Nurbs(co, crv, deg, lyr.pnts)

	if(hasattr(lyr, "bcrvs")):
		for crv in lyr.bcrvs:
			__MakeCurve_Bezier(co, crv, deg, lyr.pnts)

	co.transform(mxAxisConv)

	return([ob])

#### --------------------------------
def __MakeCurve_PerCrv(context, lyr, mxAxisConv, deg):
	u'''カーブ毎にオブクジェクト作成, → N obj: 1 curve: 1 spline'''
	def __Build(lyr, crv, cnt, BuildFunc):
		name = "{}_crv{}".format(lyr.name, cnt)
		(co, ob) = __Container(name)
		BuildFunc(co, crv, deg, lyr.pnts)
		co.transform(mxAxisConv)
		return(ob)
	#### ----------------
	obs = []
	cnt = 1

	for crv in lyr.crvs:
		obs.append(__Build(lyr, crv, cnt, __MakeCurve_Nurbs))
		cnt += 1

	if(hasattr(lyr, "bcrvs")):
		for crv in lyr.bcrvs:
			obs.append(__Build(lyr, crv, cnt, __MakeCurve_Bezier))
			cnt += 1

	return(obs)

#### --------------------------------
def __MakeCurve_Nurbs(co, crv, deg, pnts):
	u'''ナーブズ'''
	# crv=[flag, idx,idx,,,]
	if(crv[0]):	# Closed
		pntIdxs  = crv[2:-2]	# モデラカーブの補助点と始終の重複を除外
		isClosed = True
	else:		# Open
		pntIdxs  = crv[1:]
		isClosed = False

	sp = co.splines.new("NURBS")
	sp.points.add(len(pntIdxs )-1)	# 必要数-1, new段階で[0]が発生している.
	sp.points.foreach_set("co", tuple( v for i in pntIdxs for v in (pnts[i] + (1.0,)) ))	# (xyzw)
	sp.order_u = deg + 1			# 2~6,def:4
	sp.use_endpoint_u = True
	sp.use_cyclic_u = isClosed		# 周期時はendpnt指定は無視される

#### --------------------------------
def __MakeCurve_Bezier(co, crv, deg, pnts):
	u'''ベジェ'''
	if(crv[1] == crv[-1]):					# Closed
		# crv==[None, 点A,出A, 入B,点B,出B, 入C,点C,出C,,,入A,点A]
		# close[XA, A,A, B,B,B, C,C,C], 入Aの移動と点A'を捨てる
		tmp = (crv[-2],) + crv[1:-2]
		ancs = [ (pnts[si], pnts[mi], pnts[ei]) for (si, mi, ei) in zip(*(iter(tmp),) * 3) ]
		isClosed = True
	else:									# Open
		# crv==[None, 点A,出A, 入B,点B,出B, 入C,点C,出C,,,入N,点N]
		# open[XA, A,A, B,B,B, C,C,C, N,N, XN], 入Aと出Nを仮置きしてから収集する
		tmp = (None,) + crv[1:] + (None,)
		ancs = []
		for (si, mi, ei) in zip(*(iter(tmp),) * 3):
			st = None if(si is None) else pnts[si]
			ed = None if(ei is None) else pnts[ei]
			ancs.append( (st, pnts[mi], ed) )
		isClosed = False

	sp = co.splines.new("BEZIER")

	num = len(ancs)
	sp.bezier_points.add(num-1)	# 必要数-1, new段階で[0]が発生している.
	mts = [ v for (_, mvec, __) in ancs for v in mvec ]
	sp.bezier_points.foreach_set("co", mts)	# (x,y,z,x,y..)

	for bz in sp.bezier_points.values():
		bz.handle_left_type = bz.handle_right_type = "FREE"

	lts = []
	for (svec, mvec, _) in ancs:
		if(svec is None):
			lts.extend(mvec)
		else:
			lts.extend(svec)
	sp.bezier_points.foreach_set("handle_left", lts)

	rts = []
	for (_, mvec, evec) in ancs:
		if(evec is None):
			rts.extend(mvec)
		else:
			rts.extend(evec)
	sp.bezier_points.foreach_set("handle_right", rts)

	# "LINEAR", "CARDINAL", "BSPLINE", "EASE"
	sp.radius_interpolation = sp.tilt_interpolation = "LINEAR"
	sp.use_bezier_u = 1#False
	sp.resolution_u = 16
	sp.order_u = deg + 1	# 2~6,def:4

	sp.use_cyclic_u = isClosed		# 周期時はendpnt指定は無視される
