﻿# -*- coding: utf-8 -*-

import bpy
import bmesh

from .. import FnCalc

#### =======================================================================
__all__ = ("MakeMesh", "PreCalcVertsBasePos",)

#### =======================================================================
class BldData(object):
	__slots__ = ("newPolNum", "totalPerPolPntNum", "loop_start", "loop_total", "vertex_index")
	def __init__(self):
		self.newPolNum = 0			# 実際に作られる面ポリの数
		self.totalPerPolPntNum = 0	# ポリ毎の頂点数の累積, len(me.loops)等価
		self.loop_start = None		# mesh.polygons.loop_start, 各面の開始頂点idx, (0,4,7,10,13...)
		self.loop_total = None		# mesh.polygons.loop_total, 面の頂点数, (4,4,3,3,5...)
		self.vertex_index = None	# me.loops.vertex_index, pols/pntIdxの列挙, 頂点順はLWの逆

#### =======================================================================
def MakeMesh(context, lyr, tags, surfs, deteminedMtlDct, opts, obLoc):
	u'''メッシュ作成'''
	FnObjectLink = context.view_layer.active_layer_collection.collection.objects.link
	me = bpy.data.meshes.new(lyr.name)
	ob = bpy.data.objects.new(me.name, me)
	FnObjectLink(ob)							# link->active->modeを遵守
	bpy.context.view_layer.objects.active = ob	# OBJECTに入る為選択
	bpy.ops.object.mode_set(mode="OBJECT")		#

	# 前処理
	(bldData, buildedPolIdxs, edgeIdxs, vertIdxs) = __BuildBasisArrays(lyr.polNum, lyr.pols)

	# メッシュ作成
	__BuildGeom(me, lyr, bldData, edgeIdxs, vertIdxs, opts.mxAxisConv)

	if(opts.useEdge and lyr.hasEdge):		# エッジウェイト
		__ApplyEdgeWeight(me, lyr.edgeWgts)

	# マテリアル割当
	orderdSurfNewPolIdxs = __AssignMaters(me, lyr, tags, deteminedMtlDct, buildedPolIdxs, frozenset(edgeIdxs+vertIdxs))
	__ApplySmooth(me, surfs, bldData.newPolNum, orderdSurfNewPolIdxs)

	if(opts.useNorm and lyr.normals):		# NORM法線
		usedNormPerSurf = __PreApplyNormal(lyr.normals, surfs, orderdSurfNewPolIdxs)
		if(usedNormPerSurf):
			__ApplyNormal(me, lyr.pols, lyr.normals, usedNormPerSurf, buildedPolIdxs, bldData.totalPerPolPntNum, bldData.loop_start, opts.mxAxisConv)

	if(lyr.uvs):							# UVマップ
		__AssignUVs(me, lyr.pols, lyr.uvs)

	if(opts.useMorf and (lyr.absMorphs or lyr.relMorphs)):	# モーフマップ
		__AssignMorph(ob, me, lyr, opts.mxAxisConv, opts.scale, opts.usePivot)

	if(opts.useWght and lyr.weights):		# ウェイトマップ
		buildedVertGrps = __AssignWeights(ob, lyr.weights)
	else:
		buildedVertGrps = [[],[]]

	if(opts.useSels and lyr.picks):			# 選択セット
		__AssignSelectionSet(ob, lyr.picks)

	if(opts.useVCol):						# 頂点カラー
		if(lyr.rgb):
			__AssignRGBA(me, lyr.pols, lyr.rgb)
		if(lyr.rgba):
			__AssignRGBA(me, lyr.pols, lyr.rgba)

	if(opts.useColr and lyr.tagColors):		# カラータグ
		__AssignPTag(ob, "COLR_{}", lyr.tagColors, buildedPolIdxs)

	if(opts.usePart and lyr.tagParts):		# パートタグ
		__AssignPTag(ob, "PART_{}", lyr.tagParts, buildedPolIdxs)

	if(hasattr(lyr, "tagSelects")):
		if(opts.usePick and lyr.tagSelects):	# 選択タグ
			__AssignPTag(ob, "PICK_{}", lyr.tagSelects, buildedPolIdxs)

	valid = me.validate(clean_customdata=False)

	# 未使用頂点の掃除は最後
	__RemoveUnusedVert(me, lyr.pntNum, lyr.pols, bldData.vertex_index, edgeIdxs, vertIdxs)

	bpy.ops.object.mode_set(mode="OBJECT")

	if(opts.usePivot):
		ob.location = kwd.get("obLoc", (0.0, 0.0, 0.0))

	ob.hide_viewport = lyr.isHide	# 不可視だとEDITに入れなくなるので最後に.

	return(ob, buildedVertGrps)

#### =======================================================================
def __BuildBasisArrays(polNum, pols):
	lStart = []
	lTotal = []
	vIndex = []
	edgeIdxs = []	# 2点は別枠なので抽出
	vertIdxs = []	# 1点として未使用頂点(Bone,Curve)と区別して抽出
	buildedPolIdxs = [None] * polNum	# [0,-,1,2,-,3,...], 元[2]は新1, 元[4]はcrv/bone等

	stAPD = lStart.append
	toAPD = lTotal.append
	viEXT = vIndex.extend
	eiAPD = edgeIdxs.append
	vtAPD = vertIdxs.append

	totalPerPolPntNum = polCnt = 0
	for (i, pntIdxs) in enumerate(pols):
		n = len(pntIdxs)
		if(n > 2):		# 面ポリ
			toAPD(n)
			stAPD(totalPerPolPntNum)
			totalPerPolPntNum += n

			viEXT(pntIdxs[::-1])	# 逆順

			# 後で照合されるpols内のidxと作成した結果(1/2点ポリ除外)のidxが食違う事への対処
			buildedPolIdxs[i] = polCnt
			polCnt += 1
		elif(n == 2):	# 線ポリ, 抽出を個別のループで回すのとどちらが速いかは未確認
			eiAPD(i)
		else:			# 点ポリ
			vtAPD(i)

	bldData = BldData()
	bldData.newPolNum = polCnt
	bldData.totalPerPolPntNum = totalPerPolPntNum
	bldData.loop_start   = tuple(lStart)
	bldData.loop_total   = tuple(lTotal)
	bldData.vertex_index = tuple(vIndex)

	return(bldData, tuple(buildedPolIdxs), tuple(edgeIdxs), tuple(vertIdxs))

#### =======================================================================
def __BuildGeom(me, lyr, bldData, edgeIdxs, vertIdxs, mxAxisConv):
	u'''ジオメトリ作成'''
	me.vertices.add(lyr.pntNum)
	me.vertices.foreach_set("co", tuple( v for vec in lyr.pnts for v in vec ))

	me.loops.add(bldData.totalPerPolPntNum)
	me.loops.foreach_set("vertex_index", bldData.vertex_index)

	me.polygons.add(bldData.newPolNum)
	me.polygons.foreach_set("loop_start", bldData.loop_start)
	me.polygons.foreach_set("loop_total", bldData.loop_total)

	if(edgeIdxs):						# 線ポリとしてのエッジ
		me.edges.add(len(edgeIdxs))		# each_setでの一括処理の為update()前の空の状態で適用
		me.edges.foreach_set("vertices", tuple( pntIdxs for ei in edgeIdxs for pntIdxs in lyr.pols[ei] ))

	me.update(calc_edges=True)			# エッジを確定させるとポリを構成するエッジ(loop?)でedgesが埋まる

	me.transform(mxAxisConv)

#### -----------------------------------------------------------------------
def __RemoveUnusedVert(me, pntNum, pols, vertex_index, edgeIdxs, vertIdxs):
	u'''面線点で使用されない(ボーン/カーブ)頂点を除去'''
	bpy.ops.object.mode_set(mode="EDIT")

	bm = bmesh.from_edit_mesh(me)
	unusedVertList = __UnusedVert(pntNum, pols, vertex_index, edgeIdxs, vertIdxs)
	for i in sorted(unusedVertList, reverse=True):
		bm.verts.ensure_lookup_table()
		bm.verts.remove(bm.verts[i])
	# to_mesh()不要
	bm.free()

#### --------------------------------
def __UnusedVert(pntNum, pols, vertex_index, edgeIdxs, vertIdxs):
	polSet = set(vertex_index)
	if(edgeIdxs):
		polSet.update( pntIdxs for ei in edgeIdxs for pntIdxs in pols[ei] )
	if(vertIdxs):
		polSet.update( pntIdxs for vi in vertIdxs for pntIdxs in pols[vi] )
	return(list(set(range(pntNum)) - polSet))

#### -----------------------------------------------------------------------
def __AssignMorph(ob, me, lyr, mxAxisConv, scale, usePivot):
	u'''モーフ作成'''
	# data.meshes[ME].shape_keys[K].key_blocks[SK] == ME.K.key_blocks[SK] == data.shape_keys[K].key_blocks[SK]
 	# ob.shape_key_addは if(has Key){SK追加}else{Key作成&SK追加&Key.reference_key=SK} Return(SK)
	sk = ob.shape_key_add(name= "Basis")
	me.shape_keys.use_relative = True
	me.shape_keys.name = "{}_key".format(me.name)

	# Basis用は全頂点なのでforeach_get&setの方が速い, transformの計算精度誤差がどうでるかは不明
	vecs = [None] * (lyr.pntNum * 3)		# 必要サイズのlistを先に用意
	me.vertices.foreach_get("co", vecs)		# 戻りは1次元化された物((),())->(,,,)
	sk.data.foreach_set("co", vecs)

	# 絶対相対共に全pntが値を持つ保証は無い, me.transformはmorfに作用しないので自前変換する
	for (name, dct) in sorted(list(lyr.absMorphs.items()), key=lambda x: x[0]):	# morph名でソート
		sk = ob.shape_key_add(name= name)
		if(scale != (1.0,1.0,1.0)):
			if(usePivot):
				for (pntIdx, pos) in dct.items():
					sk.data[pntIdx].co = FnCalc.VTransform(FnCalc.VMulSub(pos, scale), mxAxisConv)
			else:
				for (pntIdx, pos) in dct.items():
					sk.data[pntIdx].co = FnCalc.VTransform(FnCalc.VMul(pos, scale), mxAxisConv)
		else:
			if(usePivot):
				for (pntIdx, pos) in dct.items():
					sk.data[pntIdx].co = FnCalc.VTransform(FnCalc.VSub(pos, lyr.piv), mxAxisConv)
			else:
				for (pntIdx, pos) in dct.items():
					sk.data[pntIdx].co = FnCalc.VTransform(pos, mxAxisConv)

	# 相対モーフは絶対値にする
	for (name, dct) in sorted(list(lyr.relMorphs.items()), key=lambda x: x[0]):	# morph名でソート
		sk = ob.shape_key_add(name= name)
		if(scale != (1.0,1.0,1.0)):
			for (pntIdx, pos) in dct.items():
				sk.data[pntIdx].co = FnCalc.VTransform(FnCalc.VMulAdd(pos, scale, lyr.pnts[pntIdx]), mxAxisConv)
		else:
			for (pntIdx, pos) in dct.items():
				sk.data[pntIdx].co = FnCalc.VTransform(FnCalc.VAdd(lyr.pnts[pntIdx], pos), mxAxisConv)

#### -----------------------------------------------------------------------
def __AssignMaters(me, lyr, tags, deteminedMtlDct, buildedPolIdxs, excludeSet):
	u'''マテリアル割り当て'''
	orderdSurfPols = [None] * len(lyr.surfTags)
	APD = orderdSurfPols.append

	idxPlot = [None] * lyr.polNum	# 最大値として元の数

	matAPD = me.materials.append
	mtlIdx = 0
	for (idx, polIdxs) in sorted(list(lyr.surfTags.items()), key=lambda x: tags[x[0]]):	# surface名でソート.

		surfName = tags[idx]
		# lxoではptagと紐付くsurfが無い(値取得していない)物もある事への対応.
		# ModoはmaterialGroup(PTAG)の明示的割り当てが無い場合 PTAG:"Default"になる。
		# 明示的にDefaultのPTAGを作れる。Defaultとdefaultの2者作成も可。
		if(surfName not in deteminedMtlDct):
			matAPD(None)	# BlenderにMayaSG層相当は無い. Mat未接続としてNone
		else:
			matAPD(deteminedMtlDct[surfName])

		surfsPolIdxsSet = frozenset(polIdxs) - excludeSet	# surf内idxから線点idxを除外
		for j in surfsPolIdxsSet:
			idxPlot[j] = mtlIdx		# [-,A,-,A,B,B,...], mtlIdxをプロットしていく

		# surf毎, 旧idx->新idx
		orderdSurfPols[mtlIdx] = (surfName, [ buildedPolIdxs[j] for j in surfsPolIdxsSet ])

		mtlIdx += 1

	# 隙間を抜くと新idxでの一連のmtlIdxListになる
	me.polygons.foreach_set("material_index", tuple(i for i in idxPlot if(i != None)))

	return(orderdSurfPols)

#### -----------------------------------------------------------------------
def __ApplyEdgeWeight(me, edgeWgts):
	u'''CCエッジウェイト適用'''
	me.use_customdata_edge_crease = True
	for ed in me.edges.values():	# items==(idx,edge)なのでvalue()で.
		wgt = edgeWgts.get( (ed.vertices[0], ed.vertices[1]) )
		if(wgt is not None):
			ed.crease = min(max(0.0, wgt), 1.0)	# 保険でクランプ

#### -----------------------------------------------------------------------
def __ApplySmooth(me, surfs, newPolNum, orderdSurfNewPolIdxs):
	u'''スムース/フラットシェード. MeshProp->normal->autoSmooth(法線スムージング)とは違う'''
	smoothLst = [False] * newPolNum
	for (name, newPolIdxs) in orderdSurfNewPolIdxs:
		# lxoでは明示的なmat割当が無い場合がある
		if(name in surfs):
			if(surfs[name].sman > 0.0):	# 無効時のsmanは－値
				for i in newPolIdxs:
					smoothLst[i] = True
	me.polygons.foreach_set("use_smooth", smoothLst)

#### -----------------------------------------------------------------------
def __PreApplyNormal(normals, surfs, orderdSurfNewPolIdxs):
	u'''perPolPnt法線適用, 前処理'''
	usedNormPerSurf = []
	APD = usedNormPerSurf.append

	for (name, _) in orderdSurfNewPolIdxs:
		# lxoでは明示的なmat割当が無い場合がある
		if(name in surfs):
			norm = surfs[name].norm
			if(norm in normals):
				APD(norm)

	return(usedNormPerSurf)

#### --------------------------------
def __ApplyNormal(me, pols, normals, usedNormPerSurf, buildedPolIdxs, totalPerPolPntNum, loop_start, mxAxisConv):
	u'''perPolPnt法線適用'''
	me.calc_normals_split()		# loopsに法線値が出るようになる

	loops = [0.0] * (totalPerPolPntNum * 3)
	me.loops.foreach_get("normal", loops)	# transform済みの法線が戻る, 一次元化される((),())->(,,,)

	me.create_normals_split()

	for norm in usedNormPerSurf:
		for (polIdx, dct) in normals[norm].items():
			offs = loop_start[buildedPolIdxs[polIdx]] * 3	# ポリの点[0]のvec.xの位置を求める
			for pntIdx in reversed(pols[polIdx]):
				# LW系のnormalVecを変換
				(loops[offs], loops[offs+1], loops[offs+2]) = FnCalc.VTransform(dct[pntIdx], mxAxisConv)
				offs += 3
	me.normals_split_custom_set(tuple(zip(*(iter(loops),) * 3)))	# ((xyz),(xyz),)が必要

	me.use_auto_smooth = True

#### -----------------------------------------------------------------------
def __AssignUVs(me, pols, uvs):
	u'''UV割り当て'''
	for (name, vo) in uvs.items():
		if((not vo.vmap) and (not vo.vmad)):
			continue

		res = []
		EXT = res.extend
		cnt = 0
		for (polIdx, pntIdxs) in enumerate(pols):
			if(len(pntIdxs) < 3): continue

			if(polIdx in vo.vmad):
				for pntIdx in reversed(pntIdxs):
					uv = vo.vmad[polIdx].get(pntIdx)
					if(uv == None):
						uv = vo.vmap.get(pntIdx) or (0.0, 0.0)
					EXT(uv)
			else:
				for pntIdx in reversed(pntIdxs):
					EXT(vo.vmap.get(pntIdx) or (0.0, 0.0))

			cnt += len(pntIdxs)

		if(res):
			ul = me.uv_layers.new(name= name, do_init=False)
			ul.data.foreach_set("uv", res)

#### -----------------------------------------------------------------------
def __AssignRGBA(me, pols, rgba):
	u'''頂点カラー
2.90:
  vertex_colors.
2.92~:
  attributesが追加. vertex_colorsがエイリアス的な位置づけになった.
  vc作成 == attr("CORNER","BYTE_COLOR")作成.
3.2~
  vertex_colorsとしてのattributesのインターフェースとしてcolor_attributesが追加.
  vertex_colorsも機能してるのでインターフェースが3つある状態？
'''
	# 3.2からは新しいインターフェースで作成する.
	useNewAttr = bool((bpy.app.version[0]*10000) + (bpy.app.version[1]*100 + bpy.app.version[2]) >= 30200)

	for (name, vo) in rgba.items():
		if((not vo.vmap) and (not vo.vmad)):
			continue

		res = []
		EXT = res.extend
		cnt = 0
		for (polIdx, pntIdxs) in enumerate(pols):
			if(len(pntIdxs) < 3):
				continue

			if(polIdx in vo.vmad):
				for pntIdx in reversed(pntIdxs):
					colr = vo.vmad[polIdx].get(pntIdx)
					if(colr == None):
						colr = vo.vmap.get(pntIdx) or (1.0, 1.0, 1.0, 1.0)	# 白
					EXT(colr)
			else:
				for pntIdx in reversed(pntIdxs):
					EXT(vo.vmap.get(pntIdx) or (1.0, 1.0, 1.0, 1.0))	# 白

			cnt += len(pntIdxs)

		if(res):
			if(useNewAttr):
				# 元の頂点カラー自体はfloatなので新式ではあえてFLOAT_COLORを使う
				vc = me.color_attributes.new(name= name, type="FLOAT_COLOR", domain="CORNER")
			else:
				vc = me.vertex_colors.new(name= name, do_init=False)

			vc.data.foreach_set("color", res)

#### -----------------------------------------------------------------------
def __AssignWeights(ob, weights):
	u'''ウェイトを頂点グループで割り当て'''
	vgNames = []
	vgs = []
	nAPD = vgNames.append
	vAPD = vgs.append

	for (name, dct) in weights.items():
		vg = ob.vertex_groups.new(name= name)
		nAPD(vg.name)
		vAPD(vg)
		for (pntIdx, wght) in dct.items():
			vg.add((pntIdx,), min(max(0.0, wght), 1.0), "REPLACE")	# blenderは0～1のみ

	return( [vgNames, vgs] )

#### -----------------------------------------------------------------------
def __AssignSelectionSet(ob, picks):
	u'''セレクションセットを頂点グループで割り当て'''
	for (name, pntIdxs) in picks.items():
		vg = ob.vertex_groups.new(name= "Sels_{}".format(name))
		vg.add(pntIdxs, 1.0, "REPLACE")	# 便宜的に1.0

#### -----------------------------------------------------------------------
def __AssignPTag(ob, prefixFMT, dct, buildedPolIdxs):
	u'''PTag:Color/PartをFaceMapとして割り当て'''
	if(not dct):
		return

	for (name, polIdxs) in dct.items():
		if(not polIdxs):
			continue

		res = []
		APD = res.append
		for i in polIdxs:
			newIdx = buildedPolIdxs[i]
			if(newIdx != None):
				APD(newIdx)

		if(res):
			fm = ob.face_maps.new(name= prefixFMT.format(name))
			fm.add(res)

#### =======================================================================
def PreCalcVertsBasePos(lyr, scale, usePivot):
	if(scale != (1.0,1.0,1.0)):
		if(usePivot):
			lyr.piv  = tuple(FnCalc.VMul(lyr.piv, scale))
			lyr.pnts = tuple( tuple(FnCalc.VMulSub(vec, scale, lyr.piv)) for vec in lyr.pnts )
		else:
			lyr.pnts = tuple( tuple(FnCalc.VMul(vec, scale)) for vec in lyr.pnts )
	else:
		if(usePivot):
			lyr.pnts = tuple( tuple(FnCalc.VSub(vec, lyr.piv)) for vec in lyr.pnts )

	if(usePivot):
		return((lyr.piv[0], lyr.piv[2], lyr.piv[1]))

	return((0.0, 0.0, 0.0))
