﻿# -*- coding: utf-8 -*-

import sys

import bpy
from bpy_extras.io_utils import ImportHelper
from bpy.props import BoolProperty, IntProperty, FloatProperty, EnumProperty, StringProperty, CollectionProperty

from .. import Parse_LWO, Parse_LXO
from . import Build_LWO, Build_LXO, Opts

if(sys.modules["Load_LWO"].DO_RELOAD):
	import importlib
	loc = set(locals().keys())
	for f in ("Parse_LWO", "Parse_LXO", "Build_LWO", "Build_LXO", "Opts",):
		if(f in loc):
			importlib.reload(eval(f))
	del f, loc

from .Opts import *

#### =======================================================================
__all__ = ("LoadLWO", "register", "unregister")

#### =======================================================================
class LoadLWO(bpy.types.Operator, ImportHelper):
	'''Load Lightwave LWO Files'''
	bl_idname = "import_scene.lwo"
	bl_label  = "Import LWO"
	bl_options= {"PRESET", "UNDO"}

	#### --------------------------------
	# ファイルブラウザがソートを覚えない事への対策.
	# FileSelectParamsと同じpropはinvoke{fileselect_all()}後のSpaceFileBrowser.paramsに継承される.
	sort_method: EnumProperty(
		items=(
			("FILE_SORT_ALPHA", "","", 1),		# idxは1から, 未指定だと0からになりずれる.
			("FILE_SORT_EXTENSION", "","", 2),
			("FILE_SORT_TIME", "","", 3),
			("FILE_SORT_SIZE", "","", 4),
			),
		default="FILE_SORT_TIME")
	filter_glob: StringProperty(default="*.lwo;*.lxo", options={"HIDDEN"})
	files: CollectionProperty(type= bpy.types.OperatorFileListElement)
	directory: StringProperty(subtype="DIR_PATH")	# FILE_PATH同等

	#### --------------------------------
	for obj in (
		pMeshs,pCrvs,pBones,
		pMEdge,pMMorf,pMWght,pMSels,pMColr,pMPart,pMPick,pMNorm,pMVCol,
		pBWght,pBAMod,pBSPar,
		pGHide,pGPivt,pGMatl,pGImgs
		):
		__annotations__[obj.KEY] = BoolProperty(**obj.kwds())

	__annotations__[pCMode.KEY] = EnumProperty(**pCMode.kwds())
	__annotations__[pCDegr.KEY] = IntProperty(**pCDegr.kwds())
	__annotations__[pBLayr.KEY] = IntProperty(**pBLayr.kwds())
	__annotations__[pGScl.KEY]  = FloatProperty(**pGScl.kwds())

	meshDefines = (
		pMEdge, pMWght,
		pMMorf, pMVCol,
		pMSels, pMPick,
		pMColr, pMPart,
		pMNorm,
		)
	#### ------------------------------------------------------------------------
	def execute(self, context):
		'''self.filepathはImportHelperで定義. execは同invoke(fileselect_add)を経由して起動される'''
		kwd = self.as_keywords(ignore=("sort_method", "filter_glob", "files", "directory"))
		kwd[pCMode.KEY] = bool(kwd[pCMode.KEY] == pCMode.DEF)
		opts = OptsObject(**kwd)

		# 単品ならself.filepathを見ればいい
		selectedFiles = tuple( "{}{}".format(self.directory, ob.name) for ob in self.files if(ob.name.lower().endswith((".lwo",".lxo"))) )
		for fileName in selectedFiles:
			if(fileName.lower().endswith(".lwo")):
				inst = Parse_LWO.ParseLWO(fn= fileName)
			elif(fileName.lower().endswith(".lxo")):
				inst = Parse_LXO.ParseLXO(fn= fileName)
			else:
				continue

			if(not inst.lwo.layers):
				continue

			if(inst.fileType == "LWO2"):
				Build_LWO.BuildObjects(context, inst.lwo, opts)
			else:
				Build_LXO.BuildObjects(context, inst.lwo, opts)

		return({"FINISHED"})

	#### ------------------------------------------------------------------------
	def draw(self, context):
		self.__UI_Common()
		self.__UI_Mesh()
		self.__UI_Crv()
		self.__UI_Bone()

	#### --------------------------------
	def __UI_Common(self):
		baseBox = self.layout.box()
		baseClm = baseBox.column()

		row = baseClm.row()
		row.scale_y = .8
		row.use_property_split = True
		row.label(text="")
		row.prop(self, pGScl.KEY)

		row = baseClm.row()
		row.scale_y = 0.82
		row.prop(self, pGHide.KEY, icon= pGHide.ICON[self.ignoreHide])
		row.prop(self, pGPivt.KEY, icon= pGPivt.ICON[self.usePivot])

		# mtl→imageの為に分断
		row = baseClm.row()
		row.scale_y = 0.82
		row.prop(self, pGMatl.KEY, icon= pGMatl.ICON)

		row = row.row()
		row.prop(self, pGImgs.KEY, icon= pGImgs.ICON[self.useImgs])
		row.enabled = self.useExistMtl ^ 1

	#### --------------------------------
	def __UI_Mesh(self):
		baseBox = self.layout.box()
		baseClm = baseBox.column()

		baseClm.row().prop(self, pMeshs.KEY, icon= pMeshs.ICON)

		subClm = baseClm.box().column()

		flw = subClm.grid_flow(row_major=True, columns=2, even_columns=True, even_rows=False, align=False)
		flw.scale_y = .82
		for obj in self.meshDefines:
			flw.prop(self, obj.KEY, icon= obj.ICON)

		subClm.enabled = self.buildMesh

	#### --------------------------------
	def __UI_Crv(self):
		baseBox = self.layout.box()
		baseClm = baseBox.column()

		baseClm.row().prop(self, pCrvs.KEY, icon= pCrvs.ICON)

		subClm = baseClm.box().column()

		row = subClm.row()
		row.scale_y = .82
		row.prop(self, pCMode.KEY, expand=True)

		row = subClm.row()
		row.scale_y = .8
		row.use_property_split = True
		row.label(text="")
		row.prop(self, pCDegr.KEY)

		subClm.enabled = self.buildCrv

	#### --------------------------------
	def __UI_Bone(self):
		baseBox = self.layout.box()
		baseClm = baseBox.column()

		baseClm.row().prop(self, pBones.KEY, icon= pBones.ICON)

		subClm = baseClm.box().column()

#		flw = subClm.column_flow(columns=2)
#		flw.scale_y = .82
#		flw.prop(self, pBWght.KEY, icon= pBWght.ICON)
#		flw.scale_y = .8
#		flw.use_property_split = True
#		flw.prop(self, pBLayr.KEY)

		row = subClm.row()
		row.scale_y = .8
		row.use_property_split = True
		row.label(text="")
		row.prop(self, pBLayr.KEY)

		flw = subClm.grid_flow(row_major=True, columns=2, even_columns=True, even_rows=False, align=False)
		flw.scale_y = .82
		flw.prop(self, pBSPar.KEY, icon= pBSPar.ICON)
		flw.prop(self, pBWght.KEY, icon= pBWght.ICON)

		row = subClm.row()
		row.scale_y = .82
		row.prop(self, pBAMod.KEY, icon= pBAMod.ICON)

		subClm.enabled = self.buildBone

#### =======================================================================
def Add_menu(self, context):
	self.layout.operator(LoadLWO.bl_idname, text="Lightwave/Modo (.lwo, .lxo)")

def register():
	bpy.utils.register_class(LoadLWO)
	bpy.types.TOPBAR_MT_file_import.append(Add_menu)

def unregister():
	bpy.utils.unregister_class(LoadLWO)
	bpy.types.TOPBAR_MT_file_import.remove(Add_menu)
