﻿# -*- coding: utf-8 -*-

from __future__ import absolute_import

import sys, os, struct

from . import LWO_Object

# 重複の都合上, 必要な物だけを拾う
from .LWO_ParseHeader import ReadT, RSize
from .LWO_ParseHeader import ChunkIdT as lwoChunkIdT
from .LXO_ParseHeader import ChunkIdT as lxoChunkIdT

#### =======================================================================
__all__ = ("FnReadObject",)

#### ========================================================================
class FnReadObject(object):
	__slots__ = (
		"FnUnpckTAG",
		"FnUnpckTAGU2",
		"FnUnpckLWID",
		"FnUnpckFLAG",
		"FnUnpckV4",
		"FnUnpckV3",
		"FnUnpckANG",		# lwo
		"FnUnpckF4",
		"FnUnpckU4",
		"FnUnpckU2",
		"FnUnpckU1",
		"FnUnpckI4",
		"FnUnpckI2",
		"FnUnpckI1",
		"FnUnpckU2U2",
		"FnUnpckUV",
		"FnUnpckOPAC",		# lwo
		"FnUnpckISEQ",		# lwo
		"Read_VX",
		"Read_nameString",
		"FnPackLWID",
		)

	#### -----------------------------------------------------------------------
	def __init__(self, *arg, **kwd):
		u''''''
		self.FnUnpckTAG  = struct.Struct(ReadT.TAG).unpack
		self.FnUnpckTAGU2= struct.Struct(ReadT.TAGU2).unpack
		self.FnUnpckLWID = self.FnUnpckTAG
		self.FnUnpckFLAG = struct.Struct(ReadT.FLAG).unpack
		self.FnUnpckV4   = struct.Struct(ReadT.VEC4).unpack
		self.FnUnpckV3   = struct.Struct(ReadT.VEC3).unpack
		self.FnUnpckANG  = struct.Struct(ReadT.ANG4).unpack		# lwo
		self.FnUnpckF4   = struct.Struct(ReadT.F4).unpack
		self.FnUnpckU4   = struct.Struct(ReadT.U4).unpack
		self.FnUnpckU2   = struct.Struct(ReadT.U2).unpack
		self.FnUnpckU1   = struct.Struct(ReadT.U1).unpack
		self.FnUnpckI4   = struct.Struct(ReadT.I4).unpack
		self.FnUnpckI2   = struct.Struct(ReadT.I2).unpack
		self.FnUnpckI1   = struct.Struct(ReadT.I1).unpack
		self.FnUnpckU2U2 = struct.Struct(ReadT.U2U2).unpack
		self.FnUnpckUV   = struct.Struct(ReadT.UV).unpack
		self.FnUnpckOPAC = struct.Struct(ReadT.OPAC).unpack		# lwo
		self.FnUnpckISEQ = struct.Struct(ReadT.ISEQ).unpack		# lwo
		self.FnPackLWID  = struct.Struct(ReadT.LWID).pack

		# py2/3で関数切り替え.
		if(sys.version_info[0] == 2):
			self.Read_VX = self.Read_VX_Py2
			self.Read_nameString = self.Read_nameString_Py2
		else:
			self.Read_VX = self.Read_VX_Py3
			self.Read_nameString = self.Read_nameString_Py3

	#### ========================================================================
	@staticmethod
	def GetKeyFromKwd(kwd, sn, ln, defaultVal):
		return(kwd.get(sn, kwd.get(ln, defaultVal)))

	#### ========================================================================
	def _Do_Read(self, fileName):
		u'''ファイルロードの前後処理'''
		(fp, errMsg) = self.__Check_FileState(fileName)
		if(errMsg):
			print(errMsg),
			return(None)

		if(self.VisFlags.VERB):
			print(u"読み込み開始: {}".format(fileName))

		self._Read(fp)
		fp.close()

		if(self.VisFlags.VERB):
			print(u"読み込み終了.")

		return(True)

	#### ========================================================================
	def Read_nameString_Py2(self, data):
		u'''文字列読取'''
		# "" -> 00 00, idx:0, len:1, %:True, len:2
		# A -> nn 00, idx:1, len:2 %:False, len:2
		# LWID -> 4byte, idx:5, len:6 %:False, len:6
		# 文字が偶数時は+NULLで奇数, paddingで偶数化されてる
		# 文字が奇数時は+NULLで偶数, padding無し
		idx = data.find(b'\0')
		len = idx + (2 - (idx & 0x1))
		name = "" if(idx < 1) else data[:idx]
		return(name, len)

	#### --------------------------------
	def Read_nameString_Py3(self, data):
		u'''文字列読取'''
		idx = data.find(b'\0')
		len = idx + (2 - (idx & 0x1))

		if(idx < 1):
			name = ""
		else:
			# LWO2はcp932/utf8, LWO3はutf8. 何かは不明だから例外を拾って対処.
			try:
				name = data[:idx].decode("utf-8")
			except UnicodeDecodeError:
				name = data[:idx].decode("cp932")
		return(name, len)

	#### ========================================================================
	def Read_VX_Py2(self, data):
		u'''可変長U2/4読取, 先立ってord[0~4]するのはNG. U2だった時は[3~4]は存在しない.'''
		if(ord(data[0]) != 255):	# 2xではdata==str, ordで数値化.
			return(((ord(data[0]) << 8)| ord(data[1])), 2)
		return(((ord(data[1]) << 16)|(ord(data[2]) << 8)| ord(data[3])), 4)

	#### --------------------------------
	def Read_VX_Py3(self, data):
		u'''可変長U2/4読取, 先立ってord[0~4]するのはNG. U2だった時は[3~4]は存在しない.'''
		if(data[0] != 255):			# 3xではbytes, 直接扱う.
			return( ((data[0] << 8)| data[1]), 2)
		return( ((data[1] << 16)|(data[2] << 8)| data[3]), 4)

	#### ========================================================================
	def Read_LWID(self, curChunk):
		lwid = self.FnUnpckLWID(curChunk.read(RSize.TAG))[0]
		size = curChunk.chunksize - RSize.TAG
		return(lwid, size)

	#### ========================================================================
	def Read_Tags(self, data, dataSize):
		u'''各種Tag名. [[tag],[tag],...]
		文末の偶数化0埋めの分が空文字として含まれる."ab"\0[pad]"cde"\0"fg"...'''
		# 2xはdata==str, 3xは==bytes,  2xのfilterはlist, 3xはiter.
		# 後工程を考えbytes→str変換.1回/fileなので分岐コストは考えない.
		if(sys.version_info[0] > 2):
			data = data.decode()
		tags = tuple([ s for s in data.split('\0') if(s!="") ])

		self.DebugPrint_Tags(tags)
		return(tags)

	#### ========================================================================
	#### PNTS
	#### ========================================================================
	def Read_Pnts(self, data, dataSize, lyr):
		u'''頂点配列. [(x,y,z),(...),...]
		PNTS { point-location[VEC12] * }'''
#		APD = lyr.pnts.append
#		offs = 0
#		while(offs < dataSize):
#			APD( self.FnUnpckV3(data[offs:(offs + RSize.VEC3)]) )	# (x,y,z)
#			offs += RSize.VEC3
		lyr.pnts = [ self.FnUnpckV3(data[offs: offs + RSize.VEC3]) for offs in range(0, dataSize, RSize.VEC3) ]
		lyr.pntNum = len(lyr.pnts)

		self.DebugPrint_Pnts(lyr.name, lyr.pnts, lyr.pntNum)

	#### ========================================================================
	#### PTAG
	#### ========================================================================
	def Read_PTag_Surf(self, data, dataSize, lyr, curPolIdxOffs):
		u'''Poly&SurfaceIdx. {TagsIdx:[polIdx,polIdx,...]}
		PTAG { type[ID4], ( poly[VX], tag[U2] )* } 
		POL->TAGS(name)->Surface'''
		offs = 0
		while(offs < dataSize):
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			tagIdx = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
			offs += RSize.U2

			if(tagIdx not in lyr.surfTags):
				lyr.surfTags[tagIdx] = []	# 既設でなければ作成

			lyr.surfTags[tagIdx].append(polIdx + curPolIdxOffs)	# POLS内のIndex

		self.DebugPrint_PTag("SURF", lyr.name, lyr.surfTags)

	#### ------------------------------------------------------------------------
	def Read_PTag(self, data, dataSize, lyr, tags, curPolIdxOffs, tagKey, ptagName):
		u'''
		PTAG_PART. {name:[polIdx,,,]}
		PTAG_COLR. {colorName:[polIdx,,,]}
		PTAG_PICK. {name:[polIdx,,,]} (lxo)
		'''
		tagDct = lyr[tagKey]

		offs = 0
		while(offs < dataSize):
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			tagIdx = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
			offs += RSize.U2

			tag = tags[tagIdx]	# カラー/パート/選択名
			if(tag not in tagDct):
				tagDct[tag] = []

			tagDct[tag].append(polIdx + curPolIdxOffs)

		self.DebugPrint_PTag(ptagName, lyr.name, tagDct)

	#### ========================================================================
	#### EDGE
	#### ========================================================================
	def Read_SUBV_VMed(self, data, dataSize, lyr):
		u'''Modo EdgeWeight. {{id1 id2:val},{id1 id2:val},...}
		VMED { type[ID4], dimension[U2], name[S0],( vert[VX], vert[VX], value[F4] # dimension )* }
		SUBV, dim[U2], name[S0],( vert[VX], vert[VX], value[F4] # dimension )*'''
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])

		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(idx1, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(idx2, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			wgt = self.FnUnpckF4(data[offs:(offs + RSize.F4)])[0]
			offs += RSize.F4
			lyr.edgeWgtsM[(idx1, idx2)] = wgt

		lyr.hasEdgeM = True

		self.DebugPrint_EdgeWeight("VMed(Modo)", name, lyr.edgeWgtsM)

	#### ------------------------------------------------------------------------
	def Read_EPCK_VMed(self, data, dataSize, lyr):
		u'''Modo EdgeSelect. {{id1 id2:val},{id1 id2:val},...}
		VMED { type[ID4], dimension[U2], name[S0],( vert[VX], vert[VX], value[F4] # dimension )* }
		EPCK, dim[U2], name[S0],( vert[VX], vert[VX], value[F4] # dimension )*'''
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		offs = RSize.U2 + nameLen

		APD = lyr.edgeSel.append
		while(offs < dataSize):
			(idx1, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(idx2, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx

			APD( (idx1, idx2) )

		self.DebugPrint_EdgeSelect("VMed(Modo)", name, lyr.edgeSel)

	#### ========================================================================
	#### 簡易走査
	#### ========================================================================
	def PreScan(self, fileName, **kwd):
		u'''レイヤ簡易走査'''
		(fp, errMsg) = self.__Check_FileState(fileName)
		if(errMsg):
			print(errMsg),
			return(None)

		print(u"走査開始: {}".format(fileName))

		layers = self.ScanLayer(fp)
		fp.close()

		print(u"-------: {} 個のレイヤ".format(len(layers)))

		if(self.VisFlags.VERB):
			print(u"走査完了."),

		return(layers)

	#### ------------------------------------------------------------------------
	def Scan_Morph_VMap(self, data, morphs):
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		morphs[name] = 1

	#### --------------------------------
	def Scan_RGB_VMap(self, data, lyr):
		(name, nameLen)= self.Read_nameString(data[RSize.U2:])
		if(name not in lyr.rgb):
			lyr.rgb[name] = LWO_Object.VMapObject()
		lyr.rgb[name].vmap.update({1:1})

	#### --------------------------------
	def Scan_RGB_VMad(self, data, lyr):
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		if(name not in lyr.rgb):
			lyr.rgb[name] = LWO_Object.VMapObject()
		lyr.rgb[name].vmad.update({1:1})

	#### --------------------------------
	def Scan_RGBA_VMap(self, data, lyr):
		(name, nameLen)= self.Read_nameString(data[RSize.U2:])
		if(name not in lyr.rgba):
			lyr.rgba[name] = LWO_Object.VMapObject()
		lyr.rgba[name].vmap.update({1:1})

	#### --------------------------------
	def Scan_RGBA_VMad(self, data, lyr):
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		if(name not in lyr.rgba):
			lyr.rgba[name] = LWO_Object.VMapObject()
		lyr.rgba[name].vmad.update({1:1})

	#### --------------------------------
	def Scan_NORM(self, data, lyr):
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		lyr.normals[name] = 1

	#### --------------------------------
	def Scan_SUBV_VMed(self, data, lyr):
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		lyr.hasEdgeM = True

	#### --------------------------------
	def Scan_Weight_VMad(self, data, lyr):
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		if(name == self.sdsEdgeWgtName):
			lyr.hasEdge = True

	#### --------------------------------
	def Scan_Bones(self, data, lyr):
		lyr.bones["dummy"] = 1
		lyr.bonNum = 1

	#### --------------------------------
	def Scan_Crvs(self, data, lyr):
		lyr.crvs = ((1,))
		lyr.crvNum = 1

	#### ========================================================================
	def __Check_FileState(self, fileName):
		u'''ファイルチェック. 読み進んだfileObjectを返す.'''
		if(not fileName) or (not isinstance(fileName, str)):
			return(None, u"ファイルが未指定.\n")

		if(not os.path.exists(fileName)):	# filenameを含まないpathも通る
			return(None, u"ファイル/パスが無い.\n")

		if(not os.path.isfile(fileName)):	# pathがfilenameを含むか
			return(None, u"ファイルが無い.\n")

		fp = open(fileName, 'rb')
		try:
			# FORM:size:<LWO2|LWO3|LXOB>
			(tag, size, lwid) = struct.unpack(ReadT.FORM, fp.read(RSize.FORM))
		except:
			fp.close()
			return(None, u"ファイルがおかしい.\n")

		if(self.fileType == "LWO2"):
			idSet = { lwoChunkIdT.LWO2, lwoChunkIdT.LWO3 }
			header = "LWO2/LWO3"
		else:
			idSet = { lxoChunkIdT.LXOB, }
			header = "LXOB"

		if(lwid not in idSet):
			fp.close()
			return(None, u"{}以外は非サポート.\n".format(header))

		return(fp, "")


"""
	def Read_PTag_Pick(self, data, dataSize, lyr, tags, curPolIdxOffs):
		u'''ポリゴン選択(lxo). PTAG_PICK. {name:[polIdx,,,]}'''
		tagDct = lyr.tagSelects

		offs = 0
		while(offs < dataSize):
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			tagIdx = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
			offs += RSize.U2

			tag = tags[tagIdx]	# 選択名
			if(tag not in tagDct):
				tagDct[tag] = []

			tagDct[tag].append(polIdx + curPolIdxOffs)

		self.DebugPrint_PTag("PICK", lyr.name, tagDct)

	#### ------------------------------------------------------------------------
	def Read_PTag_Color(self, data, dataSize, lyr, tags, curPolIdxOffs):
		u'''PTAG_COLR. {"col":[polIdx,,,]}'''
		tagDct = lyr.tagColors

		offs = 0
		while(offs < dataSize):
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			tagIdx = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
			offs += RSize.U2

			tag = tags[tagIdx]	# 色名
			if(tag not in tagDct):
				tagDct[tag] = []

			ltagDct[tag].append(polIdx + curPolIdxOffs)

		self.DebugPrint_PTag("COLR", lyr.name, tagDct)

	#### ------------------------------------------------------------------------
	def Read_PTag_Part(self, data, dataSize, lyr, tags, curPolIdxOffs):
		u'''PTAG_PART. {name:[polIdx,,,]}'''
		tagDct = lyr.tagParts

		offs = 0
		while(offs < dataSize):
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			tagIdx = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
			offs += RSize.U2

			tag = tags[tagIdx]	# パート名
			if(tag not in tagDct):
				tagDct[tag] = []

			tagDct[tag].append(polIdx + curPolIdxOffs)

		self.DebugPrint_PTag("PART", lyr.name, tagDct)
"""
