﻿# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import print_function

import sys, os, re
from functools import partial

import shiboken2 as shiboken
from PySide2.QtGui import QColor, QBrush
from PySide2.QtWidgets import QListWidget

import maya.OpenMayaUI as omUI
import maya.cmds as mc

from .. import Parse_LWO, Parse_LXO, Vis_Flags
from . import Build_LWO, Build_LXO, Opts, ApplyWeightsByAttr

if(sys.modules["Load_LWO"].DO_RELOAD):
	if(sys.version_info[0] >= 3):
		import importlib
		Fn = importlib.reload
	else:
		Fn = reload
	loc = set(locals().keys())
	for f in ("Parse_LWO", "Build_LWO", "Parse_LXO", "Build_LXO", "Opts", "ApplyWeightsByAttr", "Vis_Flags"):
		if(f in loc):
			Fn(eval(f))
	del f, loc

#### =======================================================================
__all__ = ("LoadLWO")

# py2/3兼用対策
if(sys.version_info[0] >= 3):
	basestring = str

#### =======================================================================
# form文字定数 -----------
class FPos(object):
	TP = "top"
	BT = "bottom"
	LT = "left"
	RT = "right"
	__slots__ = ()

# popupMenu(button=) -----------
class PClick(object):
	RMB = 0
	LMB = 1
	MMB = 2
	__slots__ = ()

# LabelFont -----
class LblFont(object):
	PLN   = "plainLabelFont"		# プロポーショナル
	PLN_S = "smallPlainLabelFont"
	BLD   = "boldLabelFont"			# defaultFont
	BLD_S = "smallBoldLabelFont"
	TIN   = "tinyBoldLabelFont"
	FIX   = "fixedWidthFont"		# 固定幅
	FIX_S = "smallFixedWidthFont"
	OBL   = "obliqueLabelFont"		# 斜体
	OBL_S = "smallObliqueLabelFont"
	__slots__ = ()

# IconStyle (iconTextCB/RadBtn(st=),)-----
class ICStyle(object):
	ICON    = "iconOnly"
	TEXT    = "textOnly"
	HORIZON = "iconAndTextHorizontal"
	VERTICAL= "iconAndTextVertical"
	CENTER  = "iconAndTextCentered"
	__slots__ = ()

# 空object -----
class BaseObject(object):
	'''inst objを辞書風に使う為の物. クラス変数には非対応.'''
	def items(self):					# obj.items()
		return(self.__dict__.items())
	def values(self):					# obj.values()
		return(self.__dict__.values())
	def keys(self):						# obj.keys()
		return(self.__dict__.keys())
	def __setitem__(self, k,v):			# obj[k]=v
		self.__dict__[k] = v
	def __getitem__(self, k):			# =obj[k]
		return(self.__dict__[k])
	def __contains__(self, k):
		return(k in self.__dict__.keys())

# BGC object -----
class BGCols(object):
	DEF  = 0.27
	BASE = 0.32
	HCON = 0.42
	BG   = 0.1647

	DEF3 = (DEF,)*3
	BASE3= (BASE,)*3
	HCON3= (HCON,)*3
	BG3  = (BG,)*3

	MPOP_GREEN = (BASE, BASE+.15, BASE+.1)
	LPOP_BLUE  = (BASE, BASE+.1,  BASE+.15)

	# STAT_BLUE(.32, .62, .72)をbgcに入れるとhlcと同等になる(2015はbgcのみだった事への対策).
	# 同値をhlcに入れると明るくなる.
	STAT_BLUE = (BASE, BASE+.3, BASE+.4)
	HLC = (0.3215686274509804, 0.5215686274509804, 0.6509803921568628)
	__slots__ = ()

#### =======================================================================
class OptVarObject(object):
	__slots__=()
	#### ------------------------
	@classmethod
	def members(cls):
		res = []
		APD = res.append
		for m in dir(cls):
			if("__" not in m):
				try:
					r = cls.__dict__[m]
				except:
					continue
				if(not isinstance(r, classmethod)):
					APD(r)
		return(res)

	#### ------------------------
	@classmethod
	def Remove(cls, **kwd):
		u'''Remove(<window=windowName>)'''
		if("window" in kwd):
			name = kwd["window"]
			if(mc.windowPref(name, ex=True, q=True)):
				mc.windowPref(name, r=True)	# edit=不要
				print("{}\n".format(name)),

		res = [ "removed {}".format(o) for o in cls.members() if(cls.RemoveOptionVar(o)) ]
		print("\n".join(res) + "\n")

	#### ------------------------
	@staticmethod
	def RemoveOptionVar(opt):
		if(mc.optionVar(ex= opt)):
			mc.optionVar(rm= opt)
			return(True)
		return(False)

	#### ------------------------
	@staticmethod
	def GetOptionVar(opt, defVal):
		if(mc.optionVar(ex= opt)):
			return(mc.optionVar(q= opt))

		if(isinstance(defVal, int)):
			mc.optionVar(iv= (opt, defVal))
		elif(isinstance(defVal, float)):
			mc.optionVar(fv= (opt, defVal))
		else:
			mc.optionVar(sv= (opt, defVal))
		return(defVal)

#### =======================================================================
class UI_Opts(Opts.OptsObject):
	__slots__ = Opts.OptsObject.__slots__ + ("isLyrSel", "outXML", "outATTR")
	def __init__(self, *arg, **kwd):
		super(UI_Opts, self).__init__(*arg, **kwd)
		self.isLyrSel = True
		self.outXML   = False
		self.outATTR  = False

#### =======================================================================
class LoadLWO(object):
	__slots__ = (
		"ctls",			# UI ctl
		"subStore",		# 履歴リスト
		"presetPath",	# プリセットpathリスト
		"lyrVisList",	# 時点のレイヤ可視属性リスト
		"tmpSelList",	# レイヤ選択On/Off時のレイヤ選択状態退避用
		"dialogPath",	# 前回取得時のpath
		"wndName",
		"optsUI",
		"hasMentalray",
		"VisFlags",
		"currentExt",
	)

	TEMPLA_BASE = "LOAD_LWO_PANEL_TEMPLATE_BASE"	# UI template名

	WINDOWNAME  = "Load_LWO_panels"

	FILE_CAPTION = "Load LWO/LXO"
	FILE_FILTER  = u"Lightwave/Modo Files (*.lwo *.lxo);;Lightwave Files (*.lwo);;Modo Files (*.lxo);;全てのファイル (*.*)"

	LWO_EXT = ".lwo"
	LXO_EXT = ".lxo"

	#### ----------------
	class OptVar(OptVarObject):
		__slots__ = OptVarObject.__slots__
		HEAD = "LoadLightWaveObject_"

		SCALING   = HEAD+ "Scaling"
		LYR_SEL   = HEAD+ "LayerSelection"
		ING_HID   = HEAD+ "IgnoreHidden"
		MAKE_POLY = HEAD+ "Make_Polygon"
		MAKE_BONE = HEAD+ "Make_Bone"
		MAKE_CURV = HEAD+ "Make_Curve"
		MERGEUV   = HEAD+ "MergeUVs"
		SHARE_SG  = HEAD+ "ShareShadingGroup"
		USE_EDGEWGT  = HEAD+ "Use_EdgeWeight"
		USE_PIVOT    = HEAD+ "Use_Pivot"
		USE_VERTCOLR = HEAD+ "Use_VertexColor"
		USE_MR_VCOLR = HEAD+ "Use_miVertexColor"
		USE_DBL_SIDE = HEAD+ "Use_DoubleSide"
		USE_SELSET   = HEAD+ "Use_SelectionSet"
		USE_COLR_TAG = HEAD+ "Use_ColorTag"
		USE_PART_TAG = HEAD+ "Use_PartsTag"
		USE_PICK_TAG = HEAD+ "Use_PickTag"
		USE_WIRECOLR = HEAD+ "Use_WireColor"
		SHADER_TYPE  = HEAD+ "ShaderType"
		USE_IMAGE_COL= HEAD+ "Use_ColorImageTexture"
		USE_IMAGE_BMP= HEAD+ "Use_BumpImageTexture"
		LOAD_MORF_REL= HEAD+ "Load_MorphMap_Relative"
		LOAD_MORF_ABS= HEAD+ "Load_MorphMap_Absolute"
		MAKE_BLD_REL = HEAD+ "Make_BlendShape_Relative"
		MAKE_BLD_ABS = HEAD+ "Make_BlendShape_Absolute"
		BONE_TYPE    = HEAD+ "BoneType"
		OUT_WGT_XML  = HEAD+ "Output_BoneWeight_XML"
		OUT_WGT_ATTR = HEAD+ "Output_BoneWeight_ATTR"
		CURV_DEGREE  = HEAD+ "CurveDegree"

		del HEAD

	#### ----------------
	class CSize(object):
		RHGT = 20
		FHGT = 18
		SCL  = 50
		CRV  = 35
		SHD  = 60
		W    = 102	#4
		Wh   = W / 2
		BASE= (Wh, Wh, W*2, W)
		POLY= (W,)*5
		SETS= (W,)*5
		SHAD= (W, W*4)
		IMGS= (W, Wh, Wh)
		MORF= (W, Wh, Wh)
		BONE= ((1,W), (2,Wh), (3,W), (4,W), (5,1), (6,W),)
		CURV= (W, Wh-24, Wh+24)
#		BASE= (5, Wh, Wh, W*2, W)
#		SHAD= (5, W, W*4)
#		IMGS= (5, W, Wh, Wh)
#		MORF= (5, W, Wh, Wh)
#		BONE= ((1,5), (2,W), (3,Wh-8), (4,W), (5,W), (6,1), (7,W),)
#		CURV= (5, W, Wh, Wh)

		SHADER_CW4 = (70, SHD, SHD, 70)
		__slots__=()

	#### ----------------
	class Anno(object):
		SCL = u"倍率. LW: m, Maya: cmの場合100倍."
		LYR_SEL = u"読み込むレイヤーを個別に選択する."
		IGN_HIDE=(
			u"ファイル単位で読み込む場合に不可視レイヤを除外する.\n"
			u"全てが不可視なら何も読み込まない."
			)
		EXEC = (
			u"実行\n"
			u"MMB:\n"
			u"  レイヤリストの更新: 現在のファイルを走査する.\n"
			u"  設定初期化: 設定を初期値に変更する."
			)
		MESH = u"ポリゴンを読み込む."
		MERG = (
			u"　不連続UVのみの頂点(UV島外縁等)が不連続なままにならないように\n"
			u"mesh作成後に距離0.0001でUVをマージする."
			)
		SHSG = (
			u"　サーフェスに相当するshadingEngine/shadingGroupが存在する場合\n"
			u"新設せずに利用する."
			)
		UEDG = (
			u"　エッジウェイトがあればmeshに適用する. creaseSetの作成では無い.\n"
			u"SmoothMeshPrev/SmoothNode(MayaCC/OpenSubDiv):\n"
			u"  分割値が上限. div=2ならweightは2が最大効果.\n"
			u"Arnold(catclark):\n"
			u"  分割値+1が上限. div=2ならweightは3が最大効果."
			)
		UPIV = (
			u"レイヤーのpivotを適用する.\n"
			u"world上のpivot座標がトランスフォームの位置になる."
			)
		VCOL = (
			u"頂点カラーをmeshに適用する.\n"
			u"createColorSet/polyColorPerVertexの作成では無い."
			)
		DBLS = u"ポリゴンの両面フラグをONにする."
		MIVC = u"Mentalrayがある場合miVertexColorNodeを適用する."

		UTXT = u"objectSetへの登録."
		PICK = u"選択セットから作成する."
		COLR = u"カラータグから作成する."
		PART = u"パートタグから作成する."
		PPIC = u"選択タグから作成する."

		UWCL = u"ワイヤカラーをobjectColorRGBとしてmeshに適用する."

		STXT = u"使用するshaderNode. 互換は無いのでダミーにしかならない."
		UBLN = STXT
		TTXT = u"サーフェスに貼られているテクスチャをshaderに適用する."
		UCOL = u"ディフーズカラーを適用する."
		UBMP = u"バンプマップを適用する."
		MTXT = (
			u"モーフマップから形状を作成する.\n"
			u"作成したmeshへのshadingGroup適用はしない."
			)
		LREL = u"相対モーフを対象にする."
		LABS = u"絶対モーフを対象にする."
		BTXT = u"モーフマップから作成した形状を使ってblendShapeを適用する."
		UREL = LREL
		UABS = LABS
		BONE = u"スケルゴンをボーンとして読み込む."
		ZAXZ = (
			u"スケルゴンをZ-Axisボーンとして扱う.\n"
			u"　OFF: jointでは1階層ずれるので注意."
			)
		WXML = (
			u"　ウェイトマップを[Mayaメニュー]→[デフォーム]→[ウェイトを読み込み]で\n"
			u"使用出来るXMLファイルとして, プロジェクトパスの/scenesに出力する."
			)
		UWGT = u"ウェイトマップを独自のアトリビュートとしてmeshに埋め込む."
		UWGT_MENU = (
			u"アトリビュートを用いた場合の補助機能.\n"
			u"LMB:\n"
			u"  Attrから転送: skinClusetに転送する. 頂点増減があった場合は不正になる.\n"
			u"  Attrを除去: アトリビュートを除去する."
			)
		CURV = (
			u"CatmulRom/モデラカーブをnurbsCurveとして読み込む.\n"
			u"CatmulRomからNURBSへの数学的な変換はしない."
			)
		DEG  = u"nurbsCurveの次数"

		LYRS = (
			u"Modeler内のGUI順.\n"
			u"ファイル内の格納順. ( )は可視. [ ]は非表示.\n"
			u"レイヤ名.\n"
			u"VC: 頂点カラーの有無.\n"
			u"EW: エッジウェイトの有無.\n"
			u"Rel: 相対モーフの有無.\n"
			u"Abs: 絶対モーフの有無.\n"
			u"Bon: ボーンの有無.\n"
			u"Crv: カーブの有無."
			)

		__slots__=()

	#### ----------------
	class Lbls(object):
		PATH = u"File:"
		REF  = u"参照"
#		REF  = u"　参照　"
		PHT  = u"LWOファイル"
		SCL  = u"形状倍率"
		LYR_SEL  = u"[レイヤを選択]"
		IGN_HIDE = u"不可視を除外"
		DOIT    = u"※ 読込み実行 ※"
		REFRESH = u"レイヤリストの更新"
		INIT = u"設定初期化"

		MESH = u"[[ポリゴンを作成]]"
		MERG = u"不連続UVマージ"
		SHSG = u"重複SGを共用"
		UEDG = u"ｴｯｼﾞｳｪｲﾄを適用"
		UPIV = u"Pivotを適用"
		VCOL = u"頂点色を適用"
		MIVC = u"miVColorを適用"
		DBLS = u"両面をON"

		PICK = u"頂点選択セット"
		COLR = u"Colrタグ"
		PART = u"Partタグ"
		PPIC = u"Pickタグ"

		UWCL = u"Wire色を適用"

		BONE = u"[[ボ－ンを作成]]"
		ZAXZ = u"[ZAxis]"
		WXML = u"[ｳｪｲﾄのXML出力]"
		UWGT = u"[ｳｪｲﾄをAttr付与]"
		UWGT_MENU = u"▽ウェイト操作"
		UWGT_MENU_APPLY = u"Attrから転送"
		UWGT_MENU_REMOVE= u"Attrを除去"
		CURV = u"[[カーブを作成]]"

		ABS  = u"絶対"
		REL  = u"相対"
		UTXT = u"Setsへの登録:"
		BTXT = u"ﾌﾞﾚﾝﾄﾞｼｪｲﾌﾟを設定:"
		MTXT = u"モーフ形状を作成 :"
		TTXT = u"画像を適用:"
		STXT = u"使用シェーダ:"
		DTXT = u"次数"

		T_COLR = u"Color"
		T_BUMP = u"Bump"

		S_LAMB = u"Lambert"
		S_BLIN = u"Blinn"
		S_PHNG = u"Phong"
		S_STD  = u"StdSurf"
		SHADERS= (S_LAMB, S_BLIN, S_PHNG, S_STD)

		M_INV = u"選択反転"
		M_ALL_SEL = u"全て選択"
		M_ALL_REM = u"全て解除"
		M_SEL_VIS = u"選択: 可視"
		M_SEL_HID = u"選択: 不可視"
		M_REM_VIS = u"除外: 可視"
		M_REM_HID = u"除外: 不可視"
		M_PRESET = u"プリセット"
		M_RECENT = u"履歴"
		__slots__=()

	#### ----------------
	class Msgs(object):
		NOSEL_FILE = u"ファイル選択が無い."
		NOSEL_LAYER= u"レイヤー選択が無い."
		NOSEL_MAKE = u"読み込み対象が未指定."
		__slots__=()

	#### ----------------
	qTxColorBrush = QBrush(QColor("#9FAFBF"))
	qBGColorBrush = QBrush(QColor("#3F3F3F"))

	#### =======================================================================
	def __init__(self, *arg, **kwd):
		self.wndName = self.WINDOWNAME
		self.ctls = BaseObject()
		self.subStore   = []
		self.presetPath = []
		self.tmpSelList = []
		self.lyrVisList = []
		self.dialogPath = ""

		self.currentExt = ""

		self.hasMentalray = (mc.pluginInfo("Mayatomr", l=True, q=True))

		self.VisFlags = Vis_Flags.VisFlags(**kwd)

		self.optsUI = UI_Opts()

		OptVar = self.OptVar
		self.optsUI.scl      = OptVar.GetOptionVar(OptVar.SCALING, 100.0)
		self.optsUI.deg      = OptVar.GetOptionVar(OptVar.CURV_DEGREE, 3)
		self.optsUI.isLyrSel = OptVar.GetOptionVar(OptVar.LYR_SEL, False)
		self.optsUI.skipHide = OptVar.GetOptionVar(OptVar.ING_HID, True)
		self.optsUI.isMesh   = OptVar.GetOptionVar(OptVar.MAKE_POLY, True)
		self.optsUI.isBone   = OptVar.GetOptionVar(OptVar.MAKE_BONE, False)
		self.optsUI.isCrv    = OptVar.GetOptionVar(OptVar.MAKE_CURV, False)
		self.optsUI.mergeUV  = OptVar.GetOptionVar(OptVar.MERGEUV, False)#True)
		self.optsUI.shareSG  = OptVar.GetOptionVar(OptVar.SHARE_SG, True)
		self.optsUI.useEdge  = OptVar.GetOptionVar(OptVar.USE_EDGEWGT, True)
		self.optsUI.usePivot = OptVar.GetOptionVar(OptVar.USE_PIVOT, False)
		self.optsUI.useVCol  = OptVar.GetOptionVar(OptVar.USE_VERTCOLR, False)
		self.optsUI.useMiVC  =(OptVar.GetOptionVar(OptVar.USE_MR_VCOLR, False) & self.hasMentalray)
		self.optsUI.isDblSide= OptVar.GetOptionVar(OptVar.USE_DBL_SIDE, False)
		self.optsUI.usePick  = OptVar.GetOptionVar(OptVar.USE_SELSET, False)
		self.optsUI.useColr  = OptVar.GetOptionVar(OptVar.USE_COLR_TAG, False)
		self.optsUI.usePart  = OptVar.GetOptionVar(OptVar.USE_PART_TAG, False)
		self.optsUI.usePPic  = OptVar.GetOptionVar(OptVar.USE_PICK_TAG, False)
		self.optsUI.useWire  = OptVar.GetOptionVar(OptVar.USE_WIRECOLR, False)
		self.optsUI.mtlType  = OptVar.GetOptionVar(OptVar.SHADER_TYPE, 0)
		self.optsUI.useColT  = OptVar.GetOptionVar(OptVar.USE_IMAGE_COL, False)
		self.optsUI.useBmpT  = OptVar.GetOptionVar(OptVar.USE_IMAGE_BMP, False)
		self.optsUI.loadRM   = OptVar.GetOptionVar(OptVar.LOAD_MORF_REL, False)
		self.optsUI.loadAM   = OptVar.GetOptionVar(OptVar.LOAD_MORF_ABS, False)
		self.optsUI.useRM    = OptVar.GetOptionVar(OptVar.MAKE_BLD_REL, False)
		self.optsUI.useAM    = OptVar.GetOptionVar(OptVar.MAKE_BLD_ABS, False)
		self.optsUI.isZAxis  = OptVar.GetOptionVar(OptVar.BONE_TYPE, True)
		self.optsUI.useXML   = OptVar.GetOptionVar(OptVar.OUT_WGT_XML, False)
		self.optsUI.useWgt   = OptVar.GetOptionVar(OptVar.OUT_WGT_ATTR, False)

	#### =======================================================================
	@classmethod
	def RemOptVar(cls):
		cls.OptVar.Remove(window = cls.WINDOWNAME)

	#### =======================================================================
	def GetListWidget(self, ctl):
		return(shiboken.wrapInstance(long(omUI.MQtUtil.findControl(ctl)), QListWidget))

	#### =======================================================================
	def SubWind_Reuse(self, wndName):
		if(mc.window(wndName, ex=True)):
			mc.showWindow(wndName)
			return(wndName, True)
		else:
			return(mc.window(wndName, t=wndName, s=True, rtf=True), False)

	#### ------------------------
	def SubWind_Rebuild(self, wndName):
		if(mc.window(wndName, ex=True)): mc.deleteUI(wndName)
		return(mc.window(wndName, t=wndName, s=True, rtf=True))

	#### =======================================================================
	def MenuSeparator(self, m, *arg, **kwd):
		mc.menuItem(p=m, d=True, **kwd)

	def MenuDelete(self, parent, *arg, **kwd):
		mc.popupMenu(parent, dai=True, e=True)	# 項目削除

	#### =======================================================================
	def SetupPopupMenu_LyrList(self, parent):
		menuCtl = mc.popupMenu(p= parent, b= PClick.MMB)
		mc.menuItem(l= self.Lbls.M_INV, c= partial(self.SelList_Mod_Rev, parent))
		self.MenuSeparator(menuCtl)
		mc.menuItem(l= self.Lbls.M_ALL_SEL, c= partial(self.SelList_Mod_AllSel, parent))
		mc.menuItem(l= self.Lbls.M_ALL_REM, c= partial(self.SelList_Mod_DeSel, parent))
		self.MenuSeparator(menuCtl)
		mc.menuItem(l= self.Lbls.M_SEL_VIS, c= partial(self.SelList_Mod_Vis,  parent, "sii"))
		mc.menuItem(l= self.Lbls.M_SEL_HID, c= partial(self.SelList_Mod_Hide, parent, "sii"))
		self.MenuSeparator(menuCtl)
		mc.menuItem(l= self.Lbls.M_REM_VIS, c= partial(self.SelList_Mod_Vis, parent, "dii"))
		mc.menuItem(l= self.Lbls.M_REM_HID, c= partial(self.SelList_Mod_Hide,parent, "dii"))
		mc.setParent(m=True)	# 階層を閉じる

	#### ------------------------
	def SelList_Mod_Rev(self, ctl, *arg):
		u'''選択の反転'''
		n = mc.textScrollList(ctl, q=True, ni=True)			# 全項目数
		allLst = list(range(1, n+1))						# lst[0]=1,[1]=2...
		curSels = mc.textScrollList(ctl, q=True, sii=True)	# idxは 1 から

		if(not curSels):			# 選択無し
			mc.textScrollList(ctl, e=True, sii= allLst)
		else:
			mc.textScrollList(ctl, e=True, da=True)	# 全選択解除

			# 選択が指すidx範囲を[1:n]から[0:n-1]とする
			# 選択されてる[idx]の状態をフラグ 0
			for i in curSels:
				allLst[(i -1)] = 0					# フラグとして 0.

			reducedLst = list(set(allLst[:]))		# フラグ 0で重複削除,ソート.
			if(reducedLst):
				# 先頭にフラグ 0が残っていれば[1:].
				mc.textScrollList(ctl, e=True, sii= reducedLst[int(reducedLst[0] == 0):])

	#### ------------------------
	def SelList_Mod_AllSel(self, ctl, *arg):
		u'''全て選択'''
		n = mc.textScrollList(ctl, q=True, ni=True)					# 全項目数
		mc.textScrollList(ctl, e=True, sii= list(range(1, n+1)))	# idxは 1 から

	#### ------------------------
	def SelList_Mod_DeSel(self, ctl, *arg):
		u'''選択の全解除'''
		mc.textScrollList(ctl, e=True, da=True)

	#### ------------------------
	def SelList_Mod_Vis(self, ctl, optStr, *arg):
		u'''可視レイヤを選択'''
		self.__SelList_Mod_State(ctl, optStr, False)

	#### ------------------------
	def SelList_Mod_Hide(self, ctl, optStr, *arg):
		u'''不可視レイヤを選択'''
		self.__SelList_Mod_State(ctl, optStr, True)

	#### ------------------------
	def __SelList_Mod_State(self, ctl, optStr, state):
		# 論理順で前詰めしたリストから可視属性を見る. 選択操作のidxは 1 から(start=1).
		# (lyr内論理idx, 属性)
		lst = [ cnt for (cnt, (_, isHide)) in enumerate(self.lyrVisList, start=1) if(isHide == state) ]
		if(lst):
			mc.textScrollList(ctl, e=True, **{ optStr: lst })

	#### =======================================================================
	def SetupPopupMenu_RecentName(self, parent, ctl):
		menuCtl = mc.popupMenu(p= parent, b= PClick.MMB)
		mc.popupMenu(menuCtl, pmc= partial(self.MakePopupMenu_RecentName, menuCtl, ctl), e=True)

	#### ------------------------
	def MakePopupMenu_RecentName(self, parent, ctl, *arg):
		self.MenuDelete(parent)
		self.MakeSubMenu_Recent(parent, ctl)
		self.MakeSubMenu_PresetPath(parent, ctl)

	#### ------------------------
	def MakeSubMenu_Recent(self, parent, ctl):
		selID = mc.menuItem(p= parent, l= self.Lbls.M_RECENT, sm=True)
		if(len(self.subStore) == 0):
			mc.menuItem(selID, en=False, e=True)	#
		else:
			for n in self.subStore:
				mc.menuItem(p= selID, l= n, c= partial(self.Func_Set_FileName_ByPresetMenu, ctl, n))
		mc.setParent(m=True)	# 階層閉じ

	#### ------------------------
	def MakeSubMenu_PresetPath(self, parent, ctl):
		selID = mc.menuItem(p= parent, l= self.Lbls.M_PRESET, sm=True)
		for n in self.presetPath:
			mc.menuItem(p= selID, l= n, c= partial(self.Func_Set_FileName_ByPresetMenu, ctl, n))
		mc.setParent(m=True)	# 階層閉じ

	#### ------------------------
	def RefineRecentNameArray(self, s):
		try:
			idx = self.subStore.index(s)
			del self.subStore[idx]
		except:
			pass
		self.subStore.insert(0, s)

	#### =======================================================================
	def Get_FileName_FromDialog(self, *arg):
		u'''dialog経由のパス設定時の処理'''
		dialogPref = mc.optionVar(q= "FileDialogStyle")

		fn = mc.fileDialog2(cap= self.FILE_CAPTION, dir= self.dialogPath, ff= self.FILE_FILTER, ds= dialogPref, fm=1)
		if(not fn):
			return	# キャンセルはNone, 戻りは[u"drv:/path/file", ...], fm=4時はds=1/2でセパレータが変わる

		self.dialogPath = mc.workspace(q=True, dir=True)	# 前回作業Dir

		mc.textFieldButtonGrp(self.ctls.FNAME, e=True, tx= fn[0])

		self.RefineRecentNameArray(fn[0])	# ファイル履歴

		self.Set_LayerList(fn[0])			# lyr情報

	#### ------------------------
	@staticmethod
	def Determine_DialogPath(fn, *arg):
		u'''path or path/fileの有無チェック. 無効の時はワークスペースRoot. 末尾 / 付きで統一.'''
		if(fn):
			if(os.path.exists(fn)):				# file/dirとして存在する
				# fileならdir抽出, dirならそのまま
				pt = os.path.dirname(fn) if(os.path.isfile(fn)) else fn
				if(not pt.endswith(("/","\\"))):
					pt = "{}/".format(pt)
				return(pt.replace("\\", "/"))
		return(mc.workspace(q=True, rd=True))	# 無効なpathならワークスペースRoot

	#### ------------------------
	def Func_Set_FileName_ByPresetMenu(self, ctl, filePathName, *arg):
		mc.textFieldButtonGrp(ctl, e=True, tx= filePathName)

		self.dialogPath = self.Determine_DialogPath(filePathName)	# 現在dirを変更

		self.Set_LayerList(filePathName)

	#### --------------------------------
	def Func_Set_FileName_FromPanel(self, filePathName, *arg):
		u'''パスの直接入力時の処理. txField::ccとmenuから呼ばれる'''
		if(isinstance(filePathName, bool)):		# menu時はBoolを受けている
			filePathName = mc.textFieldButtonGrp(self.ctls.FNAME, q=True, tx=True)

		# / 終端でなく, 拡張子が無い場合はlwo→lxo順で探る
		if(not filePathName.endswith(("/", "\\"))):
			(base, ext) = os.path.splitext(filePathName)
			if(not ext):
				if(os.path.exists("{}{}".format(filePathName, self.LWO_EXT))):
					filePathName += self.LWO_EXT
				elif(os.path.exists("{}{}".format(filePathName, self.LXO_EXT))):
					filePathName += self.LXO_EXT
				mc.textFieldButtonGrp(self.ctls.FNAME, e=True, tx= filePathName)

		self.dialogPath = self.Determine_DialogPath(filePathName)	# 現在dirを変更

		self.Set_LayerList(filePathName)

	#### ------------------------
	def __Get_CurrentExt(self, filename):
		if((not filename) or (not os.path.exists(filename))):
			self.currentExt = ""
			return(True)

		(base, ext) = os.path.splitext(filename)	# "drv:/path/fn", ".ext"
		if(ext.lower() == self.LWO_EXT):
			self.currentExt = self.LWO_EXT
		elif(ext.lower() == self.LXO_EXT):
			self.currentExt = self.LXO_EXT
		else:
			self.currentExt = ""

		self.SyncStat_LXO()
		return(False)

	#### ------------------------
	def Get_InitialPath_FromArg(self, fn, *arg):
		u'''引数のfilePathの検証, 文字列としての評価, 文字列以外なら空文字, 配列なら[0]'''
		res = ""	# None,空文字,空配列
		if(fn):
			if(isinstance(fn, (list, tuple))):
				if(isinstance(fn[0], basestring)):
					res = fn[0]
			elif(isinstance(fn, basestring)):
				res = fn
			else:
				pass
		self.dialogPath = self.Determine_DialogPath(res)	# 現在dirを変更
		return(res)

	#### ------------------------
	@staticmethod
	def Get_PresetName_FromArg(*arg, **kwd):
		u'''引数のプリセットパスの検証'''
		if(kwd):
			p = kwd.get("p", kwd.get("preset", None))
			if(not p):
				return([])

			if(isinstance(p, basestring)):
				return([p])	# 単一文字列なら配列にして返す

			if(isinstance(p, (list, tuple))):	# 配列なら文字列だけをピックアップ
				return( [ s for s in p if(isinstance(s, basestring)) ] )
#				return(filter(lambda x: isinstance(x, basestring), p))
		return([])

	#### =======================================================================
	def Create_Ctl_FileLine(self, parent, initialPath, Lbls, *arg):
		ctl = mc.textFieldButtonGrp(p= parent, l= Lbls.PATH, tx= initialPath, bl= Lbls.REF,
			pht= Lbls.PHT, cw=((1,25),(2,185),(3,50)), adj=2, cal=(1,FPos.LT),
			bc= partial(self.Get_FileName_FromDialog), cc= partial(self.Func_Set_FileName_FromPanel))

		(lbl, field, btn) = self.Get_GrpLayoutChildCtls(ctl)
		mc.control(field, h=self.CSize.RHGT, e=True)
		mc.button(btn, w=50, h=self.CSize.FHGT, e=True)

		self.SetupPopupMenu_RecentName(ctl, ctl)
		mc.setParent(parent)
		return(ctl)

	#### =======================================================================
	def Create_Ctl_BaseLine(self, parent, ctls, CSize, Lbls, Anno, *arg):
#		r1 = mc.rowLayout(p= parent, nc=5, cw5= CSize.BASE, adj=5)
#		mc.text()
		r1 = mc.rowLayout(p= parent, nc=4, cw4= CSize.BASE, adj=4)
		mc.text(l= Lbls.SCL, ann= Anno.SCL)
		ctls.SCL = mc.floatField(v= self.optsUI.scl, ann= Anno.SCL)

		subForm = mc.formLayout(p= r1)
		ctls.LYR_SEL  = mc.iconTextCheckBox(l= Lbls.LYR_SEL, ebg=True, w= CSize.W, v= self.optsUI.isLyrSel, ann= Anno.LYR_SEL)
		ctls.IGN_HIDE = mc.checkBox(l= Lbls.IGN_HIDE, v= self.optsUI.skipHide, ann= Anno.IGN_HIDE)
		mc.iconTextCheckBox(ctls.LYR_SEL, cc= partial(self.SyncStat_IgnoreHideLayer, ctls.LYR_SEL, ctls.IGN_HIDE), e=True)

		ctl = mc.button(p= r1, l= Lbls.DOIT, bgc= BGCols.MPOP_GREEN, c= partial(self.Exec), ann= Anno.EXEC)
		menuCtl = mc.popupMenu(parent= ctl, b= PClick.MMB)
		mc.menuItem(p= menuCtl, l= Lbls.REFRESH, c= partial(self.Func_Set_FileName_FromPanel))
		mc.menuItem(p= menuCtl, l= Lbls.INIT, c= partial(self.Init_UI_Options))

		aco = ((ctls.IGN_HIDE, FPos.LT, 5, ctls.LYR_SEL),)	# ctl1の辺をctl2の辺に合わせる
		mc.formLayout(subForm, ac= aco, e=True)

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_Polygons(self, parent, ctls, CSize, Lbls, Anno, *arg):
		mc.rowLayout(p= parent, nc=5, cw5= CSize.POLY)
		ctls.MESH = mc.iconTextCheckBox(l= Lbls.MESH, w= CSize.W, v= self.optsUI.isMesh, ann= Anno.MESH)
		mc.iconTextCheckBox(ctls.MESH, cc= partial(self.SyncStat_Mesh, ctls.MESH), e=True)
#		ctls.MERG = mc.checkBox(l= Lbls.MERG, v= self.optsUI.mergeUV, ann= Anno.MERG)
		ctls.SHSG = mc.checkBox(l= Lbls.SHSG, v= self.optsUI.shareSG, ann= Anno.SHSG)
		ctls.UEDG = mc.checkBox(l= Lbls.UEDG, v= self.optsUI.useEdge, ann= Anno.UEDG)
		ctls.UPIV = mc.checkBox(l= Lbls.UPIV, v= self.optsUI.usePivot, ann= Anno.UPIV)
		ctls.VCOL = mc.checkBox(l= Lbls.VCOL, v= self.optsUI.useVCol, cc= partial(self.SyncStat_VCol), ann= Anno.VCOL)

		#### --------------------------------
		ctls.R_POLY = mc.rowLayout(p= parent, nc=5, cw5= CSize.POLY)
		mc.text()
		ctls.DBLS = mc.checkBox(l= Lbls.DBLS, v= self.optsUI.isDblSide, ann= Anno.DBLS)
		ctls.UWCL = mc.checkBox(l= Lbls.UWCL, v= self.optsUI.useWire, ann= Anno.UWCL)
		mc.text()
		mc.text()
#		ctls.MIVC = mc.checkBox(l= Lbls.MIVC, v=(self.optsUI.useMiVC & self.hasMentalray), ann= Anno.MIVC)
		ctls.MIVC = ""

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_Sets(self, parent, ctls, CSize, Lbls, Anno, *arg):
		ctls.R_SETS = mc.rowLayout(p= parent, nc=5, cw5= CSize.SETS)
		mc.text(l= Lbls.UTXT, ann= Anno.UTXT)
		ctls.PICK = mc.checkBox(l= Lbls.PICK, v= self.optsUI.usePick, ann= Anno.PICK)
		ctls.COLR = mc.checkBox(l= Lbls.COLR, v= self.optsUI.useColr, ann= Anno.COLR)
		ctls.PART = mc.checkBox(l= Lbls.PART, v= self.optsUI.usePart, ann= Anno.PART)
		ctls.PPIC = mc.checkBox(l= Lbls.PPIC, v= self.optsUI.usePPic, ann= Anno.PPIC)

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_Shader(self, parent, ctls, CSize, Lbls, Anno, *arg):
#		ctls.R_SHAD = mc.rowLayout(p= parent, nc=3, cw3= CSize.SHAD)
#		mc.text()
		ctls.R_SHAD = mc.rowLayout(p= parent, nc=2, cw2= CSize.SHAD)
		mc.text(l= Lbls.STXT, ann= Anno.STXT)
		ctls.UBLN = mc.radioButtonGrp(sl= self.optsUI.mtlType +1, nrb=4, ann= Anno.UBLN,
			la4= Lbls.SHADERS, cl4= (FPos.LT,)*4, cw4= CSize.SHADER_CW4)
		# idxが1からなので+1

		#### --------------------------------
#		ctls.R_IMGS = mc.rowLayout(p= parent, nc=4, cw4= CSize.IMGS)
#		mc.text()
		ctls.R_IMGS = mc.rowLayout(p= parent, nc=3, cw3= CSize.IMGS)
		mc.text(l= Lbls.TTXT, ann= Anno.TTXT)
		ctls.UCOL = mc.checkBox(l= Lbls.T_COLR, v= self.optsUI.useColT, ann= Anno.UCOL)
		ctls.UBMP = mc.checkBox(l= Lbls.T_BUMP, v= self.optsUI.useBmpT, ann= Anno.UBMP)

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_Morphs(self, parent, ctls, CSize, Lbls, Anno, *arg):
#		ctls.R_MORF = mc.rowLayout(p= parent, nc=4, cw4= CSize.MORF)
#		mc.text()
		ctls.R_MORF = mc.rowLayout(p= parent, nc=3, cw3= CSize.MORF)
		mc.text(l= Lbls.MTXT, ann= Anno.MTXT)
		ctls.LREL = mc.checkBox(l= Lbls.REL, v= self.optsUI.loadRM, ann= Anno.LREL)
		ctls.LABS = mc.checkBox(l= Lbls.ABS, v= self.optsUI.loadAM, ann= Anno.LABS)

		#### --------------------------------
#		ctls.R_BLDS = mc.rowLayout(p= parent, nc=4, cw4= CSize.MORF)
#		mc.text()
		ctls.R_BLDS = mc.rowLayout(p= parent, nc=3, cw3= CSize.MORF)
		mc.text(l= Lbls.BTXT, ann= Anno.BTXT)
		ctls.UREL = mc.checkBox(l= Lbls.REL, v= self.optsUI.useRM, ann= Anno.UREL)
		ctls.UABS = mc.checkBox(l= Lbls.ABS, v= self.optsUI.useAM, ann= Anno.UABS)

		mc.checkBox(ctls.LREL, e=True, cc= partial(self.SyncStat_Morph, ctls.LREL, ctls.UREL, ctls.LABS))
		mc.checkBox(ctls.LABS, e=True, cc= partial(self.SyncStat_Morph, ctls.LABS, ctls.UABS, ctls.LREL))

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_Bones(self, parent, ctls, CSize, Lbls, Anno, *arg):
#		mc.rowLayout(p= parent, nc=7, cw= CSize.BONE, cal=(6,FPos.RT), adj=6)
#		mc.text()
		mc.rowLayout(p= parent, nc=6, cw= CSize.BONE, cal=(5,FPos.RT), adj=5)
		ctls.BONE = mc.iconTextCheckBox(l= Lbls.BONE, w= CSize.W, v= self.optsUI.isBone, ann= Anno.BONE)
		mc.iconTextCheckBox(ctls.BONE, cc= partial(self.SyncStat_Bone, ctls.BONE), e=True)

		ctls.ZAXZ = mc.iconTextCheckBox(l= Lbls.ZAXZ, w= CSize.Wh-8, v= self.optsUI.isZAxis, ann= Anno.ZAXZ)
		mc.iconTextCheckBox(ctls.ZAXZ, cc= partial(self.SyncStat_Wght, ctls.ZAXZ), e=True)

		ctls.WXML = mc.iconTextCheckBox(l= Lbls.WXML, v= self.optsUI.useXML, ann= Anno.WXML)
		mc.iconTextCheckBox(ctls.WXML, cc= partial(self.SyncStat_Wght, ctls.WXML), e=True)

		ctls.UWGT = mc.iconTextCheckBox(l= Lbls.UWGT, v= self.optsUI.useWgt, ann= Anno.UWGT)
		mc.iconTextCheckBox(ctls.UWGT, cc= partial(self.SyncStat_Wght, ctls.UWGT), e=True)

		mc.text()

		ctl = mc.button(l= Lbls.UWGT_MENU, ann= Anno.UWGT_MENU)
		menuCtl = mc.popupMenu(p= ctl, b= PClick.LMB)
		mc.menuItem(p= menuCtl, l= Lbls.UWGT_MENU_APPLY,  c= partial(ApplyWeightsByAttr.Apply))
		mc.menuItem(p= menuCtl, l= Lbls.UWGT_MENU_REMOVE, c= partial(ApplyWeightsByAttr.Remove))

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_Curves(self, parent, ctls, CSize, Lbls, Anno, *arg):
#		mc.rowLayout(p= parent, nc=4, cw4= CSize.CURV)
#		mc.text()
		mc.rowLayout(p= parent, nc=3, cw3= CSize.CURV)
		ctls.CURV = mc.iconTextCheckBox(l= Lbls.CURV, w= CSize.W, v= self.optsUI.isCrv, ann= Anno.CURV)
		mc.iconTextCheckBox(ctls.CURV, cc= partial(self.SyncStat_Crv, ctls.CURV), e=True)
		ctls.DTXT = mc.text(l= Lbls.DTXT, ann= Anno.DEG)
		ctls.DEG  = mc.intField(min=1, max=7, v= self.optsUI.deg, ann= Anno.DEG)

		mc.setParent(parent)

	#### =======================================================================
	def Create_Ctl_LayerList1(self, parent, baseClm, ctls, *arg):
		u'''レイヤーリスト作成1'''
#		subClm = mc.colmnLayout(p= rootForm, adj=True, bgc=(0,1,0))
		subClm = mc.rowLayout(p= parent, nc=1, rat=(1,"both",0), adj=1)#, bgc=(0,1,0))
		afo = (
			(baseClm, FPos.LT, 0), (baseClm, FPos.RT, 0), (baseClm, FPos.TP, 0),
			(subClm,  FPos.LT, 0), (subClm,  FPos.RT, 0),#  (subClm, FPos.BT, 0),
		)
		aco = ((subClm, FPos.TP, 0, baseClm),)
		apo = ((subClm, FPos.BT, 0, 100),)
		mc.formLayout(parent, e=True, af= afo, ac= aco, ap=apo)

		lstForm = mc.formLayout(p= subClm)

		lstCtl = mc.textScrollList(p= lstForm, fn= LblFont.FIX, ams=True, ann= self.Anno.LYRS)

#		afo = ((lstCtl, FPos.TP, 0), (lstCtl, FPos.LT, 0), (lstCtl, FPos.RT, 0),)
#		afo = ((lstCtl, FPos.TP, 0), (lstCtl, FPos.LT, 0), (lstCtl, FPos.RT, 0), (lstCtl, FPos.BT, 0),)
		apo = ((lstCtl, FPos.BT, 0, 100),)
		afo =()
		mc.formLayout(lstForm, e=True, af= afo, ap= apo)

		return(lstCtl)

	#### =======================================================================
	def Create_Ctl_LayerList2(self, parent, baseClm, ctls, *arg):
		u'''レイヤーリスト作成2'''
		lstForm = mc.formLayout(p= parent)
		afo = (
			(baseClm, FPos.LT, 0), (baseClm, FPos.RT, 0), (baseClm, FPos.TP, 0),
			(lstForm, FPos.LT, 0), (lstForm, FPos.RT, 0), (lstForm, FPos.BT, 0),
		)
		aco = ((lstForm, FPos.TP, 0, baseClm),)
		mc.formLayout(parent, e=True, af= afo, ac= aco)

		lstCtl = mc.textScrollList(p= lstForm, fn= LblFont.FIX, ams=True, ann= self.Anno.LYRS)

		afo = ((lstCtl, FPos.TP, 0), (lstCtl, FPos.LT, 0), (lstCtl, FPos.RT, 0),)
		apo = ((lstCtl, FPos.BT, 0, 100),)
		mc.formLayout(lstForm, e=True, af= afo, ap= apo)

		return(lstCtl)

	#### =======================================================================
	def SetupUITemplate_Base(self, name):
		u'''UIテンプレート'''
		CSize = self.CSize

		if(mc.uiTemplate(name, ex=True)): mc.deleteUI(name, uit=True)
		mc.uiTemplate(name)

		mc.rowLayout(dt=name, h= CSize.RHGT)
		mc.button(dt=name, h= CSize.FHGT)
		mc.intField(dt=name, h= CSize.FHGT, w= CSize.CRV, s=1)
		mc.floatField(dt=name, h= CSize.FHGT, w= CSize.SCL, pre=2, s=1.0, min=0.01, max=10000.0)
		mc.text(dt=name, l=u"", h= CSize.FHGT)
		mc.textFieldButtonGrp(dt=name, h= CSize.RHGT+2)
		mc.iconTextCheckBox(dt=name, h= CSize.FHGT, st= ICStyle.TEXT, ebg=True, v=False, bgc=BGCols.HCON3)
		mc.checkBox(dt=name, h= CSize.FHGT, v=False)

	#### =======================================================================
	def Window(self, fn, *arg, **kwd):
		if(1):
			self.wndName = self.SubWind_Rebuild(self.wndName)
		else:
			(self.wndName, isReopen) = self.SubWind_Reuse(self.wndName)
			if(isReopen): return

		self.presetPath = self.Get_PresetName_FromArg(**kwd)
		ctls = self.ctls

		self.SetupUITemplate_Base(self.TEMPLA_BASE)		# make Template

		mc.setUITemplate(self.TEMPLA_BASE, pst=True)	# push Template

		rootForm = mc.formLayout()
		baseClm  = mc.columnLayout(p=rootForm, adj=True)

		#### --------------------------------
		initialPath = self.Get_InitialPath_FromArg(fn)
		ctls.FNAME = self.Create_Ctl_FileLine(baseClm, initialPath, self.Lbls)			# ファイル

		self.Create_Ctl_BaseLine(baseClm, self.ctls, self.CSize, self.Lbls, self.Anno)	# 読込み
		mc.separator()

		self.Create_Ctl_Polygons(baseClm, self.ctls, self.CSize, self.Lbls, self.Anno)	# ポリ
		mc.separator()

		self.Create_Ctl_Sets(baseClm, self.ctls, self.CSize, self.Lbls, self.Anno)		# セット
		mc.separator()

		self.Create_Ctl_Shader(baseClm, self.ctls, self.CSize, self.Lbls, self.Anno)	# シェーダ
		mc.separator()

		self.Create_Ctl_Morphs(baseClm, self.ctls, self.CSize, self.Lbls, self.Anno)	# モーフ
		mc.separator()

		self.Create_Ctl_Bones(baseClm, self.ctls, self.CSize, self.Lbls, self.Anno)		# ボーン
		mc.separator()

		self.Create_Ctl_Curves(baseClm, self.ctls, self.CSize, self.Lbls, self.Anno)	# カーブ
		mc.separator()

		#### レイヤリスト --------------------------------
		if(0):
			ctls.LYRS = self.Create_Ctl_LayerList1(rootForm, baseClm, self.ctls)
		else:
			ctls.LYRS = self.Create_Ctl_LayerList2(rootForm, baseClm, self.ctls)
			#(こっちが有効)
		self.SetupPopupMenu_LyrList(ctls.LYRS)

		# ステータス反映
		self.SyncStat_All()

#		lstC = mc.columnLayout(adj=True)
#		afo = ((lstC, FPos.TP, 0), (lstC, FPos.LT, 0), (lstC, FPos.RT, 0),)
#		apo = ((lstC, FPos.BT, 0, 100),)
		self.Set_LayerList(initialPath)	# ID_LYRS作成後に処理

		mc.setUITemplate(ppt=True)	# pop Template(BASE)

		mc.showWindow(self.wndName)

	#### =======================================================================
	"""
	def SetCtrlHeight_ToSomeGroup(self, ctl, hgt):
		fullpath = mc.layout(ctl, fpn=True, q=True)
		chs = mc.layout(ctl, ca=True, q=True)
		mc.control("{}|{}".format(fullpath, chs[-1]), h= hgt, e=True)
	"""
	#### -----------------------------------------------------------------------
	def Get_GrpLayoutChildCtls(self, ctl):
		u'''ctlはフルパス. 子はlabel | field<N> | button | check<N>等.
通常のgrp系ならctl名無しは無いはず. 深堀すると"layout|"のケースがある.
fieldの数が可変→N=1:n, 単数の物は N 無し.'''
		return( [ "{}|{}".format(ctl, c) for c in mc.layout(ctl, ca=True, q=True) ] )

	#### -----------------------------------------------------------------------
	def EnableControl(self, ctl, stat):
		mc.control(ctl, en= stat, e=True)

	#### -----------------------------------------------------------------------
	def Get_CtlValue_Int(self, ctl, *arg):
		typ = mc.objectTypeUI(ctl, sc=True)[0]

		if(typ == "QmayaIntField"):
			# (intField, [QmayaIntField, QmayaNumericField, QmayaField, QLineEdit, QWidget, QObject])
			return(mc.intField(ctl, q=True, v=True))

		elif(typ == "QmayaIntFieldGrp"):
			# (rowGroupLayout, [QmayaIntFieldGrp, QmayaWidgetGrp, QWidget, QObject])
			v = mc.intFieldGrp(ctl, q=True, v=True)
			return(v[0] if(len(v) == 1) else v)

		elif(typ == "QmayaCheckBox"):
			# (checkBox, [QmayaCheckBox, QCheckBox, QAbstractButton, QWidget, QObject])
			return(mc.checkBox(ctl, q=True, v=True))

		elif(typ == "QmayaCheckBoxGrp"):
			# (rowGroupLayout, [QmayaCheckBoxGrp, QmayaWidgetGrp, QWidget, QObject])
			return(mc.checkBoxGrp(ctl, q=True, va4=True))

		elif(typ == "QmayaIconTextCheckBox"):
			# (iconTextCheckBox, [QmayaIconTextCheckBox, QPushButton, QAbstractButton, QWidget, QObject])
			return(mc.iconTextCheckBox(ctl, q=True, v=True))

		elif(typ == "QmayaIntSlider"):
			# (intSlider, [QmayaIntSlider, QmayaSlider, QSlider, QAbstractSlider, QWidget, QObject])
			return(mc.intSlider(ctl, q=True, v=True))

		elif(typ == "QmayaIntSliderGrp"):
			# (rowGroupLayout, [QmayaIntSliderGrp, QmayaSliderGrp, QmayaConnectableWidgetGrp, QmayaWidgetGrp, QWidget, QObject])
			return(mc.intSliderGrp(ctl, q=True, v=True))

		elif(typ == "QmayaOptionMenu"):
			# (popupMen, [QmayaOptionMen, QmayaLabelledComboBox, QComboBox, QWidget, QObject])
			return(mc.optionMenu(ctl, q=True, sl=True))

		elif(typ == "QmayaRadioButtonGrp"):
			# (rowGroupLayout, [QmayaRadioButtonGrp, QmayaWidgetGrp, QWidget, QObject])
			return(mc.radioButtonGrp(ctl, q=True, sl=True))

		elif(typ == "QmayaColorIndexSliderGrp"):
			# (TcolorIndexSlider, [QmayaColorIndexSliderGrp, QmayaWidgetGrp, QWidget, QObject])
			return(mc.colorIndexSliderGrp(ctl, q=True, v=True))

		return(0)

	#### -----------------------------------------------------------------------
	def Get_CtlValue_Float(self, ctl, *arg):
		typ = mc.objectTypeUI(ctl, sc=True)[0]

		if(typ == "QmayaFloatField"):
			# (floatField, [QmayaFloatField, QmayaNumericField, QmayaField, QLineEdit, QWidget, QObject])
			return(mc.floatField(ctl, q=True, v=True))

		elif(typ == "QmayaFloatFieldGrp"):
			# (rowGroupLayout, [QmayaFloatFieldGrp, QmayaWidgetGrp, QWidget, QObject])
			v = mc.floatFieldGrp(ctl, q=True, v=True)
			return(v[0] if(len(v) == 1) else v)

		return(0)

	#### -----------------------------------------------------------------------
	def SyncStat_Crv(self, ctl, *arg):
		u'''カーブを作成に反応'''
		stat = self.Get_CtlValue_Int(ctl)
		self.EnableControl(self.ctls.DEG,  stat)
		self.EnableControl(self.ctls.DTXT, stat)

	#### --------------------------------
	def SyncStat_Bone(self, ctl, *arg):
		u'''ボーンを作成に反応'''
		self.EnableControl(self.ctls.ZAXZ,  self.Get_CtlValue_Int(ctl))
		self.SyncStat_Wght(self.ctls.ZAXZ)

	#### --------------------------------
	def SyncStat_Wght(self, ctl, *arg):
		u'''ウェイトを読込に反応'''
		pass

	#### --------------------------------
	def SyncStat_Mesh(self, ctl, *arg):
		u'''ポリゴンを作成に反応'''
		stat = self.Get_CtlValue_Int(ctl)
		ctls = self.ctls
		tgt = (
			ctls.R_POLY, ctls.R_SETS, ctls.R_SHAD, ctls.R_IMGS, ctls.R_MORF,
			ctls.UEDG, ctls.SHSG,#ctls.MERG,
			ctls.WXML, ctls.UWGT, ctls.UWCL,
			ctls.UPIV, ctls.VCOL,
		)
		for ctl in tgt:
			self.EnableControl(ctl, stat)

	#### --------------------------------
	def SyncStat_IgnoreHideLayer(self, ctlSelLyr, ctlIgnore, *arg):
		u'''レイヤ選択/不可視を除外'''
		stat = self.Get_CtlValue_Int(ctlSelLyr)
		self.EnableControl(ctlIgnore, stat ^ 1)

		self.EnableControl(self.ctls.LYRS, stat)
		if(stat):
			mc.textScrollList(self.ctls.LYRS, e=True, si= self.tmpSelList)	# 選択
		else:
			self.tmpSelList = mc.textScrollList(self.ctls.LYRS, q=True, si=True) or []
			mc.textScrollList(self.ctls.LYRS, e=True, da=True)				# 選択クリア

	#### --------------------------------
	def SyncStat_Morph(self, ctlMakeMorf, ctlBld, ctlPairMorf, *arg):
		u'''モーフを作成に反応'''
		# MakeMorf<abs|rel>に応じて対になるBldのstate変更
		# MakeMorfと組みになるPairMorfに応じてbldのrowのstate変更
		stat1 = self.Get_CtlValue_Int(ctlMakeMorf)
		stat2 = self.Get_CtlValue_Int(ctlPairMorf)

		self.EnableControl(ctlBld, stat1)
		self.EnableControl(self.ctls.R_BLDS, (stat1|stat2))

	#### --------------------------------
	def SyncStat_VCol(self, *arg):
		u'''頂点色を適用に反応'''
		pass
#		self.EnableControl(self.ctls.MIVC, (self.Get_CtlValue_Int(self.ctls.VCOL) & self.hasMentalray))

	#### --------------------------------
	def SyncStat_LXO(self, *arg):
		u'''LXOに反応'''
		stat = bool(self.currentExt == self.LXO_EXT)

		self.EnableControl(self.ctls.UWCL, stat)
		self.EnableControl(self.ctls.PPIC, stat)

	#### --------------------------------
	def SyncStat_All(self, *arg):
		u'''UI有効無効切り替え'''
		ctls = self.ctls
		self.SyncStat_Mesh(ctls.MESH)
		self.SyncStat_VCol()
		self.SyncStat_Morph(ctls.LREL, ctls.UREL, ctls.LABS)
		self.SyncStat_Morph(ctls.LABS, ctls.UABS, ctls.LREL)
		self.SyncStat_Bone(ctls.BONE)
		self.SyncStat_Wght(ctls.ZAXZ)
		self.SyncStat_Wght(ctls.UWGT)
		self.SyncStat_Wght(ctls.WXML)
		self.SyncStat_Crv(ctls.CURV)
		self.SyncStat_IgnoreHideLayer(ctls.LYR_SEL, ctls.IGN_HIDE)
		self.SyncStat_LXO()

	#### =======================================================================
	def Set_LayerList(self, filename=None, *arg):
		u'''レイヤーリストの更新'''
		# 起動直後、ダイアログ、パス入力時に呼ばれる
		if(self.__Get_CurrentExt(filename)):
			return

		# layersは格納順/物理順. lyr.idxは論理順(ModelerGUI).
		if(self.currentExt == self.LWO_EXT):
			layers = Parse_LWO.ParseLWO().PreScan(filename.encode())
		elif(self.currentExt == self.LXO_EXT):
			layers = Parse_LXO.ParseLXO().PreScan(filename.encode())
		else:
			mc.textScrollList(self.ctls.LYRS, e=True, ra=True)	# リストクリア
			return

		if(layers is None):
			return

		# レイヤ名の最大長
		maxNameLen = 22
		for l in layers:
			maxNameLen = max(maxNameLen, len(l.name))
		maxNameLen += 1

		lyrVisList = []
		layerList  = []
		visAPD = lyrVisList.append
		lyrAPD = layerList.append
		FMT_BASE = ": {{:<{PAD}}}:VC{{}} EW{{}} Rel{{}} Abs{{}} Bon{{}} Crv{{}}".format(PAD=maxNameLen)
		FMT_VIS  = "{:03}({:03})" + FMT_BASE
		FMT_HIDE = "{:03}[{:03}]" + FMT_BASE
		mk = ("-", "+")
		for (fi, lyr) in enumerate(layers):	# lwo内の格納順で列挙
			hasVC  = mk[bool(lyr.rgb)|bool(lyr.rgba)]
			hasEdg = mk[bool(lyr.hasEdge|lyr.hasEdgeM)]
			hasRel = mk[bool(lyr.relMorphs)]
			hasAbs = mk[bool(lyr.absMorphs)]
			hasBon = mk[bool(lyr.bonNum)]
			hasCrv = mk[bool(lyr.crvNum)]

			# 表示上,ModelerのGUI合わせでlyr.idx +1する
			FMT = FMT_HIDE if(lyr.isHide) else FMT_VIS
			lyrAPD(FMT.format(lyr.idx +1, fi, lyr.name, hasVC, hasEdg, hasRel, hasAbs, hasBon, hasCrv))
			visAPD([lyr.idx, lyr.isHide])	# [lwoの論理idx,属性]

		# 論理順でソートして格納
		self.lyrVisList = sorted(lyrVisList, key= lambda x: x[0])

		layerList.sort()

		mc.textScrollList(self.ctls.LYRS, e=True, ra=True, a= layerList)	# リストクリア, 項目追加は同時可

		if(self.Get_CtlValue_Int(self.ctls.LYR_SEL)):
			mc.textScrollList(self.ctls.LYRS, e=True, si= layerList)		# 選択はクリア/項目追加と同時不可

		self.SetListViewColor()			# 着色

		self.tmpSelList = layerList[:]	# 新規項目の退避

	#### =======================================================================
	def SetListViewColor(self):
		u'''リストの着色'''
		qlw = self.GetListWidget(self.ctls.LYRS)
		cnt = 0
		# lwoの論理idxはtextScrollListの行idxと合致する保証は無い
		for (_, isHide) in self.lyrVisList:	# [lwoの論理idx,属性]
			if(isHide):
				it = qlw.item(cnt)
				qlw.editItem(it.setForeground(self.qTxColorBrush))
				qlw.editItem(it.setBackground(self.qBGColorBrush))
			cnt += 1

	#### =======================================================================
	def ConvertOpts_FromUI(self, optsUI):
		u'''UI設定から実行オプションへコピー. UI固有の項目はコピー不要'''
		opts = Opts.OptsObject()

		for k in opts.keys():
			if("__" not in k):
				opts[k] = optsUI[k]
		opts.scl = optsUI.scl
		return(opts)

	#### =======================================================================
	def Get_UI_Options(self):
		u'''UIから設定取得'''
		ctls = self.ctls
		optsUI = self.optsUI

		for (o, c) in (
			("isLyrSel", "LYR_SEL"),
			("isMesh","MESH"), ("isBone","BONE"), ("isCrv","CURV"),
			("useWgt","UWGT"), ("useXML","WXML"), ("isZAxis", "ZAXZ")
			):
			optsUI[o] = self.Get_CtlValue_Int(ctls[c])

		for (o, c) in (
			("useEdge",  "UEDG"), ("usePivot","UPIV"),
			("loadRM",   "LREL"), ("loadAM",  "LABS"), ("useRM", "UREL"), ("useAM", "UABS"),
			("useVCol",  "VCOL"),
			("shareSG",  "SHSG"), ("useColT", "UCOL"), ("useBmpT", "UBMP"),
			("usePick",  "PICK"), ("useColr", "COLR"), ("usePart", "PART"),
			("usePPic", "PPIC"), ("useWire", "UWCL"),
			("isDblSide","DBLS"), ("skipHide","IGN_HIDE")
			#("mergeUV","MERG"), ("useMiVC", "MIVC"),
			):
			optsUI[o] = self.Get_CtlValue_Int(ctls[c])

		optsUI.scl = self.Get_CtlValue_Float(ctls.SCL)
		optsUI.deg = self.Get_CtlValue_Int(ctls.DEG)
		optsUI.mtlType= mc.radioButtonGrp(ctls.UBLN, q=True, sl=True) -1	# 戻りは1からなので-1

		return(optsUI)

	#### =======================================================================
	def Store_UI_Options(self):
		u'''optionVar保存'''
		OptVar = self.OptVar
		optsUI = self.optsUI

		mc.optionVar(fv= (OptVar.SCALING, optsUI.scl))

		for o in (
			(OptVar.CURV_DEGREE, optsUI.deg),
			(OptVar.SHADER_TYPE, optsUI.mtlType),
			(OptVar.LYR_SEL, optsUI.isLyrSel),
			(OptVar.ING_HID, optsUI.skipHide),
			(OptVar.MAKE_POLY, optsUI.isMesh),
			(OptVar.MAKE_BONE, optsUI.isBone),
			(OptVar.MAKE_CURV, optsUI.isCrv),
			(OptVar.MERGEUV, optsUI.mergeUV),
			(OptVar.SHARE_SG, optsUI.shareSG),
			(OptVar.USE_EDGEWGT, optsUI.useEdge),
			(OptVar.USE_PIVOT, optsUI.usePivot),
			(OptVar.USE_VERTCOLR, optsUI.useVCol),
#			(OptVar.USE_MR_VCOLR, (optsUI.useMiVC & self.hasMentalray)),
			(OptVar.USE_DBL_SIDE, optsUI.isDblSide),
			(OptVar.USE_SELSET, optsUI.usePick),
			(OptVar.USE_COLR_TAG, optsUI.useColr),
			(OptVar.USE_PART_TAG, optsUI.usePart),
			(OptVar.USE_PICK_TAG, optsUI.usePPic),
			(OptVar.USE_WIRECOLR, optsUI.useWire),
			(OptVar.USE_IMAGE_COL, optsUI.useColT),
			(OptVar.USE_IMAGE_BMP, optsUI.useBmpT),
			(OptVar.LOAD_MORF_REL, optsUI.loadRM),
			(OptVar.LOAD_MORF_ABS, optsUI.loadAM),
			(OptVar.MAKE_BLD_REL, optsUI.useRM),
			(OptVar.MAKE_BLD_ABS, optsUI.useAM),
			(OptVar.BONE_TYPE, optsUI.isZAxis),
			(OptVar.OUT_WGT_XML, optsUI.useXML),
			(OptVar.OUT_WGT_ATTR, optsUI.useWgt),
			):
			mc.optionVar(iv= o)

	#### =======================================================================
	def Init_UI_Options(self, *arg):
		u'''UIを初期値に.'''
		ctls = self.ctls
		optsUI = self.optsUI

		for (o, c, v) in (
			("isLyrSel","LYR_SEL",False),
			("isMesh","MESH",True), ("isBone","BONE",False), ("isCrv",  "CURV",False),
			("useWgt","UWGT",False),("useXML","WXML",False), ("isZAxis","ZAXZ",True)
			):
			mc.iconTextCheckBox(ctls[c], e=True, v= v)

		for (o, c, v) in (
			("useEdge",  "UEDG",True), ("usePivot","UPIV",False),
			("loadRM",   "LREL",False),("loadAM",  "LABS",False), ("useRM", "UREL",False), ("useAM", "UABS",False),
			("useVCol",  "VCOL",False),
			("shareSG",  "SHSG",True), ("useColT", "UCOL",False), ("useBmpT", "UBMP",False),
			("usePick",  "PICK",False),("useColr", "COLR",False), ("usePart", "PART",False),
			("usePPic",  "PPIC",False),("useWire", "UWCL",False),
			("isDblSide","DBLS",False),("skipHide","IGN_HIDE",True)
			#("mergeUV","MERG",True), ("useMiVC", "MIVC",False),
			):
			mc.checkBox(ctls[c], e=True, v= v)

		mc.floatField(ctls.SCL, e=True, v= optsUI.defaultScl)
		mc.intField(  ctls.DEG, e=True, v= optsUI.defaultDeg)
		mc.radioButtonGrp(ctls.UBLN, e=True, sl= 1)

		self.Store_UI_Options()
		self.SyncStat_All()

	#### =======================================================================
	def Exec(self, *arg, **kwd):
		ctls = self.ctls

		# ファイル有無に関係なくオプション取得と保存
		self.Get_UI_Options()
		self.Store_UI_Options()

		fname= mc.textFieldButtonGrp(ctls.FNAME, q=True, tx=True)
		if(not os.path.isfile(fname)):
			mc.warning(self.Msgs.NOSEL_FILE),
			return

		# actListはfile内格納順に応じたboolのリスト
		actList = None

		if(self.optsUI.isLyrSel):
			selList = mc.textScrollList(ctls.LYRS, q=True, si=True) or []	# GUI上の項目名のリスト
			if(not selList):
				mc.warning(self.Msgs.NOSEL_LAYER),
				return

			allNum  = mc.textScrollList(ctls.LYRS, q=True, ni=True)	# 時点のGUI上のリスト総数
			actList = [False] * allNum

			# GUIIdx1(FileIdx0): name~ | GUIIdx1[FileIdx0]: name~, 可視属性によって()[]が違う
			pat = re.compile("^\d+[\(\[](\d+)[\)\]]:")
			for s in selList:
				idx = int(pat.findall(s)[0])	# 戻りはリスト
				actList[idx] = True

		if((self.optsUI.isMesh|self.optsUI.isBone|self.optsUI.isCrv) == 0):
			mc.warning(self.Msgs.NOSEL_MAKE),
			return

		if(fname.lower().endswith(self.LWO_EXT)):
			inst = Parse_LWO.ParseLWO(fn= fname.encode(), **self.VisFlags.GetKwd())
		elif(fname.lower().endswith(self.LXO_EXT)):
			inst = Parse_LXO.ParseLXO(fn= fname.encode(), **self.VisFlags.GetKwd())
		else:
			return

		if(not inst.lwo.layers):
			return

		"""
		# 2017? 2016? から追加されたuvEditorの[テクスチャを自動ロード]によるtextureWindowUpdateUVSets,
		# textureWindowAutoLoadTexturesが同内でgetAttr(node.attr)での階層違い同名node(戻りは配列)を考慮せず,
		# 文字列として決め打ち処理をしているせいでエラーになる.
		# uvEditorの平時も同名nodeがあるとエラーになる.
		# データbuild中はcmds.mergeUVに際してchangeSelect/changeUVSetからコールバックが発生するので,無効にして対処.
		# cmds.mergeUV相当をMFnMeshでする分にはselectが発生しない為,駆動されない.
		autoLoadState = None
		Ge2017 = int(mc.about(v=1)) >= 2017)
		if(Ge2017 and mc.optionVar(ex="textureEditorAutoLoadTexturesState")):
			autoLoadState = mc.optionVar(q="textureEditorAutoLoadTexturesState")
			mc.optionVar(iv= ("textureEditorAutoLoadTexturesState", 0))
		"""

		uiOpts = self.ConvertOpts_FromUI(self.optsUI)

		undoState = mc.undoInfo(st=True, q=True)
		if(undoState):
			# stateWithoutFlushでFalseにする. state=,edit=は不要
			mc.undoInfo(swf=False)

		try:
			if(inst.fileType == "LWO2"):
				Build_LWO.BuildObjects(inst.lwo, o= uiOpts, al=actList)
			else:
				Build_LXO.BuildObjects(inst.lwo, o= uiOpts, al=actList)
		except:
			pass
		finally:
			if(undoState):
				# stateWithoutFlushでTrueにする. state=,edit=は不要
				mc.undoInfo(swf=True)

		"""
		# 無効にしていたコールバックを戻す.
		if(Ge2017 and (autoLoadState != None)):
			mc.optionVar(iv= ("textureEditorAutoLoadTexturesState", autoLoadState))
		"""

		self.RefineRecentNameArray(fname)	# ファイル履歴
