﻿# -*- coding: utf-8 -*-

from __future__ import absolute_import

import math

import maya.OpenMaya as om
import maya.cmds as mc

from ... import LXO_ParseHeader
from ... import FnCalc
from .. import Defines
from . import FnOM1

#### =======================================================================
__all__ = ("Grouping", "MeshInstancing", "SetMeshesNodeAttr", "MakeNodes", "ParentAndMove", "Rename",)

#### ========================================================================
def Grouping(isVis, arTrs, name):
	'''グループ化.
空トランスフォームを作り, arTrs内のmoを子にする.
Grouping(isVis, arTrs, name)
isVis: 可視フラグ
arTrs: [mo,...]
name: トランスフォーム名
return: トランスフォームのMDagPath'''
	if(not arTrs):
		return(None)		# moでなくNoneで.

	FnTrs = om.MFnTransform()
	FnTrs.create()
	FnTrs.setName(name)

	FnTrs.findPlug("v", False).setBool(isVis)

	for mo in arTrs:
		FnTrs.addChild(mo)

	dagp = om.MDagPath()
	FnTrs.getPath(dagp)
	return(dagp)

#### ========================================================================
#### 以下 LXO用
def __Set_Attrs(FnTrs, obj, optScl):
	FnTrs.findPlug("v", False).setBool(obj.isVis)

	if(obj.pivot != (0.0, 0.0 ,0.0)):
		piv = FnCalc.VScl(obj.pivot, optScl)
		mpPiv = om.MPoint(*FnCalc.VScl(obj.pivot, optScl))
		FnTrs.setRotatePivot(mpPiv, om.MSpace.kTransform, False)
		FnTrs.setScalePivot(mpPiv, om.MSpace.kTransform, False)
#		"pivotRotate", "pivotRotateOrder",

	if(0):	# overrideで色付けの場合
		FnTrs.findPlug("ove", False).setBool(True)
		FnTrs.findPlug("ovrgbf", False).setBool(True)
		FnTrs.findPlug("ovcr", False).setFloat(obj.lyrRGBA[0])
		FnTrs.findPlug("ovcg", False).setFloat(obj.lyrRGBA[1])
		FnTrs.findPlug("ovcb", False).setFloat(obj.lyrRGBA[2])

#### ------------------------------------------------------------------------
def __Set_MeshAttrs(obj, useWire):
	dagp = om.MDagPath(obj.buildedObject)
	dagp.extendToShape()

	FnMe = om.MFnMesh(dagp)

	FnMe.findPlug("displaySmoothMesh", False).setInt(obj.isSds * 2)	# 0/1/2.

	FnMe.findPlug("useGlobalSmoothDrawType", False).setBool(False)
	FnMe.findPlug("smoothDrawType", False).setInt(0)	# 0:MCC, 2:OSD, 3:adpOSD

	FnMe.findPlug("smoothLevel", False).setShort(0)
	FnMe.findPlug("useSmoothPreviewForRender", False).setBool(False)
	FnMe.findPlug("renderSmoothLevel", False).setShort(int(obj.sdsLvl))

	# OSD
	FnMe.findPlug("osdFvarBoundary", False).setInt(1)	# 0:None, 1:Edge&Corner, 2:Edge, 3:MCC
	# MCC
	FnMe.findPlug("smoothUVs", False).setBool(True)
	FnMe.findPlug("keepMapBorders", False).setInt(0)	# 0:None, 1:Internal, 2:All

	# .setSmoothMeshDisplayOptions(om.MMeshSmoothOptions())はosd要素に対応していない

	if(useWire  and (obj.lyrRGBA is not None)):
		# RGBが効くのはmesh/crv/jointのみ? locatorはnd.use=1/nd.ocのみ?(nd.use=2だとnd.oc=0で作用)
		# shapeに色を付けるのでインスタンス時はModoと非互換になる.
		# 互換にするにはtransform.overrideで色付けする.
		# overrideと違って親から継承されない
#		FnMe.useObjectColor = bool						# bool(objectColorType) 自動更新.代入不要
		FnMe.setObjectColorType(FnMe.kUseRGBColor)		# nd.useObjectColor(0:none, 1:index, 2:rgb)
		FnMe.setObjectColor(om.MColor(*obj.lyrRGBA))	# nd.wireColorRGB
#		FnMe.setObjectColor(index)						# nd.objectColor(index 0-7)

#### ------------------------------------------------------------------------
def __Set_CurveAttrs(rootDagp, useWire, lyrRGBA):
	FnDag = om.MFnDagNode()
	Fn = om.MFnDagNode()

	for i in range(rootDagp.childCount()):
		FnDag.setObject(rootDagp.child(i))
		dagp = om.MDagPath()
		FnDag.getPath(dagp)
		dagp.extendToShape()

		Fn.setObject(dagp)
		Fn.findPlug("useObjectColor", False).setInt(2)
		Fn.findPlug("wireColorR", False).setFloat(lyrRGBA[0])
		Fn.findPlug("wireColorG", False).setFloat(lyrRGBA[1])
		Fn.findPlug("wireColorB", False).setFloat(lyrRGBA[2])
		# shapeに色を付けるのでインスタンス時はModoと非互換になる.
		# 互換にするにはModo上での要素のインスタンス継承を考慮した上で
		# transform.overrideで色付けする.

#### ------------------------------------------------------------------------
def __Make_Transform(obj, optScl):
	FnTrs = om.MFnTransform()
	FnTrs.create()
	FnTrs.setName(obj.name)

	__Set_Attrs(FnTrs, obj, optScl)

	dagp = om.MDagPath()
	FnTrs.getPath(dagp)	# .dagPath()不可
	return(dagp)

#### ------------------------------------------------------------------------
def __Make_Locator(obj, optScl):
	dagp = __Make_Transform(obj, optScl)

	mdg = om.MDagModifier()
	mo = mdg.createNode("locator", dagp.node())
	mdg.doIt()

	size = obj.size * optScl
	FnDep = om.MFnDependencyNode(mo)
	FnDep.findPlug("lsx", False).setFloat(size)
	FnDep.findPlug("lsy", False).setFloat(size)
	FnDep.findPlug("lsz", False).setFloat(size)

	return(dagp)

#### ------------------------------------------------------------------------
def SetMeshesNodeAttr(lxo, useWire, optScl):
	# layerから作成された/インスタンス作成した分のattrを設定
	FnTrs = om.MFnTransform()
	for obj in lxo.meshes.values():
		if(obj.buildedObject):
			FnTrs.setObject(obj.buildedObject)	# Meshのtransform
			__Set_Attrs(FnTrs, obj, optScl)
			if("meshInst" != obj.ownItem.type):	# 非instanceならmeshShapeのsds設定
				__Set_MeshAttrs(obj, useWire)

		if(obj.buildedCrvs):
			FnTrs.setObject(obj.buildedCrvs[0])	# CurveGroupのtransform
			__Set_Attrs(FnTrs, obj, optScl)
			if("meshInst" != obj.ownItem.type):	# 非instanceならcurveShapeの色
				if(useWire and (obj.lyrRGBA is not None)):
					__Set_CurveAttrs(obj.buildedCrvs[0], useWire, obj.lyrRGBA)

#### ------------------------------------------------------------------------
def MakeNodes(lxo, optScl, totalDct):
	u'''trs/locの既存Nodeとの重複を考慮しない作成'''
	tmp = []
	EXT = tmp.extend
	for t in [ obj.treeParentsList for (uniq, obj) in lxo.meshes.items() if(obj.buildedObject) ]:
		EXT(t)
	meshUniqNamesSet = set(tmp)

	# 未作成のmeshが作成済meshの階層経路になる場合は代替の空transformを作る
	for (uniq, obj) in lxo.meshes.items():
		if(obj.buildedObject is None):
			if(uniq in meshUniqNamesSet):	# 必要な経路に含まれているなら
				obj.buildedObject = __Make_Transform(obj, optScl)
				obj.isPlaceHolder = True
				break

	for (uniq, obj) in lxo.transforms.items():
		if(uniq in meshUniqNamesSet):	# 必要な経路に含まれているなら
			obj.buildedObject = __Make_Transform(obj, optScl)

	for (uniq, obj) in lxo.locators.items():
		if(uniq in meshUniqNamesSet):	# 必要な経路に含まれているなら
			obj.buildedObject = __Make_Locator(obj, optScl)

#### ------------------------------------------------------------------------
def ParentAndMove(lxo, optScl, totalDct):
	u'''親子組'''
	def __RoundRadian(rad):
		return(math.radians(round(math.degrees(rad) * 100000.0) * 0.00001))
	#### ----------------
	def __Round_Scaling(v, scl):
		return(round(v * 100000000.0) * 0.00000001 * scl)
	#### ----------------
	FnTrs = om.MFnTransform()
	for obj in totalDct.values():
		hasCrv = False if(obj.type != "mesh") else bool(obj.buildedCrvs)

		if(obj.buildedObject is None):		# mesh/instance無し?
			if(not hasCrv):					# カーブ要素も無いなら次へ
				continue
		# mesh/curve何れかがある

		if(obj.parentUniqueIdentifier is not None):
			FnTrs.setObject(totalDct[obj.parentUniqueIdentifier].buildedObject)	# 直親のdagp
			if(obj.buildedObject):
				FnTrs.addChild(obj.buildedObject.node())
			if(hasCrv):
				FnTrs.addChild(obj.buildedCrvs[0].node())

		# lxoは単精度なので変に精度が低い. 3.0→2.9~9nnnnなので丸めてからスケールする
		tx = __Round_Scaling(obj.translate[0], optScl)
		ty = __Round_Scaling(obj.translate[1], optScl)
		tz = __Round_Scaling(obj.translate[2], optScl)
		mv = om.MVector(tx, ty, tz)

		# lxoは単精度のradianなので変に精度が低い. deg変換してから丸めてradianに戻す
		rx = __RoundRadian(obj.rotate[0])
		ry = __RoundRadian(obj.rotate[1])
		rz = __RoundRadian(obj.rotate[2])
		# omではEulerのorderはkwd指定は出来ない. omではsetRotのMSpace指定は無い
		er = om.MEulerRotation(rx, ry, rz, Defines.EulerRotateOrderT[obj.rotateOrder])

		MSU = om.MScriptUtil()	# omではMSU経由が必要
		MSU.createFromDouble(*obj.scale)
		ptr = MSU.asDoublePtr()

		if(obj.buildedObject):
			FnTrs.setObject(obj.buildedObject)
			FnTrs.setTranslation(mv, om.MSpace.kObject)
			FnTrs.setRotation(er)
			# FnTrs.setRotation(om.MEulerRotation(obj.rotate[0], obj.rotate[1], obj.rotate[2], Defines.EulerRotateOrderT[obj.rotateOrder]))
			# FnTrs.setRotationOrder()はMEulerRotationの段で指定済なので不要. orderのenumは2者で違うので注意
			FnTrs.setScale(ptr)

		if(hasCrv):
			FnTrs.setObject(obj.buildedCrvs[0])
			FnTrs.setTranslation(mv, om.MSpace.kObject)
			FnTrs.setRotation(er)
			FnTrs.setScale(ptr)

#### ------------------------------------------------------------------------
def Rename(lxo):
	u'''リネーム'''
	def __Rename_Shape(Fn, srcDagp, name):
		dagp = om.MDagPath(srcDagp)
		dagp.extendToShape()
		Fn.setObject(dagp)
		Fn.setName("{}Shape".format(name))
	#### ----------------
	def __Rename_Transform(Fn, obj):
		if(obj.buildedObject):
			Fn.setObject(obj.buildedObject)
			Fn.setName(obj.name)
	#### ----------------
	def __Rename_Mesh(Fn, obj):
		if(obj.buildedObject):
			Fn.setObject(obj.buildedObject)

			# 未作成Meshのダミーtransform
			name = "{}_PlaceHolder".format(obj.name) if(obj.isPlaceHolder) else obj.name
			Fn.setName(name)

			if(not Fn.isInstanced()):
				# インスタンスならシェイプの変更はしない
				__Rename_Shape(Fn, obj.buildedObject, Fn.name())	# transformの変化後の名
	#### ----------------
	def __Rename_Locator(Fn, obj):
		if(obj.buildedObject):
			Fn.setObject(obj.buildedObject)
			Fn.setName(obj.name)

			if(not Fn.isInstanced()):
				# インスタンスならシェイプの変更はしない
				__Rename_Shape(Fn, obj.buildedObject, Fn.name())	# transformの変化後の名
	#### ----------------
	Fn = om.MFnDagNode()

	for obj in lxo.transforms.values():
		__Rename_Transform(Fn, obj)

	for obj in lxo.meshes.values():
		__Rename_Mesh(Fn, obj)

	for obj in lxo.locators.values():
		__Rename_Locator(Fn, obj)

#### ========================================================================
def MeshInstancing(lxo):
	u'''インスタンス作成'''
#	FnSrcDag = om.MFnDagNode()
#	FnDstDag = om.MFnDagNode()
	for (uniq, obj) in lxo.meshes.items():
		if("meshInst" == obj.ownItem.type):
			if("source" in obj.ownItem.itemLinks):
				srcObj = lxo.meshes[lxo.items[obj.ownItem.itemLinks.source[0].id].uniqueIdentifier]
				if(srcObj.buildedObject is not None):
					# srcMeshが作成されていたら作成. Shapeだけインスタンス.
					r = mc.duplicate(srcObj.buildedObject.fullPathName(), ilf=True, po=False, ic=False, rc=True)[0]
					obj.buildedObject = FnOM1.GetDagPath(FnOM1.MakeSelectionList(r), 0)

					# SDSかどうかの継承. MayaはMeshで処理するから不要.
					obj.isSds = srcObj.isSds

					# Modoでのインスタンス→変更可能なtrs&mesh.
					# Modoではtrsに明示的な色指定が無い場合は元trsの状態が継承される.
					if(srcObj.lyrRGBA is not None):
						if(isinstance(obj.lyrRGBA, list)):	# list時: ModoのUI設定値
							obj.lyrRGBA = srcObj.lyrRGBA

				if(srcObj.buildedCrvs):		# [dagp]
					# カーブがあったら作成. Shapeだけインスタンス.
					r = mc.duplicate(srcObj.buildedCrvs[0].fullPathName(), ilf=True, po=False, ic=False, rc=True)[0]
					obj.buildedCrvs.append(FnOM1.GetDagPath(FnOM1.MakeSelectionList(r), 0))

#					FnSrcDag.setObject(srcObj.buildedObject)
					# 複製のmo出力だけで, Fn内の更新は起きない. SG適用がされない.
#					FnDstDag.setObject(FnSrcDag.duplicate(True, False))
#					dagp = om.MDagPath()
#					FnDstDag.getPath(dagp)	# .dagPath()不可
#					obj.buildedObject = dagp


#### ========================================================================
def Set_OutlinerColor(lxo):
	u'''アイテムカラー適用(outliner色として)'''
	def __OutlinerColor(FnDep, mo, r, g, b):
		FnDep.setObject(mo)
		FnDep.findPlug("useOutlinerColor", False).setBool(True)
		FnDep.findPlug("outlinerColorR", False).setFloat(r)
		FnDep.findPlug("outlinerColorG", False).setFloat(g)
		FnDep.findPlug("outlinerColorB", False).setFloat(b)
	#### ----------------
	def __Set(objDct):
		FnDep = om.MFnDependencyNode()
		for (uniq, obj) in objDct.items():
			r = g = b = 0.0
			for itg in obj.ownItem.itemTags:
				if(itg.lwid == "COLR"):
					(r,g,b) = LXO_ParseHeader.ITEM_ITAG_COLR[itg.v]
					break
			else:
				continue

			if(obj.buildedObject):
				__OutlinerColor(FnDep, obj.buildedObject.node(), r, g, b)

			if((obj.type=="mesh") and obj.buildedCrvs):
				__OutlinerColor(FnDep, obj.buildedCrvs[0].node(), r, g, b)
	#### ----------------
	__Set(lxo.meshes)
	__Set(lxo.transforms)
	__Set(lxo.locators)
