﻿# -*- coding: utf-8 -*-

from __future__ import absolute_import

import sys

import maya.OpenMaya as om
import maya.OpenMayaAnim as oa
import maya.api.OpenMaya as om2
import maya.cmds as mc

from ... import FnCalc
from .. import Defines, Opts, PrintMsg
from . import Conv1To2

if(sys.modules["Load_LWO"].DO_RELOAD):
	if(sys.version_info[0] >= 3):
		import importlib
		Fn = importlib.reload
	else:
		Fn = reload
	loc = set(locals().keys())
	for f in ("Defines", "Opts", "FnCalc", "PrintMsg", "Conv1To2",):
		if(f in loc):
			Fn(eval(f))
	del f, loc

# メモ: 2022になってもoa2.MFnIkJointは無い.
# 途中でom1 -> om2をする.

#### ========================================================================
__all__ = ("MakeJointObject",)

#### ========================================================================
def MakeJointObject(lyr, vScl= Opts.OptsObject.defaultVScl, isZAxis=True):
	u'''ジョイント作成, Joint風'''
	if(lyr.bonNum == 0):
		return([])

	PrintMsg.BuildJoint(lyr)

	arDagp1 = __Make_Joints(lyr.pnts, lyr.bones, vScl, isZAxis)

	#### om1からom2へ.
	arDagp = Conv1To2.arDagPath(arDagp1)

	if(isZAxis):
		__AppendSkelgonEditorParent(lyr.bones)
	__Parenting(arDagp, lyr.bones)

	arTrs = __RootTransform(arDagp)
	__Renaming(arDagp, lyr.bones)

	return(arTrs)

#### ========================================================================
def __RootTransform(arDagp):
	u'''ジョイントチェーンの各上流にトランスフォームを作成, 親子付け'''
	MMx = om2.MTransformationMatrix()	# 単位マトリクスとして
	FnTrs = om2.MFnTransform()
	FnParent = om2.MFnTransform()

	arTrs = []
	APD = arTrs.append

	for dagp in arDagp:
		FnTrs.setObject(dagp)									# dagPath可
		if(FnTrs.parent(0).apiType() == om2.MFn.kWorld):		# 親がワールド
			APD(FnParent.create())								# 親になるTransform mo
			FnParent.setTransformation(FnTrs.transformation())	# 自身と同じ位置
			FnParent.setName(Defines.NameFMT.JNTGRP.format(NAME= FnTrs.name()))	# 自身名＋サフィックス
			FnParent.addChild(dagp.node())						# dagPath不可, 要MObj
			FnTrs.setTransformation(MMx)						# 自身の位置を初期化
	return(arTrs)

#### ========================================================================
def __RebuildMx(mx, pt0, pt1, upvec):
	fwd = om.MVector(pt1[0]-pt0[0], pt1[1]-pt0[1], (-pt1[2])-(-pt0[2])).normal()
	rgt = (fwd ^ om.MVector(upvec[0], upvec[1], -upvec[2]).normal()).normal()
	up  = (rgt ^ fwd).normal()
	return([fwd.x, fwd.y, fwd.z, 0.0, up.x, up.y, up.z, 0.0, rgt.x, rgt.y, rgt.z, 0.0] + mx[12:16])

#### -----------------------------------------------------------------------
def __IsZeroVector(name, vec):
	# DrawSkelegonによってBNUP(vec)が<0,0,0>の場合が有り得る
	if(vec == (0.0, 0.0, 0.0)):
		mc.warning(u"{} の BNUP がゼロベクトル.".format(name)),
		return(True)
	return(False)

#### -----------------------------------------------------------------------
def __Make_Joints(pnts, bones, vScl, isZAxis):
	u'''ジョイント作成, ワールドにばら撒く, 結果をdagPathとして返す'''
	# この関数だけom1系.
	FnJt = oa.MFnIkJoint()
	FnJt2= oa.MFnIkJoint()

	FnMDG = om.MDGModifier()

	mmx = om.MMatrix()
	rmx = [1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0]

	ups = (None, None,
		( 1.0, 0.0, 0.0), (-1.0, 0.0, 0.0),
		( 0.0, 1.0, 0.0), ( 0.0,-1.0, 0.0),
		( 0.0, 0.0, 1.0), ( 0.0, 0.0,-1.0),
	)

	arDagp = []
	APD = arDagp.append

	for (_, bone) in sorted(bones.items(), key= lambda x: x[0]):
		# [0]根本,[1]先端
		pt0 = pnts[bone.pnt0]
		pt1 = pnts[bone.pnt1]
		mx = rmx + FnCalc.VMul(pt0 if(isZAxis) else pt1, vScl) + [1.0]	# 4x3行列+スケールしたwpos(3)+埋め

		# バンクモード, enum 0:BNUP依存 1:無変更 2:+X 3:-X 4:+Y 5:-Y 6:+Z 7:-Z 8:任意値
		if(bone.skbm != None):			# SkelegonEditor製Skelegon
			bankMode = bone.skbm
			if(bankMode == 0):
				if((bone.bnup != None) and (not __IsZeroVector(bone.name, bone.bnup))):
					mx = __RebuildMx(mx, pt0, pt1, bone.bnup)
			elif(2 <= bankMode <= 7):
				mx = __RebuildMx(mx, pt0, pt1, ups[bankMode])

			elif((bankMode == 8) and (bone.skbr != None)):
#				localBank = bone[BoneT.SKBR]
				pass
		else:								# 素のSkelegon
			if((bone.bnup != None) and (not __IsZeroVector(bone.name, bone.bnup))):
				mx = __RebuildMx(mx, pt0, pt1, bone.bnup)

		om.MScriptUtil.createMatrixFromList(mx, mmx)

		# joint作成
		FnJt.create()
		FnJt.set(om.MTransformationMatrix(mmx))

		dagp = om.MDagPath()
		FnJt.getPath(dagp)
		APD(dagp)

		if(isZAxis and (not bone.childs)):
			dstmo = FnJt2.create()
			FnJt.addChild(dstmo)	# dagPath不可, 要MObj

			mx[12:15] = FnCalc.VMul(pt1, vScl)
			om.MScriptUtil.createMatrixFromList(mx, mmx)
			mmx *= dagp.inclusiveMatrixInverse()	# dagPathから逆行列, 親座標系, MMatrix型
			FnJt2.set(om.MTransformationMatrix(mmx))

			FnMDG.connect(FnJt.object(), FnJt.attribute("s"), dstmo, FnJt2.attribute("is"))	# inverseScale
			FnJt2.findPlug("radi", False).setDouble(__DetermineJointRadius(FnJt2.getTranslation(om.MSpace.kObject).length()))	# radius
			FnJt2.setName("{}_end".format(bone.name))

	if(isZAxis):
		FnMDG.doIt()

	return(arDagp)

#### ========================================================================
def __AppendSkelgonEditorParent(bones):
	u'''skelegonEditorの親指定をbones[].childsに追加'''
	for (i, bone) in sorted(bones.items(), key=lambda x: x[0]):
		if(bone.skpr == None): continue
		for (k, b) in bones.items():
			if(b.skid == bone.skpr):
				b.childs.append(i)
				break

	# 頂点共有親子へもskprが付くので重複がありえる
	for (i, bone) in bones.items():
		if(bone.childs):
			bone.childs = list(set(bone.childs[:]))

#### -----------------------------------------------------------------------
def __DetermineJointRadius(length):
	return(min(Defines.JtSizeT.MAX, max(Defines.JtSizeT.MIN, (length * Defines.JtSizeT.COEF))))

#### -----------------------------------------------------------------------
def __Parenting(arDagp, bones):
	u'''dagPathのジョイントをlwoに合わせて親子付け'''
	FnTrs  = om2.MFnTransform()
	FnChild= om2.MFnTransform()
	FnMDG  = om2.MDGModifier()

	for (i, bone) in sorted(bones.items(), key= lambda x: x[0]):
		if(not bone.childs):
			continue

		FnTrs.setObject(arDagp[i])					# dagPath可, 骨をset
		iMx = arDagp[i].inclusiveMatrixInverse()	# dagPathから逆行列, 親座標系, MMatrix型
		srcmo = FnTrs.object()						# mo取り出し
		srcma = FnTrs.attribute("s")				# scale

		isFirst = True
		for c in bone.childs:						# [idx,idx,,]
			# 親座標系へ, MMatrix型同士なので掛け算可, incMatは階層変更前に取る必要有り
			iMx0 = arDagp[c].inclusiveMatrix() * iMx

			FnChild.setObject(arDagp[c])			# dagPath可, 子になる骨をset
			dstmo = FnChild.object()
			FnTrs.addChild(dstmo)					# dagPath不可, 要MObj
			FnChild.setTransformation(om2.MTransformationMatrix(iMx0))	# MMatrixから変換して適用.

			FnMDG.connect(srcmo, srcma, dstmo, FnChild.attribute("is"))	# inverseScale

			radius = __DetermineJointRadius(FnChild.translation(om2.MSpace.kObject).length())
			FnChild.findPlug("radi", False).setDouble(radius)	# radius

			if(isFirst):	# 第1子と同じ直径を使う
				FnTrs.findPlug("radi", False).setDouble(radius)	# radius
				isFirst = False
	FnMDG.doIt()

#### ========================================================================
def __Renaming(arDagp, bones):
	u'''ワールドで作る都合上, 名前付けは階層を組んでから行う'''
	FnDag = om2.MFnDagNode()
	for (i, bone) in sorted(bones.items(), key= lambda x: x[0]):
		FnDag.setObject(arDagp[i])
		FnDag.setName(bone.name)
