﻿# -*- coding: utf-8 -*-

from __future__ import absolute_import

import sys, re

import maya.cmds as mc
import maya.api.OpenMaya as om2

from ... import FnCalc
from .. import Weights, Shaders, PrintMsg, Defines, Smooth
from . import BaseMesh, VColors, Edges, Surfs, SelectionSet, MorfTargetMesh, BlendShape

if(sys.modules["Load_LWO"].DO_RELOAD):
	if(sys.version_info[0] >= 3):
		import importlib
		Fn = importlib.reload
	else:
		Fn = reload
	loc = set(locals().keys())
	for f in (
		"FnCalc",
		"Weights", "Shaders", "PrintMsg", "Defines",
		"BaseMesh", "VColors", "Edges", "Surfs", "Smooth", "SelectionSet", "MorfTargetMesh",
		"BlendShape",
		):
		if(f in loc):
			Fn(eval(f))
	del f, loc

#### ========================================================================
__all__ = ("MakeMeshObject",)

#### =======================================================================
__DEBUG = sys.modules["Load_LWO"].DEBUG

#### =======================================================================
def __Verts_Collection(lyr, opts):
	# 係数,±Z方向適用後の頂点座標値リスト
	fpaVerts = __Make_VertsFPA(lyr.pntNum, lyr.pnts, lyr.piv, opts)
	# ポリ頂点数リストと反時計回り化した頂点順リスト
	(newPolNum, buildedPolIdxs, miaPols, miaOrd) = __Make_PolPntNum_Order_Collection(lyr.polNum, lyr.pols)
	return(newPolNum, buildedPolIdxs, fpaVerts, miaPols, miaOrd)

#### --------------------------------
def __Make_VertsFPA(pntNum, pnts, pivot, opts):
	u'''頂点座標値リストを作成. LW/MayaのZ方向を吸収. result: MFloatPointArray'''
#	2018はfpa[i].x = 1, fpa[i][0] = 1共に有効.
#	2020はどちらも無効. docのサンプルも通らない.
#	fpa[i] = [,,,]は可能. 但しfpa[i] = floatPoint(,,,)による置き換え相当.
#	xyz値のリストを直接与えるのがベスト?
	if(opts.usePivot):
		offs = FnCalc.VMul(pivot, opts.vScl)
		fpa = om2.MFloatPointArray([ FnCalc.VMulSub(p, opts.vScl, offs) for p in pnts ])
	else:
		fpa = om2.MFloatPointArray([ FnCalc.VMul(p, opts.vScl) for p in pnts ])

	if(__DEBUG):
		print(u"vert: {}".format(pntNum))
	return(fpa)

#### --------------------------------
def __Make_PolPntNum_Order(polNum, pols):
	arPntNums = []			# ポリ構成頂点数の配列
	arPolPntIdxs = []		# 使用頂点Idxの配列
	pnAPD = arPntNums.append
	ppEXT = arPolPntIdxs.extend

	buildedPolIdxs = [None] * polNum
	polCnt = 0
	for (i, pntIdxs) in enumerate(pols):
		n = len(pntIdxs)
		if(n > 2):
			pnAPD(n)
			ppEXT(pntIdxs[::-1])	# 左右逆回りのIdx
			buildedPolIdxs[i] = polCnt
			polCnt += 1
	return(polCnt, buildedPolIdxs, arPntNums, arPolPntIdxs)

#### ----------------
def __Make_PolPntNum_Order_Collection(polNum, pols):
	'''perポリ頂点数/順リスト(MIntArray).
polPntNums[pntNum,pntNum,...]. order[pntIdx,pntIdx,...]
各ポリの頂点数の配列(MIntArray),
各頂点idx配列を全て連結した(MIntArray)'''
	(polCnt, buildedPolIdxs, arPntNums, arPolPntIdxs) = __Make_PolPntNum_Order(polNum, pols)

	if(__DEBUG):
		print(u"pol: {}\norder: {}".format(len(pols), len(arPolPntIdxs)))

	iaPntNums = om2.MIntArray(arPntNums)		# ポリ構成頂点数の配列
	iaPolPntIdxs = om2.MIntArray(arPolPntIdxs)	# 使用頂点Idxの配列
	return(polCnt, tuple(buildedPolIdxs), iaPntNums, iaPolPntIdxs)

#### ========================================================================
#### ========================================================================
def Set_UVSet(moMesh, newPolNum, buildedPolIdxs, lyr):
	u'''UVセット作成とカレントをセット'''
	(hasUV, mapNames) = __Make_UVSet(moMesh, lyr.uvs)
	if(hasUV):	# UVの適用(vmap&vmadの都合で適用漏れポリゴンが発生するかも.要検証)
		for name in lyr.uvs.keys():
			uvs = lyr.uvs[name]	# 旧名uvmap

			(uvDct, mfaU, mfaV) = __UV_Dict_Array(uvs)

			pntUVIdxDct = __UV_PntUV_ForVmapDict(uvDct, uvs.vmap)

			vmad = __UV_DictAsVMad(lyr.pols, uvs, uvDct, mfaU, mfaV, pntUVIdxDct)

			newMapName = mapNames[name]	# 旧名→真名uvmap

			FnMe = om2.MFnMesh(moMesh)
			FnMe.setUVs(mfaU, mfaV, newMapName)

			__AssignUV(FnMe, lyr.polNum, buildedPolIdxs, vmad, newMapName)

#### ------------------------------------------------------------------------
def __Make_UVSet(moMesh, uvs):
	'''UVセットのみを作成'''
	uvNames = uvs.keys()	# .[name][vmap/vmad]...
	if(len(uvNames) == 0):
		return(False, None)

	reg = re.compile(r"[^a-zA-Z0-9_]")

	FnMe = om2.MFnMesh(moMesh)
	mapNames = {}	# {旧名:真名}
	for name in uvNames:
		if(name == Defines.DEFAULT_UV):
			mapNames.update({ Defines.DEFAULT_UV: Defines.DEFAULT_UV })
			# 重複作成はしない.値を設置するまでゴースト状態？
		else:
			newName = FnMe.createUVSet(reg.sub("_", name))
			mapNames.update({ name: newName })

	if(Defines.DEFAULT_UV in mapNames):	# map1があるなら優先してカレントにする
		FnMe.setCurrentUVSetName(Defines.DEFAULT_UV)
	else:	# 旧名[0]の真名を辞書から拾う
		FnMe.setCurrentUVSetName(mapNames[uvNames[0]])
		# 新規作成に際してmap1が存在するが消せない

	return(True, mapNames)

#### ------------------------------------------------------------------------
def __UV_Dict_Array(uvs):
	u'''uv値からuvList内のindexを得る為の辞書{ (u,v):totalIdx }. u/v個々のlist.'''
	def __make_dct_array(totalIdx, uvDct, uAPD, vAPD, pntDct):
		for (_, uv) in pntDct.items():		# {pntIdx:(u,v)},{},...
			if(uv not in uvDct):
				uvDct[uv] = totalIdx		# { (u,v): totalIdx }
				uAPD(uv[0])
				vAPD(uv[1])
				totalIdx += 1
		return(totalIdx)

	#### ----------------
	uvDct = {}
	fu = []
	fv = []
	uAPD = fu.append
	vAPD = fv.append
	totalIdx = 0

	# lwoはvmap優先でvmapは残る. lxoはvmadのみとしてvmap無しもあり得る.
	if(uvs.vmap):
		totalIdx = __make_dct_array(0, uvDct, uAPD, vAPD, uvs.vmap)	# vmap{pntIdx:(u,v)},{},...

	for (polIdx, pntDct) in uvs.vmad.items():	# vmad{polIdx:{pntIdx:(u,v)}},{},...
		totalIdx = __make_dct_array(totalIdx, uvDct, uAPD, vAPD, pntDct)

	fau = om2.MFloatArray(fu)
	fav = om2.MFloatArray(fv)
	return(uvDct, fau, fav)

#### ------------------------------------------------------------------------
def __UV_PntUV_ForVmapDict(uvDct, vmap):
	if(not vmap):
		return({})

	# vmap{pntIdx:(u,v)},{},...
	return( dict([ (pntIdx, uvDct[uv]) for (pntIdx, uv) in vmap.items() ]) )
#	pntUVIdxDct = {}
#	UPD = pntUVIdxDct.update
#	for (pntIdx, uv) in vmap.items():	# vmap{pntIdx:(u,v)},{},...
#		UPD({ pntIdx: uvDct[uv] })
#	return(pntUVIdxDct)

#### ------------------------------------------------------------------------
def __Hoge_VMap(polIdx, pntIdxs, pntUVIdxDct, vmad):
	polPnt = [ pntUVIdxDct[pntIdx] for pntIdx in reversed(pntIdxs) if(pntIdx in pntUVIdxDct) ]
	if(len(pntIdxs) == len(polPnt)):		# 全点がuv/vmapを持つ
		 vmad.update({ polIdx: polPnt })	# { 旧polIdx:[uvIdx,...] }
#### -----------------------------------
def __Hoge_VMad(uvDct, polIdx, pntIdxs, uvs, pntUVIdxDct, vmad):
	polPnt = []
	ppnAPD = polPnt.append

	for pntIdx in reversed(pntIdxs):
		if(pntIdx in uvs.vmad[polIdx]):
			ppnAPD(uvDct[uvs.vmad[polIdx][pntIdx]])
		else:
			if(pntIdx in pntUVIdxDct):
				ppnAPD(pntUVIdxDct[pntIdx])

	if(len(pntIdxs) == len(polPnt)):		# 全点がuv/vmapを持つ
		vmad.update({ polIdx: polPnt })		# { 旧polIdx:[uvIdx,...] ]

#### -----------------------------------
def __UV_DictAsVMad(pols, uvs, uvDct, mfaU, mfaV, pntUVIdxDct):
	'''u/vの配列分離.polIdxから指す{polIdx:{pntIdx:u/v idx}...}の作成'''
	vmad = {}
	polNum = len(pols)
	if(uvs.vmad):	# 不連続UV有り
		for (polIdx, pntIdxs) in enumerate(pols):
			if(len(pntIdxs) > 2):
				if(polIdx in uvs.vmad):			# polyがVMADを持つなら
					__Hoge_VMad(uvDct, polIdx, pntIdxs, uvs, pntUVIdxDct, vmad)
				else:							# polyがVMAPのみ
					__Hoge_VMap(polIdx, pntIdxs, pntUVIdxDct, vmad)
	else:			# 連続UVのみ
		for (polIdx, pntIdxs) in enumerate(pols):
			if(len(pntIdxs) > 2):
				__Hoge_VMap(polIdx, pntIdxs, pntUVIdxDct, vmad)
	return(vmad)

#### ------------------------------------------------------------------------
def __AssignUV(FnMe, polNum, buildedPolIdxs, vmad, name):
	polPntNums = []
	uvIdxs = []
	ppnAPD = polPntNums.append
	uviEXT = uvIdxs.extend
	for i in buildedPolIdxs:
		if(i is not None):
			if(i in vmad):
				ppnAPD(len(vmad[i]))
				uviEXT(vmad[i])
			else:
				ppnAPD(0)
	# assignUVsはperPolUVとして全てのポリに作用する為、
	# perPolUVが成立しないポリ[Idx]部は無効フラグ"0"で埋める必要がある

	miaPolPntNums = om2.MIntArray(polPntNums)
	miaUvIdxs = om2.MIntArray(uvIdxs)
	FnMe.assignUVs(miaPolPntNums, miaUvIdxs, name)

#### ========================================================================
def __SetAttrToBaseMesh_withUpdateSurf(moMesh, hasVertColor, opts):
	FnMe = om2.MFnMesh(moMesh)
	# カラーチャネル dcc/displayColorChannel
	# "None", "Ambient", "Ambient+Diffuse", "Diffuse", "Specular", "Emission"
	FnMe.findPlug("dcc", False).setString(Defines.gMeshDispColrChans[3])
	# 両面 doubleSided
	FnMe.findPlug("ds", False).setBool(opts.isDblSide)
	# .dcol/displayColors
	if(hasVertColor is not None):
		FnMe.displayColors = hasVertColor

	FnMe.updateSurface()

#### ========================================================================
def MakeMeshObject(obj, lyr, opts):

	PrintMsg.BuildGeometry(lyr, opts)

	if((not lyr.pntNum) or (not lyr.polNum)):
		return(None)

	(newPolNum, buildedPolIdxs, fpaVerts, miaPols, miaOrd) = __Verts_Collection(lyr, opts)
	(dagpMesh, moMesh) = BaseMesh.Make(fpaVerts, miaPols, miaOrd, lyr, opts)

	# UVセット作成とカレントをセット
	Set_UVSet(moMesh, newPolNum, buildedPolIdxs, lyr)

	# エッジウェイト適用
	if(opts.useEdge):
		Edges.ApplyCreaseEdge(moMesh, lyr)

	# perPolノーマル適用

	# 頂点カラー
	(rgbNames, rgbaNames, hasVertColor) = ([], [], None)
	if(opts.useVCol):
		# 頂点カラー適用
		rgbNames  = VColors.ApplyVColors(moMesh, buildedPolIdxs, lyr.pols, lyr.rgb,  opts.useMiVC)
		rgbaNames = VColors.ApplyVColors(moMesh, buildedPolIdxs, lyr.pols, lyr.rgba, opts.useMiVC)
		# 頂点カラー表示
		hasVertColor = bool(rgbNames or rgbaNames)

	# サーフェス(SG)適用
	mtlDct = Surfs.Make(obj.tags, lyr.surfTags, obj.surfs, buildedPolIdxs, dagpMesh, opts)
	# マテリアルの設定
	Shaders.SetSurfaceAttr(obj.surfs, obj.clips, mtlDct, rgbNames, rgbaNames, opts)

	# セレクションセット作成
	if(opts.usePick):
		SelectionSet.Make(dagpMesh, lyr.name, lyr.picks, Defines.SetFMT.SELS, None, True)
	if(opts.useColr):
		SelectionSet.Make(dagpMesh, lyr.name, lyr.tagColors, Defines.SetFMT.COLR, buildedPolIdxs, False)
	if(opts.usePart):
		SelectionSet.Make(dagpMesh, lyr.name, lyr.tagParts, Defines.SetFMT.PART, buildedPolIdxs, False)
	if(opts.usePPic and (obj.type=="LXO")):
		SelectionSet.Make(dagpMesh, lyr.name, lyr.tagSelects, Defines.SetFMT.PICK, buildedPolIdxs, False)

	# レイヤがPoly/Skelegon/Curve混在の場合、頂点が未使用頂点としてmesh内に残る.
	# 削除によるindexズレ(uv等)対策をするのではなく、ジオメトリ構築後に削除する.
	# [ 1点でも削除すれば全ての未使用頂点が消え、indexが詰められる.]
	# MayaUI操作ではエッジの削除等でトポロジ変更を起こせば消える(頂点単位の明示的削除は出来ない).
	if(lyr.bonNum):
		FnMe = om2.MFnMesh(moMesh)
#		FnMe.deleteVertex(lyr.bones[0][BoneT.PNTID][0])	# 1本目の第1頂点のidx.
		FnMe.deleteVertex(lyr.bones[0].pnt0)	# 1本目の第1頂点のidx.

		if(len(lyr.weights) and (opts.useWgt or opts.useXML)):	# ボーンが参照するvmap値を保持/書き出しする場合.
			Weights.MakeWeightMap(dagpMesh.fullPathName(), lyr.bones, lyr.weights, lyr.pnts, lyr.pols, buildedPolIdxs, opts)

	elif(lyr.crvNum):
		FnMe = om2.MFnMesh(moMesh)
		FnMe.deleteVertex(lyr.crvs[0][1])	# 1本目の第1頂点のidx.

	FnMe = om2.MFnMesh(dagpMesh)	# dagpだとFix, Fixすると未使用頂点は消せない

	# モーフ形状作成
	(hasRM, hasAM, relMorfDict, absMorfDict) = MorfTargetMesh.Make(newPolNum, buildedPolIdxs, fpaVerts, miaPols, miaOrd, lyr, opts)

	# ブレンドシェイプ作成
	if(opts.useRM and opts.loadRM and hasRM):
		BlendShape.MakeNode(dagpMesh, relMorfDict)
	if(opts.useAM and opts.loadAM and hasAM):
		BlendShape.MakeNode(dagpMesh, absMorfDict)

	# 法線スムージング
	Smooth.NormalSmooth(dagpMesh.fullPathName(), obj.tags, obj.surfs, lyr.surfTags, buildedPolIdxs, FnMe.numPolygons)

	__SetAttrToBaseMesh_withUpdateSurf(moMesh, hasVertColor, opts)

#	if(opts.mergeUV):
#		Meshes.MergeUVs(dagpMesh)

	dagpMesh.pop()	# mesh -> transform

	return(dagpMesh)
