﻿# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import print_function

import sys, chunk

from . import LXO_ParseHeader, LXO_Object, Vis_Flags, DebugPrints, FnRead, Vis_Flags

if(sys.modules["Load_LWO"].DO_RELOAD):
	if(sys.version_info[0] >= 3):
		import importlib
		Fn = importlib.reload
	else:
		Fn = reload
	loc = set(locals().keys())
	for f in ("LXO_ParseHeader", "LXO_Object", "Vis_Flags", "DebugPrints", "FnRead",):
		if(f in loc):
			Fn(eval(f))
	del f, loc

from .LXO_ParseHeader import *

#### =======================================================================
__all__ = ("ParseLXO",)

#### ========================================================================
class ParseLXO(FnRead.FnReadObject, DebugPrints.DebugPrint):
	__slots__ = FnRead.FnReadObject.__slots__ + DebugPrints.DebugPrint.__slots__ + (
		"VisFlags", "lwo",
	)

	fileType = "LXOB"

	sdsPntWgtName  = "base"
	sdsEdgeWgtName = "Edge Weight"

	if(sys.version_info[0] >= 3):
		uint2ColorInstType = (int,)
	else:
		uint2ColorInstType = (int, long)

	#### -----------------------------------------------------------------------
	def __init__(self, *arg, **kwd):
		super(ParseLXO, self).__init__(*arg, **kwd)

		self.VisFlags = Vis_Flags.VisFlags(**kwd)

		self.lwo = None

		fn = kwd.get("fn", kwd.get("fileName", None))
		if(fn):
			self.Read(fn)
		else:
			self.lwo = LXO_Object.LxoObject()

	#### ========================================================================
	def Get_CHNV_Value(self, iChnv):
		# [(R,v), (G,v), (B,v)]
		return(iChnv.v[0][1], iChnv.v[1][1], iChnv.v[2][1])

	#### ========================================================================
	def __UInt2Color(self, i):
		u'''int/long(0xRRGGBBAA) → flt(r,g,b,a)'''
		#py2ではintは30bit迄. 31bitからはlongになる. py3では常にint. long型は存在しない.
		if(not isinstance(i, self.uint2ColorInstType)):
			raise TypeError(type(i))
		cof = 1.0 / 255.0
		return([ (((i >> c)&0xFF) * cof) for c in (24,16,8,0) ])	# R:24,G:16,B:8,A:0

	#### ========================================================================
	# 紐付くshader/materialを持たないptagもある.
	# Modoの適用ルールは(GUI/階層順を受けて)煩雑なので,
	# Render直下にあるadvMatの中で最大idxの物を使うか, 未割当として扱うのが妥当?
	def __Conv_I2S_dct(self):
		u'''ptagに紐付いてるidを収集. { ptag: [ (matId, mskId),...] }'''
		items = self.lwo.items

		dct = {}
		for it in items:
			if("advancedMaterial" != it.type):
				continue

			# advMatならmaskかpolyRenderを親に持つ.
			iLinks = it.itemLinks.get("parent", None)
			if(iLinks is None):
				continue

			# advMaterialが指す親
			mskItem = items[iLinks[0].id]
			# Renderツリー直下にadvMaterialが存在出来るので最低限の保険
			if("mask" != mskItem.type):
				continue

			# 適正な状態なら親Item(type:mask)はptypとptagを持つはずだけど保険
			if(("ptyp" not in mskItem.stringChans) or ("ptag" not in mskItem.stringChans)):
				continue

			if("Material" == mskItem.stringChans.ptyp.s):
				ptag = mskItem.stringChans.ptag.s
				if(ptag not in dct):
					dct[ptag] = []
				dct[ptag].append( (it.referenceId, mskItem.referenceId) )	# (matId, mskId)
		# end of items
		return(dct)

	#### --------------------------------
	def __Conv_I2S_surf(self, ptag, matItem):
		u'''advMaterialから要素取り出し. surf作成.'''
		surf = LXO_Object.SurfaceObject(n= ptag)

		surf.colr = self.Get_CHNV_Value(matItem.vectorChans.diffCol)

		for (lw, lx) in (("diff","diffAmt"), ("spec","specAmt"), ("lumi","luminousAmt"), ("tran","tranAmt"),
						 ("refl","reflAmt"), ("bump","bumpAmp"), ("side","dblSided"), ("sman","smAngle"), ):
			if(lx in matItem.chanVals):
				surf[lw] = matItem.chanVals[lx].v

		# 法線スムーズのlxo.smAngleは常に＋値(含む0.0)
		# lxo.smoothが0%の時を無効と見做し, lwo準拠の－値にする
		if("smooth" in matItem.chanVals):
			if(matItem.chanVals.smooth.v <= 0.0):
				surf.sman = -surf.sman

		self.lwo.surfs[surf.name] = surf

	#### --------------------------------
	def __MtlTraverse(self, id, APD):
		# [ownIdx,...,rootIdx]取得用
		iLinks = self.lwo.items[id].itemLinks.get("parent", None)
		if(iLinks):
			APD(iLinks[0].idx)
			self.__MtlTraverse(iLinks[0].id, APD)

	#### --------------------------------
	def __Conv_I2S_traverse(self, lst):
		u'''parentのidx値はRenderツリー表示の上(N) > 下(0).
		ptagが重複するtype:maskは表示上で上に位置する(idxが大きい)物が使われている.'''
		ar = []
		for (matId, mskId) in lst:
			tmp = []									# [ownIdx,...,rootIdx]
			self.__MtlTraverse(mskId, tmp.append)
			ar.append( ((matId, mskId), tmp[::-1]) )	# [ ((matId,mskId),[rootIdx,...,ownIdx]), ... ]
		return(ar)

	#### --------------------------------
	def __Conv_I2S_findIds(self, ar):
		u'''疑似ツリー辞書を辿ってGUI内の優先Matを拾う'''
		def __FollowDict(dct):
			# keyIdxが最大の物を辿っていく
			(d, ids) = dct[(sorted(dct.keys())[-1])]
			if(ids is None):
				return(__FollowDict(d))
			return(ids)
		#### ----------------
		# ar[ ((matId,mskId),[rootIdx,...,ownIdx]), ... ]
		# 一連の { idx:[{}, x] } を作り, 終端のxを(matId,mskId)にする
		treeDct = self.__Tree_Dct(ar)

#>>
#>>		self.DebugPrint_ItemTree_Lxo(treeDct, 0)
		return(__FollowDict(treeDct[0]))

	#### --------------------------------
	def __Convert_ItemToSurface(self, lwo):
		u'''子matelItem→親mask(ptag)Item→'''
		# 事前収集
		preDct = self.__Conv_I2S_dct()	# { ptag: [ (matId, mskId), ...]

		# surf作成
		for (ptag, idsLst) in preDct.items():
			if(len(idsLst) == 1):
				# ptagに紐付くtype:maskが1つ.
				(matId, _) = idsLst[0]
			else:
				# 同じptagのtype:maskが複数.
				ar = self.__Conv_I2S_traverse(idsLst)
				(matId, _) = self.__Conv_I2S_findIds(ar)

			self.__Conv_I2S_surf(ptag, lwo.items[matId])
		# end of items

	#### ========================================================================
	def __Relation_LayerAndMeshItem(self, lwo):
		u'''LAYRと紐付くITEMの固有名をLAYRに取り出す'''
		items = lwo.items
		for lyr in lwo.layers:
			it = items[lyr.itemRef]						# lyr.itemRefでitemを特定
			lyr.uniqueIdentifier = it.uniqueIdentifier	# meshes辞書から名前で参照可能にする

	#### ========================================================================
	def __Tree_Dct(self, ar):
		u'''ITEMの[idx]から疑似ツリー辞書を作る. material, transform'''
		obj = [{}, None]
		for (uniq, indices) in ar:
			ob = obj
			curDct = obj[0]
			for idx in indices:
				if(idx not in curDct):
					ob = [{}, None]
					curDct.update( ((idx, ob),) )
				else:
					ob = curDct[idx]
				curDct = ob[0]
			ob[1] = uniq
		return(obj)

	#### ========================================================================
	def __Collect_Parent(self, items, ownerItem):
		u'''親の固有名と自身の階層内位置idx'''
		ownerParentLinks = ownerItem.itemLinks.get("parent", [])

		for iLinks in ownerParentLinks:
			if(iLinks.id >= 0):
				return(items[iLinks.id].uniqueIdentifier, iLinks.idx)
		return(None, -1)

	#### --------------------------------
	def __ItemTraverse(self, id, APD):
		# [ownIdx,...,rootIdx]取得
		parentLinkLst = self.lwo.items[id].itemLinks.get("parent", None)
		if(parentLinkLst):
			itemId = parentLinkLst[0].id
			if(itemId >= 0):
				APD(parentLinkLst[0].idx)
				self.__ItemTraverse(itemId, APD)
			else:
				# root - (transform) - mesh/group/locator時はparent.id=-1がある?
				APD(self.lwo.items[id].lyrIdx)
		else:
			APD(self.lwo.items[id].lyrIdx)

	#### --------------------------------
	def __ItemTree_traverse(self, objDct):
		u'''ITEM(group/locator/mesh)毎の親を辿り, 階層毎のidxを取る.'''
		ar = []
		for (uniq, obj) in objDct.items():
			tmp = []								# [ownIdx,...,rootIdx]
			self.__ItemTraverse(obj.ownItem.referenceId, tmp.append)
			obj.treeParentsList = tmp[::-1]			# 一時使用. invert [rootIdx,...,ownIdx]
			ar.append( (uniq, obj.treeParentsList) )
		return(ar)

	#### --------------------------------
	def __ItemTree_Dct(self, lwo):
		u'''ITEM(group/locator/mesh)毎の[idx]から疑似ツリー辞書を作る.'''
		ar = self.__ItemTree_traverse(lwo.transforms)
		ar.extend(self.__ItemTree_traverse(lwo.locators))
		ar.extend(self.__ItemTree_traverse(lwo.meshes))

		treeDct = self.__Tree_Dct(ar)
#>>
#>>		self.DebugPrint_ItemTree_Lxo(treeDct, 0)
		return(treeDct)

	#### ========================================================================
	def __Collect_ItemParentList(self, lwo):
		u'''ITEM毎の親子階層のlist作成'''
		transformTreeDct = self.__ItemTree_Dct(lwo)

		refDct = lwo.transforms.copy()
		refDct.update(lwo.locators)
		refDct.update(lwo.meshes)

		for (uniq, obj) in refDct.items():
			ar = []								# [rootUniq, ..., ownUniq]
			curObj = transformTreeDct
			curDct = curObj[0]
			for idx in obj.treeParentsList:
				curObj = curDct[idx]
				curDct = curObj[0]
				ar.append(curObj[1])

			obj.treeParentsList = ar

		if(0):
			print(">>>")
			for (uniq, obj) in refDct.items():
				print("{}: {}:".format(uniq, obj.name), obj.treeParentsList)

	#### ========================================================================
	def __Collect_Transforms(self, lwo):
		u'''ITEMからmesh, meshInst, locator, groupLocatorを取り出し'''
		# 当座はcamera, ANYlight, textureLocatorは無視
		# t/r/sがあるとは限らない.
		def __CT_GetSomeParms(items, item):
			(parentUniqueIdentifier, parentOrder) = self.__Collect_Parent(items, item)

			wireColor = None
			if("wireColor" in item.vectorChans):
				# 色が有っても表示指定がuserでなければ無効. default/user/bgLight/bgDark
				if("wireOptions" in item.chanVals):
					if("user" == item.chanVals.wireOptions.v):
						wireColor = self.Get_CHNV_Value(item.vectorChans.wireColor) + (1.0,)	# alphaとして
						# tuple. ※明示的指定の有無判定に使う.
			else:
				if(item.lyrRGBA is not None):
					wireColor = self.__UInt2Color(item.lyrRGBA)
					# list. ※明示的指定の有無判定に使う.

			return(parentUniqueIdentifier, parentOrder, wireColor)
		#### groupLocator --------------------------------
		def __CT_MakeTrs(lwo, item, uniqueName):
			if(uniqueName not in lwo.transforms):
				(parentUniqueIdentifier, parentOrder, wireColor) = __CT_GetSomeParms(lwo.items, item)

				lwo.transforms[uniqueName] = LXO_Object.TransformObject(oi= item, n= item.name, v= (item.lyrMask&0x1),
					uui= item.userUniqueIndex, pui= parentUniqueIdentifier, poi= parentOrder, wc= wireColor,
					)
			return(lwo.transforms[uniqueName])
		#### locator ----------------
		def __CT_MakeLoc(lwo, item, uniqueName):
			if(uniqueName not in lwo.locators):
				(parentUniqueIdentifier, parentOrder, wireColor) = __CT_GetSomeParms(lwo.items, item)

				lwo.locators[uniqueName] = LXO_Object.LocatorObject(oi= item, n= item.name, v= (item.lyrMask&0x1),
					uui= item.userUniqueIndex, pui= parentUniqueIdentifier, poi= parentOrder, wc= wireColor,
					ls= item.chanVals.size.v,
					)
			return(lwo.locators[uniqueName])
		#### mesh, meshInst ----------------
		def __CT_MakeMesh(lwo, item, uniqueName):
			if(uniqueName not in lwo.meshes):
				(parentUniqueIdentifier, parentOrder, wireColor) = __CT_GetSomeParms(lwo.items, item)

				lwo.meshes[uniqueName] = LXO_Object.MeshObject(oi= item, n= item.name, v= (item.lyrMask&0x1),
					uui= item.userUniqueIndex, pui= parentUniqueIdentifier, poi= parentOrder, wc= wireColor,
					)
			return(lwo.meshes[uniqueName])
		#### --------------------------------
		def __MakeTransformObj(lwo, item):
			if("groupLocator" == item.type):
				return(__CT_MakeTrs(lwo, item, item.uniqueIdentifier))
			elif("locator" == item.type):
				return(__CT_MakeLoc(lwo, item, item.uniqueIdentifier))
			elif(item.type in {"mesh", "meshInst"}):
				return(__CT_MakeMesh(lwo, item, item.uniqueIdentifier))
			else:
				pass
			return(None)
		#### --------------------------------

		items = lwo.items

		for it in items:
			# 自体が対象タイプ
			if(it.type in {"groupLocator", "locator", "mesh", "meshInst"}):
				obj = __MakeTransformObj(lwo, it)

			# 自体はT/R/S
			elif(it.type in {"translation", "rotation", "scale"}):
				iLinks = it.itemLinks.get("xfrmLocal", [])

				for iLink in it.itemLinks.get("xfrmCore", []):
					if(iLink.id < 0):		# (xfrmCore, id=-1, idx-2)が存在する. 用途不明.
						continue

					obj = __MakeTransformObj(lwo, items[iLink.id])
					if(obj is None):
						continue

					if("type" not in it.chanVals):	# インスタンスには無い？
						thisType = "core"
					else:
						thisType = it.chanVals.type.v

					if("translation" == it.type):
						pos = self.Get_CHNV_Value(it.vectorChans.pos)
						if("core" == thisType):
							obj.translate = pos
						elif("piv" == thisType):
							obj.pivot = pos
						elif("pivc" == thisType):
							pass

					elif("rotation" == it.type):
						rot = self.Get_CHNV_Value(it.vectorChans.rot)
						# 個別指定が無い場合はt/r/s所有者がorderを持っている
						# t/r/s個々は, source有り&order無し, source無し&order有りのパターン
						if("source" in it.itemLinks):
							ro = items[it.itemLinks.source[0].id].chanVals.order.v
						else:
							ro = it.chanVals.order.v

						if("core" == thisType):
							obj.rotate = rot
							obj.rotateOrder = ro
						elif("sys" == thisType):	# PivotRotation
							obj.pivotRotate = rot
							obj.pivotRotateOrder = ro
						elif("7" == thisType):		# PivotRotationCompensation
							pass

					elif("scale" == it.type):
						# scale pivotは無い?
						obj.scale = self.Get_CHNV_Value(it.vectorChans.scl)
					# end of chanVal.type
				# end of itemLinks
			# end of it.type
		# end of items

	#### ========================================================================
	def Read(self, fileName, **kwd):
		u'''ファイルロード. インターフェースの共通化'''
		if(not self._Do_Read(fileName)):
			return(self.lwo)

		self.__Convert_ItemToSurface(self.lwo)

		self.__Collect_Transforms(self.lwo)

		self.__Relation_LayerAndMeshItem(self.lwo)

		self.__Collect_ItemParentList(self.lwo)

		if(0):
			print("\n=================================================================\n")
			def __XXX(objs):
				for (uniq, obj) in objs.items():
					print("#### {}".format(uniq))
					for k in obj.__slots__:
						print("  {}: {}".format(k, obj[k]))
			__XXX(self.lwo.meshes)
			print("----------------------")
			__XXX(self.lwo.transforms)
			print("----------------------")
			__XXX(self.lwo.locators)
			print("----------------------")

		return(self.lwo)

	#### ========================================================================
	def ScanLayer(self, fp, *arg):
		u'''簡易ファイル走査'''
		layers = []

		while(True):
			try:
				curChunk = chunk.Chunk(fp)	# FORM:size:LXOBの子階層
			except EOFError:
				break

			size = curChunk.chunksize

			chunkId = self.FnUnpckLWID(curChunk.chunkname)[0]

			if(chunkId == ChunkIdT.VRSN):
				curChunk.skip()

			elif(chunkId == ChunkIdT.LAYR):
				self.Read_Layr(curChunk.read(size), size, layers)

			elif(chunkId == ChunkIdT.VMAP):
				(lwid, size) = self.Read_LWID(curChunk)

				if(  lwid == VMapIdT.MORF):	# Rel Morph
					self.Scan_Morph_VMap(curChunk.read(size), layers[-1].relMorphs)
				elif(lwid == VMapIdT.SPOT):	# Abs Morph
					self.Scan_Morph_VMap(curChunk.read(size), layers[-1].absMorphs)
				elif(lwid == VMapIdT.RGB_):
					self.Scan_RGB_VMap(curChunk.read(size), layers[-1])
				elif(lwid == VMapIdT.RGBA):
					self.Scan_RGBA_VMap(curChunk.read(size), layers[-1])
##				elif(lwid == VMapIdT.NORM):
#				elif(lwid == VMapIdT.PICK):
#				elif(lwid == VMapIdT.TXUV):
#				elif(lwid == VMapIdT.WGHT):
#				elif(lwid == VMapIdT.TBAS):
				else:
					curChunk.skip()

			elif(chunkId == ChunkIdT.VMAD):
				(lwid, size) = self.Read_LWID(curChunk)

				if(  lwid == VMapIdT.RGB_):
					self.Scan_RGB_VMad(curChunk.read(size), layers[-1])
				elif(lwid == VMapIdT.RGBA):
					self.Scan_RGBA_VMad(curChunk.read(size), layers[-1])
				elif(lwid == VMapIdT.WGHT):
					self.Scan_Weight_VMad(curChunk.read(size), layers[-1])
				elif(lwid == VMapIdT.NORM):
					self.Scan_NORM(curChunk.read(size), layers[-1])
##				elif(lwid == VMapIdT.MORF):
##				elif(lwid == VMapIdT.SPOT):
#				elif(lwid == VMapIdT.PICK):
#				elif(lwid == VMapIdT.TXUV):
#				elif(lwid == VMapIdT.TBAS):
				else:
					curChunk.skip()

			elif(chunkId == ChunkIdT.VMED):
				(lwid, size) = self.Read_LWID(curChunk)

				if(lwid == VMedIdT.SUBV):
					self.Scan_SUBV_VMed(curChunk.read(size), layers[-1])
#				elif(lwid == VMedIdT.EPCK):
				else:
					curChunk.skip()

			elif(chunkId == ChunkIdT.POLS):
				(lwid, size) = self.Read_LWID(curChunk)

				if(lwid in {PolyIdT.HCRV, PolyIdT.BCRV, PolyIdT.CURV}):	# CURV:701で廃止?
					self.Scan_Crvs(curChunk.read(size), layers[-1])
#				elif(lwid == PolyIdT.BEZR):	# BEZR:701で廃止?
#				elif(lwid == PolyIdT.BSPL):

#				elif(lwid == PolyIdT.LINE):	# 801:linePol
#				elif(lwid == PolyIdT.OPNT):	# 13.1:pointPol

				else:
					curChunk.skip()

			else:
				curChunk.skip()
		# end of while
		return(layers)

	#### ========================================================================
	def _Read(self, fp):
		curPolIdxOffs = 0
		isExclude = False

		self.lwo = LXO_Object.LxoObject()
		lwo = self.lwo
		lwo.fileName = fp.name

		while(True):
			try:
				curChunk = chunk.Chunk(fp)	# FORM:size:LXOBの子階層
			except EOFError:
				break

			size = curChunk.chunksize

			chunkId = self.FnUnpckLWID(curChunk.chunkname)[0]

			if(chunkId == ChunkIdT.VRSN):
				self.Read_VersionData(curChunk.read(size), size)

			elif(chunkId == ChunkIdT.ENCO):
				self.Read_Encoding(curChunk.read(size), size)

			elif(chunkId == ChunkIdT.TAGS):
				lwo.tags = self.Read_Tags(curChunk.read(size), size)

			elif(chunkId == ChunkIdT.CHNM):
				lwo.channelNames = self.Read_ChannelNames(curChunk.read(size), size)

			elif(chunkId == ChunkIdT.LAYR):
				curPolIdxOffs = 0
				isExclude = False
				self.Read_Layr(curChunk.read(size), size, lwo.layers)

			elif(chunkId == ChunkIdT.PNTS):	# [LAYR]:[PNTS]:...
				self.Read_Pnts(curChunk.read(size), size, lwo.layers[-1])

#			elif(chunkId == ChunkIdT.VMPA):

			elif(chunkId == ChunkIdT.VMAP):	# [LAYR]:[VMPA][VMAP][LWID]:[VMPA][VMAP][LWID]...
				(lwid, size) = self.Read_LWID(curChunk)

				if(  lwid == VMapIdT.WGHT):
					self.Read_Weight_VMap(curChunk.read(size), size, lwo.layers[-1].weights)
				elif(lwid == VMapIdT.MORF):	# Rel Morph
					self.Read_Morph_VMap(curChunk.read(size), size, lwo.layers[-1].relMorphs)
				elif(lwid == VMapIdT.SPOT):	# Abs Morph
					self.Read_Morph_VMap(curChunk.read(size), size, lwo.layers[-1].absMorphs)
				elif(lwid == VMapIdT.TXUV):
					self.Read_UV_VMap(curChunk.read(size), size, lwo.layers[-1])
				elif(lwid == VMapIdT.RGB_):
					self.Read_RGB_VMap(curChunk.read(size), size, lwo.layers[-1])
				elif(lwid == VMapIdT.RGBA):
					self.Read_RGBA_VMap(curChunk.read(size), size, lwo.layers[-1])
				elif(lwid == VMapIdT.PICK):	# Point Selection
					self.Read_Pick_VMap(curChunk.read(size), size, lwo.layers[-1].picks)
#				elif(lwid == VMapIdT.NORM):
#				elif(lwid == VMapIdT.TBAS):
#				elif(lwid == VMapIdT.EPCK):
				else:
					curChunk.skip()

			elif(chunkId == ChunkIdT.VMAD):	# [LAYR]:[POLS][...][VMPA][VMAD][LWID]:[POLS][...][VMPA][VMAD][LWID]:...
				(lwid, size) = self.Read_LWID(curChunk)

				# VMADのpolIndexはPOLS内でのIndex.
				if(  lwid == VMapIdT.TXUV):
					self.Read_UV_VMad(curChunk.read(size), size, lwo.layers[-1], curPolIdxOffs)
				elif(lwid == VMapIdT.RGB_):
					self.Read_RGB_VMad(curChunk.read(size), size, lwo.layers[-1], curPolIdxOffs)
				elif(lwid == VMapIdT.RGBA):
					self.Read_RGBA_VMad(curChunk.read(size), size, lwo.layers[-1], curPolIdxOffs)
				elif(lwid == VMapIdT.WGHT):
					self.Read_Weight_VMad(curChunk.read(size), size, lwo.layers[-1], curPolIdxOffs)
				elif(lwid == VMapIdT.NORM):
					self.Read_Normal_VMad(curChunk.read(size), size, lwo.layers[-1], curPolIdxOffs)
#				elif(lwid == VMapIdT.MORF):
#				elif(lwid == VMapIdT.SPOT):
#				elif(lwid == VMapIdT.PICK):
#				elif(lwid == VMapIdT.TBAS):
				else:
					curChunk.skip()

			elif(chunkId == ChunkIdT.VMED):	# [LAYR]:[VMED][LWID]:...
				(lwid, size) = self.Read_LWID(curChunk)

				if(  lwid == VMedIdT.SUBV):
					self.Read_SUBV_VMed(curChunk.read(size), size, lwo.layers[-1])
				elif(lwid == VMedIdT.EPCK):
					self.Read_EPCK_VMed(curChunk.read(size), size, lwo.layers[-1])
				else:
					curChunk.skip()

			elif(chunkId == ChunkIdT.POLS):	# [LAYR]:[POLS][LWID]:[POLS][LWID]:...
				(lwid, size) = self.Read_LWID(curChunk)

				isExclude = True

				if(lwid in { PolyIdT.FACE, PolyIdT.PSUB, PolyIdT.SUBD }):
					isExclude = False
					curPolIdxOffs = self.Read_Pols(curChunk.read(size), size, lwo.layers[-1], lwid)

				elif(lwid == PolyIdT.HCRV):
					isExclude = True
					self.Read_HCrvs(curChunk.read(size), size, lwo.layers[-1])
				elif(lwid == PolyIdT.BCRV):
					isExclude = True
					self.Read_BCrvs(curChunk.read(size), size, lwo.layers[-1])
#				elif(lwid == PolyIdT.BEZR):	# BEZR:701で廃止?
				elif(lwid == PolyIdT.CURV):	# CURV:701で廃止?
					isExclude = True
					self.Read_Crvs(curChunk.read(size), size, lwo.layers[-1])
#				elif(lwid == PolyIdT.BSPL):

#				elif(lwid == PolyIdT.LINE):	# 801:linePol
#				elif(lwid == PolyIdT.OPNT):	# 13.1:pointPol

#				elif(lwid == PolyIdT.SPCH):
#				elif(lwid == PolyIdT.TEXT):
				else:
					curChunk.skip()

			elif(chunkId == ChunkIdT.PTAG):	# [LAYR]:[PTAG]size[LWID]size:[PTAG]size[LWID]size:...
				(ptag, size) = self.Read_LWID(curChunk)

				if(ptag == PTagIdT.MATR):
					if(isExclude):
						isExclude = False
						curChunk.skip()
					else:
						self.Read_PTag_Surf(curChunk.read(size), size, lwo.layers[-1], curPolIdxOffs)

				elif(ptag == PTagIdT.COLR):
					if(isExclude):
						curChunk.skip()
					else:
						self.Read_PTag(curChunk.read(size), size, lwo.layers[-1], lwo.tags, curPolIdxOffs, "tagColors", "COLR")

				elif(ptag == PTagIdT.PART):
					if(isExclude):
						curChunk.skip()
					else:
						self.Read_PTag(curChunk.read(size), size, lwo.layers[-1], lwo.tags, curPolIdxOffs, "tagParts", "PART")

				elif(ptag == PTagIdT.PICK):
					if(isExclude):
						curChunk.skip()
					else:
						self.Read_PTag(curChunk.read(size), size, lwo.layers[-1], lwo.tags, curPolIdxOffs, "tagSelects", "PICK")

#				elif(ptag == PTagIdT.FONT):
#				elif(ptag == PTagIdT.JUST):
#				elif(ptag == PTagIdT.TEXT):
#				elif(ptag == PTagIdT.SMGP):
				else:
					curChunk.skip()

#			elif(chunkId == ChunkIdT.BBOX):

			elif(chunkId == ChunkIdT.ITEM):
				self.Read_Item(curChunk.read(size), size, lwo.items)

			elif(chunkId == ChunkIdT.ACTN):
				self.Read_Actn(curChunk.read(size), size, lwo.actions)

			else:
				curChunk.skip()

	#### ========================================================================
	#### ACTN
	#### ========================================================================
	def Read_Actn_ITEM(self, it, data, offs):
		it.referenceId = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]

	#### ----------------
	def Read_Actn_CHAN(self, it, data, offs):
		(CHNM_idx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
		offs += vx

		typ = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
		offs += RSize.U2

		(ENVL_idx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
		offs += vx

		if(typ in ITEM_CHAN_Type.INTS):
			v = self.FnUnpckI4(data[offs:(offs + RSize.I4)])[0]
		elif(typ in ITEM_CHAN_Type.FLTS):
			v = self.FnUnpckF4(data[offs:(offs + RSize.F4)])[0]
		elif(typ in ITEM_CHAN_Type.STRS):
			(v, nameLen) = self.Read_nameString(data[offs:])
		else:
			raise TypeError("CHAN:CHNM_idx:{} is unknown typ {}".format(CHNM_idx, typ))

		chnm = self.lwo.channelNames[CHNM_idx]
		it.chanVals[chnm] = LXO_Object.aCHAN(CHNM_idx, typ, ENVL_idx, v, chnm)

	#### ----------------
	def Read_Actn_GRAD(self, it, data, offs):
		(CHNM_idx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
		offs += vx

		(ENVL_idx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
		offs += vx

		interp = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]

		chnm = self.lwo.channelNames[CHNM_idx]
		it.gradientChans[chnm] = LXO_Object.aGRAD(CHNM_idx, ENVL_idx, interp, chnm)

	#### ----------------
	def Read_Actn_CHNS(self, it, data, offs):
		(name, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen

		(CHNM_idx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
		offs += vx

		(v, nameLen) = self.Read_nameString(data[offs:])

		chnm = self.lwo.channelNames[CHNM_idx]
		it.stringChans[chnm] = LXO_Object.aCHNS(name, CHNM_idx, v, chnm)

	#### ----------------
	def Read_Actn_CHNN(self, it, data, offs):
		(name, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen

		typ = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
		offs += RSize.U2

		(ENVL_idx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
		offs += vx

		if(typ in ITEM_CHAN_Type.INTS):
			v = self.FnUnpckI4(data[offs:(offs + RSize.I4)])[0]
		elif(typ in ITEM_CHANL_Type.FLTS):
			v = self.FnUnpckF4(data[offs:(offs + RSize.F4)])[0]
		elif(typ in ITEM_CHAN_Type.STRS):
			(v, nameLen) = self.Read_nameString(data[offs:])
		else:
			raise TypeError("CHAN:CHNN:{} is unknown typ {}".format(name, typ))

		it.namedChans.append( LXO_Object.aCHNS(name, typ, ENVL_idx, v) )

	#### ----------------
	def Read_Actn_PRNT(self, it, data, offs):
		it.parentId = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]

	#### ----------------
	def Read_Actn_BasePart(self, it, data):
		(it.name, nameLen) = self.Read_nameString(data[:])
		offs = nameLen

		(it.layerType, nameLen) = self.Read_nameString(data[nameLen:])
		offs += nameLen

		it.layerIdx = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]
		offs += RSize.U4

		return(offs)

	#### ----------------
	def Read_Actn(self, data, dataSize, actions):
		u'''ACTN { typeName[S0], itemName[S0], referenceId[U4] :subChunks... }'''
		it = LXO_Object.ActnObject()

		offs = self.Read_Actn_BasePart(it, data)

		while(offs < dataSize):
			(subId, subLen) = self.FnUnpckTAGU2(data[offs:(offs + RSize.TAGU2)])
			offs += RSize.TAGU2

			if(subId == ActnIdT.ITEM):
				self.Read_Actn_ITEM(it, data, offs)
			elif(subId == ActnIdT.CHAN):
				self.Read_Actn_CHAN(it, data, offs)
			elif(subId == ActnIdT.GRAD):
				self.Read_Actn_GRAD(it, data, offs)
			elif(subId == ActnIdT.CHNS):
				self.Read_Actn_CHNS(it, data, offs)
			elif(subId == ActnIdT.PRNT):
				self.Read_Actn_PRNT(it, data, offs)
			elif(subId == ActnIdT.CHNN):
				self.Read_Actn_CHNN(it, data, offs)
			else:
				raise TypeError("Found in ACTN chunk. unknown subChunk {}( {:#08X} )".format(self.FnPackLWID(subId), subId))
				pass

			offs += subLen
		# end of subChunkLoop

		actions.append(it)
#>>
#		self.DebugPrint_Actions(it, len(actions)-1)

	#### ========================================================================
	#### ITEM
	#### ========================================================================
	def Read_Item_PAKG(self, it, data, offs):
		(it.package, nameLen) = self.Read_nameString(data[offs:])
		# 未使用値のU4がある
		# offs += nameLen
		# _ = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]

	#### ----------------
	def Read_Item_XREF(self, it, data, offs):
		it.xrefIdx = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]
		offs += RSize.U4

		(it.xrefPath, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen

		(it.xrefIdentifier, nameLen) = self.Read_nameString(data[offs:])

	#### ----------------
	def Read_Item_LAYR(self, it, data, offs):
		# 規定ではidx:U4. -1として0xFFFFFFFFがあるので敢えて符号付にしておく.
		it.lyrIdx = self.FnUnpckI4(data[offs:(offs + RSize.I4)])[0]
		offs += RSize.I4

		it.lyrMask = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]
		offs += RSize.U4

		it.lyrRGBA = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]

	#### ----------------
	def Read_Item_LINK(self, it, data, offs):
		(name, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen

		linkIndexId = self.FnUnpckI4(data[offs:(offs + RSize.I4)])[0]
		offs += RSize.I4

		linkIndex = self.FnUnpckI4(data[offs:(offs + RSize.I4)])[0]

		if(name not in it.itemLinks):
			it.itemLinks[name] = []
		it.itemLinks[name].append( LXO_Object.iLINK(name, linkIndexId, linkIndex) )

	#### ----------------
	def Read_Item_CLNK(self, it, data, offs):
		(name, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen

		(fromName, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen

		fromId = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]
		offs += RSize.U4

		(toName, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen

		fromLinkId = self.FnUnpckI4(data[offs:(offs + RSize.I4)])[0]
		offs += RSize.I4

		toLinkId = self.FnUnpckI4(data[offs:(offs + RSize.I4)])[0]

		it.chanLinks.append( LXO_Object.iCLNK(name, fromName, fromId, toName, fromLinkId, toLinkId) )

	#### ----------------
	def Read_Item_UCHN(self, it, data, offs):
		(n, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen

		(t, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen

		mode = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]
		offs += RSize.U4

		flag = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]
		offs += RSize.U4

		defInt = self.FnUnpckI4(data[offs:(offs + RSize.I4)])[0]
		offs += RSize.I4

		defFlt = self.FnUnpckF4(data[offs:(offs + RSize.F4)])[0]
		offs += RSize.F4

		hintCnt = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
#		offs += RSize.U2
		#### cnt > 0なら [S0, I4]が続く
		it.userChans.append( LXO_Object.iUCHN(n, t, mode, flag, defInt, defFlt, hintCnt, []) )

	#### ----------------
	def Read_Item_CHNL(self, it, data, offs):
		(n, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen

		t = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
		offs += RSize.U2

		if(t in ITEM_CHNL_Type.INTS):
			v = self.FnUnpckI4(data[offs:(offs + RSize.I4)])[0]
		elif(t in ITEM_CHNL_Type.FLTS):
			v = self.FnUnpckF4(data[offs:(offs + RSize.F4)])[0]
		elif(t in ITEM_CHNL_Type.STRS):
			(v, nameLen) = self.Read_nameString(data[offs:])
		else:
			raise TypeError("CHNL:name:{} is unknown typ {}".format(n, typ))

		it.scalarChans[n] = LXO_Object.iCHNL(n, t, v)

	#### ----------------
	def Read_Item_CHNV(self, it, data, offs):
		(n, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen

		t = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
		offs += RSize.U2

		dimension = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
		offs += RSize.U2

		vals = []
		if(t in ITEM_CHNL_Type.INTS):
			for i in range(dimension):
				(vn, nameLen) = self.Read_nameString(data[offs:])
				offs += nameLen
				v = self.FnUnpckI4(data[offs:(offs + RSize.I4)])[0]
				offs += RSize.I4
				vals.append( (vn, v) )

		elif(t in ITEM_CHNL_Type.FLTS):
			for i in range(dimension):
				(vn, nameLen) = self.Read_nameString(data[offs:])
				offs += nameLen
				v = self.FnUnpckF4(data[offs:(offs + RSize.F4)])[0]
				offs += RSize.F4
				vals.append( (vn, v) )

		elif(t in ITEM_CHNL_Type.STRS):
			for i in range(dimension):
				(vn, nameLen) = self.Read_nameString(data[offs:])
				offs += nameLen
				(v, nameLen) = self.Read_nameString(data[offs:])
				offs += nameLen
				vals.append( (vn, v) )
		else:
			raise TypeError("CHNV:name:{} is unknown typ {}".format(n, t))

		it.vectorChans[n] = LXO_Object.iCHNV(n, t, dimension, vals)

	#### ----------------
	def Read_Item_CHNS(self, it, data, offs):
		(n, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen
		(v, nameLen) = self.Read_nameString(data[offs:])

		it.stringChans[n] = LXO_Object.iCHNS(n, v)

	#### ----------------
	def Read_Item_GRAD(self, it, data, offs):
		(name, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen

		(ENVL_idx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
		offs += vx

		interp = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]

		it.gradientChans[name] =  LXO_Object.iGRAD(name, ENVL_idx, interp)

	#### ----------------
	def Read_Item_CHAN(self, it, data, offs):
		(CHNM_idx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
		offs += vx

		typ = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
		offs += RSize.U2

		# ENV時はACTN.sceneを探る
		if(typ in ITEM_CHAN_Type.INTS):
			v = self.FnUnpckI4(data[offs:(offs + RSize.I4)])[0]
		elif(typ in ITEM_CHAN_Type.FLTS):
			v = self.FnUnpckF4(data[offs:(offs + RSize.F4)])[0]
		elif(typ in ITEM_CHAN_Type.STRS):
			(v, nameLen) = self.Read_nameString(data[offs:])
		else:
			raise TypeError("CHAN:CHNM_idx:{} is unknown typ {}".format(CHNM_idx, typ))

		chnm = self.lwo.channelNames[CHNM_idx]
		it.chanVals[chnm] = LXO_Object.iCHAN(CHNM_idx, typ, v, chnm)

	#### ----------------
	def Read_Item_ITAG(self, it, data, offs):
		itemTag = self.FnUnpckLWID(data[offs:(offs + RSize.LWID)])[0]
		offs += RSize.LWID
		(s, nameLen) = self.Read_nameString(data[offs:])

		lwid = self.FnPackLWID(itemTag)
		it.itemTags.append(LXO_Object.iITAG(itemTag, s, lwid))

	#### ----------------
	def Read_Item_UNIQ(self, it, data, offs):
		(it.uniqueIdentifier, nameLen) = self.Read_nameString(data[offs:])

	#### ----------------
	def Read_Item_UIDX(self, it, data, offs):
		# 規定ではidx:U4. -1として0xFFFFFFFFがあるので敢えて符号付にしておく.
		it.userUniqueIndex = self.FnUnpckI4(data[offs:(offs + RSize.I4)])[0]

	#### ----------------
	def Read_Item_VNAM(self, it, data, offs):
		(it.visibleName, nameLen) = self.Read_nameString(data[offs:])

	#### ----------------
	def Read_Item_BCHN(self, it, data, offs):
		(it.BChanType, nameLen) = self.Read_nameString(data[offs:])
		offs += nameLen
		it.BChanDataId = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]

	#### ----------------
	def Read_Item_BasePart(self, it, data):
		(it.type, nameLen) = self.Read_nameString(data[:])
		offs = nameLen

		(it.name, nameLen) = self.Read_nameString(data[nameLen:])
		offs += nameLen

		it.referenceId = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]
		offs += RSize.U4

		return(offs)

	#### --------------------------------
	def Read_Item(self, data, dataSize, items):
		u'''ITEM { typeName[S0], itemName[S0], referenceId[U4] :subChunks... }'''
		it = LXO_Object.ItemObject()

		offs = self.Read_Item_BasePart(it, data)

		while(offs < dataSize):
			(subId, subLen) = self.FnUnpckTAGU2(data[offs:(offs + RSize.TAGU2)])
			offs += RSize.TAGU2

			if(subId == ItemIdT.PAKG):
				self.Read_Item_PAKG(it, data, offs)

			elif(subId == ItemIdT.XREF):
				self.Read_Item_XREF(it, data, offs)

			elif(subId == ItemIdT.LAYR):
				self.Read_Item_LAYR(it, data, offs)

			elif(subId == ItemIdT.BBOX):
				pass

			elif(subId == ItemIdT.LINK):
				self.Read_Item_LINK(it, data, offs)

			elif(subId == ItemIdT.CLNK):
				self.Read_Item_CLNK(it, data, offs)

			elif(subId == ItemIdT.UCHN):
				self.Read_Item_UCHN(it, data, offs)

			elif(subId == ItemIdT.CHNL):
				self.Read_Item_CHNL(it, data, offs)

			elif(subId == ItemIdT.CHNV):
				self.Read_Item_CHNV(it, data, offs)

			elif(subId == ItemIdT.CHNS):
				self.Read_Item_CHNS(it, data, offs)

			elif(subId == ItemIdT.GRAD):
				self.Read_Item_GRAD(it, data, offs)

			elif(subId == ItemIdT.CHAN):
				self.Read_Item_CHAN(it, data, offs)

			elif(subId == ItemIdT.CHNC):
				pass

			elif(subId == ItemIdT.ITAG):
				self.Read_Item_ITAG(it, data, offs)

			elif(subId == ItemIdT.UNIQ):
				self.Read_Item_UNIQ(it, data, offs)

			elif(subId == ItemIdT.UIDX):
				self.Read_Item_UIDX(it, data, offs)

			elif(subId == ItemIdT.VNAM):
				self.Read_Item_VNAM(it, data, offs)

			elif(subId == ItemIdT.BCHN):
				self.Read_Item_BCHN(it, data, offs)

			else:
				raise TypeError("Found in ITEM chunk. unknown subChunk {}( {:#08X} )".format(self.FnPackLWID(subId), subId))
				pass

			offs += subLen
		# end of subChunkLoop

		items.append(it)
#>>
		self.DebugPrint_Items(it, len(items)-1)

	#### ========================================================================
	def Read_VersionData(self, data, dataSize):
		self.lwo.majorVer = self.FnUnpckU4(data[:RSize.U4])[0]
		self.lwo.minorVer = self.FnUnpckU4(data[:RSize.U4])[0]
		(self.lwo.saverName, nameLen) = self.Read_nameString(data[(RSize.U4 *2):])

		if(self.DEBUG):
			print("{}( {}, {})".format(self.lwo.saverName, self.lwo.majorVer, self.lwo.minorVer))

	#### ========================================================================
	def Read_Encoding(self, data, dataSize):
		self.lwo.encoding = self.FnUnpckU4(data[:RSize.U4])[0]

		if(self.DEBUG):
			print(ENCO_Type.NAMES[self.lwo.encoding])

	#### ========================================================================
	def Read_ChannelNames(self, data, dataSize):
		num = self.FnUnpckU4(data[:RSize.U4])[0]
		names = self.Read_Tags(data[RSize.U4:], dataSize - RSize.U4)
		return(names)

	#### ========================================================================
	def Read_Layr(self, data, dataSize, layers):
		u'''レイヤ情報
		LAYR { number[U2], flags[U2], pivot[VEC12], name[S0], parent[U2]
			以下 lxo要素
			sdsLevel[F4], curveLevel[F4], scalePivot[VEC12], unused[U4*6],
			itemRef[U4], splinePatchLevel[U2], unused[U2*3] }'''
		lyr = LXO_Object.LayerObject()

		(lyr.idx, flags) = self.FnUnpckFLAG(data[:RSize.FLAG])

		lyr.isHide = (flags & 0x1)^1	# bitONで可視(lwoの逆)
		lyr.piv = self.FnUnpckV3(data[RSize.U4:RSize.LYRN])		# (x,y,z)
		(name, nameLen) = self.Read_nameString(data[RSize.LYRN:])

		# レイヤ名が未設定の場合, 規定名+(idx+1).を適用.
		lyr.name = name[:] if(name != "") else "{}_{}".format(lyr.defaultName, (lyr.idx +1))

		offs = (RSize.LYRN + nameLen)
		if(dataSize > offs):
			lyr.parent = self.FnUnpckI2(data[offs:(offs + RSize.I2)])[0]
		offs += RSize.I2

		lyr.sdsLvl = self.FnUnpckF4(data[offs:(offs + RSize.F4)])[0]
		offs += RSize.F4

		lyr.curveLvl = self.FnUnpckF4(data[offs:(offs + RSize.F4)])[0]
		offs += RSize.F4

		scalePiv = self.FnUnpckV3(data[offs:(offs + RSize.VEC3)])		# (x,y,z)
		offs += RSize.VEC3

		offs += (RSize.U4 * 6)

		lyr.itemRef = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]
		offs += RSize.U4

		lyr.splinePatchLvl = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]

		layers.append(lyr)

		self.DebugPrint_Layer(lyr, flags)

	#### ========================================================================
	def Read_Pick_VMap(self, data, dataSize, picks):
		u'''Pick(SelectionSet)/VMap
		VMAP { type[ID4], dimension[U2], name[S0], vert[VX]* }
		PICK, dim[U2], name[S0], vert[VX],,,,'''
		(name, nameLen)= self.Read_nameString(data[RSize.U2:])

		pick = []
		APD = pick.append
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			APD(pntIdx)

		picks[name] = tuple(pick)	# { name:(pntIdx,,,,) }

		self.DebugPrint_Pick(name, pick)

	#### ========================================================================
	def Read_Weight_VMap(self, data, dataSize, weights):
		u'''Weight/VMap
		VMAP { type[ID4], dimension[U2], name[S0],( vert[VX], value[F4] # dimension )* }
		WGHT, dim[U2], name[S0], vert[VX], value[F4]'''
		(name, nameLen)= self.Read_nameString(data[RSize.U2:])

		wgt = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			wgt[pntIdx] = self.FnUnpckF4(data[offs:(offs + RSize.F4)])[0]
			offs += RSize.F4

		weights[name] = wgt

		self.DebugPrint_Weight_VMap(name, wgt)

	#### ========================================================================
	def Read_Morph_VMap(self, data, dataSize, morphs):
		u'''Morph/VMap
		VMAP { type[ID4], dimension[U2], name[S0],( vert[VX], value[F4] # dimension )* }
		MORF, dim[U2], name[S0], vert[VX], value[VEC12]'''
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])

		mrfs = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			mrfs[pntIdx] = self.FnUnpckV3(data[offs:(offs + RSize.VEC3)])	# (x,y,z)
			offs += RSize.VEC3

		morphs[name] = mrfs

		self.DebugPrint_Morph(name, mrfs)

	#### ========================================================================
	def Read_RGB_VMap(self, data, dataSize, lyr):
		u'''VertexColor, RGB/VMap
		VMAP { type[ID4], dimension[U2], name[S0],( vert[VX], value[F4] # dimension )* }
		RGB, dim[U2], name[S0], vert[VX], value[VEC12]'''
		(name, nameLen)= self.Read_nameString(data[RSize.U2:])

		cols = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			cols[pntIdx] = self.FnUnpckV3(data[offs:(offs + RSize.VEC3)]) + (1.0,)	# (r,g,b,A)RGBAにする
			offs += RSize.VEC3

		if(name not in lyr.rgb):
			lyr.rgb[name] = LXO_Object.VMapObject()
		lyr.rgb[name].vmap.update(cols)

		self.DebugPrint_RGBA("RGB/VMap", name, lyr.rgb[name].vmap)

	#### ========================================================================
	def Read_RGBA_VMap(self, data, dataSize, lyr):
		u'''VertexColor, RGBA/VMap
		VMAP { type[ID4], dimension[U2], name[S0],( vert[VX], value[F4] # dimension )* }
		RGBA, dim[U2], name[S0], vert[VX], value[VEC16]'''
		(name, nameLen)= self.Read_nameString(data[RSize.U2:])

		cols = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			cols[pntIdx] = self.FnUnpckV4(data[offs:(offs + RSize.VEC4)])	# (r,g,b,a)
			offs += RSize.VEC4

		if(name not in lyr.rgba):
			lyr.rgba[name] = LXO_Object.VMapObject()
		lyr.rgba[name].vmap.update(cols)

		self.DebugPrint_RGBA("RGBA/VMap", name, lyr.rgba[name].vmap)

	#### ========================================================================
	def Read_RGB_VMad(self, data, dataSize, lyr, curPolIdxOffs):
		u'''VertexColor, RGB/VMad
		VMAD { type[ID4], dimension[U2], name[S0],( vert[VX], poly[VX], value[F4] # dimension )* } 
		RGB, dim[U2], name[S0],( vert[VX], poly[VX], value[VEC12])'''
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])

		cols = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx

			polIdx += curPolIdxOffs	# polIdxはPOLS内でのIndex
			col = self.FnUnpckV3(data[offs:(offs + RSize.VEC3)]) + (1.0,)	# (r,g,b,A)RGBAにする
			offs += RSize.VEC3
			if(polIdx in cols):
				cols[polIdx][pntIdx] = col
			else:
				cols[polIdx] = dict({ pntIdx: col })

		if(name not in lyr.rgb):
			lyr.rgb[name] = LXO_Object.VMapObject()
		lyr.rgb[name].vmad.update(cols)

		self.DebugPrint_RGBA("RGB/VMad", name, lyr.rgb[name].vmad)

	#### ========================================================================
	def Read_RGBA_VMad(self, data, dataSize, lyr, curPolIdxOffs):
		u'''VertexColor(RGB/RGBA)/VMad
		VMAD { type[ID4], dimension[U2], name[S0],( vert[VX], poly[VX], value[F4] # dimension )* } 
		RGBA, dim[U2], name[S0],( vert[VX], poly[VX], value[VEC16])'''
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])

		cols = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx

			polIdx += curPolIdxOffs	# polIdxはPOLS内でのIndex
			col = self.FnUnpckV4(data[offs:(offs + RSize.VEC4)])	# (r,g,b,a)
			offs += RSize.VEC4
			if(polIdx in cols):
				cols[polIdx][pntIdx] = col
			else:
				cols[polIdx] = dict({ pntIdx: col })

		if(name not in lyr.rgba):
			lyr.rgba[name] = LXO_Object.VMapObject()
		lyr.rgba[name].vmad.update(cols)

		self.DebugPrint_RGBA("RGBA/VMad", name, lyr.rgba[name].vmad)

	#### ========================================================================
	def Read_UV_VMap(self, data, dataSize, lyr):
		u'''UV/VMap. {name:{vmap:{pntIdx:[u,v]}}}
		VMAP { type[ID4], dimension[U2], name[S0], ( vert[VX], value[F4] # dimension )* }
		TXUV, dim[U2], name[S0], vert[VX], value[VEC8]'''
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])

		uvs = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			uvs[pntIdx] = self.FnUnpckUV(data[offs:(offs + RSize.UV)])	# (u,v)
			offs += RSize.UV

		if(name not in lyr.uvs):
			lyr.uvs[name] = LXO_Object.VMapObject()
		lyr.uvs[name].vmap.update(uvs)

		self.DebugPrint_UV("VMap", name, lyr.uvs[name].vmap)

	#### ========================================================================
	def Read_UV_VMad(self, data, dataSize, lyr, curPolIdxOffs):
		u'''UV/VMad. {name:{vmad:{polIdx:{pntIdx:[u,v]}}}}
		VMAD { type[ID4], dimension[U2], name[S0],( vert[VX], poly[VX], value[F4] # dimension )* } 
		TXUV,  dim[U2], name[S0],( vert[VX], poly[VX], value[VEC8])'''
		name, nameLen = self.Read_nameString(data[RSize.U2:])

		uvs = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			uvVal = self.FnUnpckUV(data[offs:(offs + RSize.UV)])	# (u,v)
			offs += RSize.UV

			polIdx += curPolIdxOffs	# polIdxはPOLS内でのIndex
			if(polIdx in uvs):	uvs[polIdx][pntIdx] = uvVal
			else:				uvs[polIdx] = dict( { pntIdx: uvVal} )

		if(name not in lyr.uvs):
			lyr.uvs[name] = LXO_Object.VMapObject()
		lyr.uvs[name].vmad.update(uvs)

		self.DebugPrint_UV("VMad", name, lyr.uvs[name].vmad)

	#### ========================================================================
	def Read_Normal_VMad(self, data, dataSize, lyr, curPolIdxOffs):
		u'''NORM/VMad. {name:{vmad:{polIdx:{pntIdx:[u,v]}}}}
		VMAD { type[ID4], dimension[U2], name[S0],( vert[VX], poly[VX], value[F4] # dimension )* } 
		NORM, dim[U2], name[S0],( vert[VX], poly[VX], value[VEC12])'''
		name, nameLen = self.Read_nameString(data[RSize.U2:])

		norm = {}
		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			vec = self.FnUnpckV3(data[offs:(offs + RSize.VEC3)])	# (x,y,z)
			offs += RSize.VEC3

			polIdx += curPolIdxOffs	# polIdxはPOLS内でのIndex

			if(polIdx in norm):	norm[polIdx][pntIdx] = vec
			else:				norm[polIdx] = dict( { pntIdx: vec} )

		lyr.normals[name] = norm

		self.DebugPrint_NORM(name, norm)

	#### ========================================================================
	def Read_Weight_VMad(self, data, dataSize, lyr, curPolIdxOffs):
		u'''Weight/VMad. {{id1 id2:val},{id1 id2:val},...}
		VMAD { type[ID4], dimension[U2], name[S0],( vert[VX], poly[VX], value[F4] # dimension )* } 
		WGHT,  dim[U2], name[S0],( vert[VX], poly[VX], value[F4])
		Some info: LW stores a face's points in a clock-wize order (with the normal pointing at you).
		This gives edges a 'direction' which is used when it comes to storing CC edge weight values.
		The weight is given to the point preceding the edge that the weight belongs to.'''
		(name, nameLen) = self.Read_nameString(data[RSize.U2:])
		if(name != self.sdsEdgeWgtName):
			return

		offs = RSize.U2 + nameLen
		while(offs < dataSize):
			(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			(polIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
			offs += vx
			wgt = self.FnUnpckF4(data[offs:(offs + RSize.F4)])[0]
			offs += RSize.F4

			polPnts = lyr.pols[(polIdx + curPolIdxOffs)]	# polIdxはPOLS内でのIndex
			try:
				idx1 = polPnts.index(pntIdx)
			except:
				continue

			pnt2 = polPnts[0] if(idx1 == len(polPnts) -1) else polPnts[idx1 + 1]
			lyr.edgeWgts[(pnt2, pntIdx)] = wgt

		lyr.hasEdge = True

		self.DebugPrint_EdgeWeight("VMad", name, lyr.edgeWgts)

	#### ========================================================================
	def Read_Pols(self, data, dataSize, lyr, lwid):
		u'''ポリ配列. [(pntIdx,pntIdx,...),(...),...]
		POLS { type[ID4], ( numvert+flags[U2], vert[VX] # numvert )* }'''
		curPolIdxOffs = len(lyr.pols)	# polIdxのoffsetとして.

		pols = []
		PolAPD = pols.append
		offs = 0
		while(offs < dataSize):
			(pntsNum,) = self.FnUnpckU2(data[offs:(offs + RSize.U2)])
			pntsNum &= 0x03FF # 上位6bitはフラグ, max1023
			offs += RSize.U2
			pntsIdxs= []
			PntAPD = pntsIdxs.append
			for i in range(pntsNum):
				(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
				offs += vx
				PntAPD(pntIdx)		# [pntIdx,pntIdx,...]
			PolAPD(tuple(pntsIdxs))	# [(pntIdx,pntIdx,...),(...),...]

		lyr.pols.extend(pols)

		self.DebugPrint_Pols(lyr.name, lwid, pols)

		lyr.polNum += len(pols)	# 累積する

		lyr.lwid = lwid

		return(curPolIdxOffs)

	#### ========================================================================
	def Read_HCrvs(self, data, dataSize, lyr):
		u'''カーブ配列. [(flag, pntIdx,pntIdx,...),(...),...]
		POLS { type[ID4], ( numvert[U2], flags[U4], vert[VX] # numvert )* }'''
		crvs = []
		CrvAPD = crvs.append
		offs = 0
		while(offs < dataSize):
			pntsNum = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
			offs += RSize.U2

			flag = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]
			offs += RSize.U4
			# フラグ ---- ---- ---- ---- ???? ???? 0000 00FF # ----1A03 : 0001 1010 0000 0011
			# F=continuFlag --SE(S=st, E=ed)
			pntsIdxs= [(flag & 0x00000003), ]
			PntAPD = pntsIdxs.append
			for i in range(pntsNum):
				(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
				offs += vx
				PntAPD(pntIdx)		# [flag, pntIdx,pntIdx,...]
			CrvAPD(tuple(pntsIdxs))	# [(flag, pntIdx,pntIdx,...),(...),...]

		lyr.crvs   = tuple(crvs)
		lyr.crvNum += len(crvs)

		self.DebugPrint_Crvs(lyr.name, "HCRV", crvs, len(crvs))

	#### ========================================================================
	def Read_BCrvs(self, data, dataSize, lyr):
		u'''カーブ配列. [(flag, pntIdx,pntIdx,...),(...),...]
		POLS { type[ID4], ( numvert[U2], flags[U4], vert[VX] # numvert )* }'''
		crvs = []
		CrvAPD = crvs.append
		offs = 0
		while(offs < dataSize):
			pntsNum = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
			offs += RSize.U2

			flag = self.FnUnpckU4(data[offs:(offs + RSize.U4)])[0]
			offs += RSize.U4
			# フラグ ---- ---- ---- ---- ???? ???? 0000 0000 # ----1A00 : 0001 1010 0000 0000
			# 閉じてる場合idxが循環している[idx0,1,2,...,0], vt*3+1個
			pntsIdxs= [None, ]
			PntAPD = pntsIdxs.append
			for i in range(pntsNum):
				(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
				offs += vx
				PntAPD(pntIdx)		# [flag, pntIdx,pntIdx,...]
			CrvAPD(tuple(pntsIdxs))	# [(flag, pntIdx,pntIdx,...),(...),...]

		lyr.bcrvs   = tuple(crvs)
		lyr.crvNum += len(crvs)

		self.DebugPrint_Crvs(lyr.name, "BCRV", crvs, len(crvs))

	#### ========================================================================
	def Read_Crvs(self, data, dataSize, lyr):
		u'''カーブ配列. [(flag, pntIdx,pntIdx,...),(...),...]
		POLS { type[ID4], ( numvert+flags[U2], vert[VX] # numvert )* }'''
		# 701で廃止. lwoのCURVと同じかどうかは未確認
		crvs = []
		CrvAPD = crvs.append
		offs = 0
		while(offs < dataSize):
			val = self.FnUnpckU2(data[offs:(offs + RSize.U2)])[0]
			offs += RSize.U2
			# カーブ NNNN FFnn nnnn nnnn, N+n=pntNum, F=continuFlag SE--(S=st, E=ed)
			# -C-- 且つ 同一のPntIdxを指定している場合は閉じたカーブ
			pntsNum = (val & 0x03FF)|((val & 0xF000) >> 2)
			pntsIdxs= [((val & 0x0C00) >> 10), ]
			PntAPD = pntsIdxs.append
			for i in range(pntsNum):
				(pntIdx, vx) = self.Read_VX(data[offs:(offs + RSize.VX)])
				offs += vx
				PntAPD(pntIdx)		# [flag, pntIdx,pntIdx,...]
			CrvAPD(tuple(pntsIdxs))	# [(flag, pntIdx,pntIdx,...),(...),...]

		lyr.crvs   = tuple(crvs)
		lyr.crvNum += len(crvs)

		self.DebugPrint_Crvs(lyr.name, "CURV", crvs, len(crvs))
